/* ************************************************************************* **
** Started Tue Jul 30 16:25:44 NZST 2002				     **
** ************************************************************************* */
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include "glView.h"
#include "glGizmo.h"
#include "gfxTypes.h"

/* ************************************************************************* */
#ifndef M_2PI
#define M_2PI	6.28318530717958647692
#endif//M_2PI

/* ************************************************************************* */
GLuint	Gizmo::_s_list;
bool	Gizmo::_s_isInit(false);
bool	Gizmo::_s_isEasy(true);

/* ************************************************************************* */
Gizmo::Gizmo(const View	&view,
	     Matrix4d	&m): 
  _m_axis(0), _m_view(&view), _m_m(&m)
{
}

// ************************************************************************* //
/// DESCRIPTION: Generates two lists: an arrow and a cube.
// ************************************************************************* //
void	Gizmo::glInitLists() const
{
  GLUquadricObj	*qobj;

  if (!_s_isInit)
    {
      _s_isInit = true;
      _s_list = glGenLists(2);
      glNewList(_s_list, GL_COMPILE);

      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();

      glPushAttrib(GL_LINE_BIT);
      glLineWidth(2.0);
      glPushAttrib(GL_LIGHTING_BIT);
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);
      Point3d(0.0, 0.0, 0.0).glVertex();
      Point3d(0.0, 0.0, 0.8).glVertex();
      glEnd();//GL_LINES
      glPopAttrib();//GL_LIGHTING_BIT
      glPopAttrib();//GL_LINE_BIT

      qobj = gluNewQuadric();
      gluQuadricDrawStyle(qobj, GLU_FILL);
      gluQuadricNormals(qobj, GLU_SMOOTH);
      glTranslated(0.0, 0.0, 0.8);
      glScaled(0.2, 0.2, 0.2);
      gluCylinder(qobj, 0.33, 0.0, 1.0, 16, 1);
      gluDeleteQuadric(qobj);

      qobj = gluNewQuadric();
      gluQuadricDrawStyle(qobj, GLU_FILL);
      gluQuadricNormals(qobj, GLU_SMOOTH);
      gluDisk(qobj, 0.33, 0.0, 16, 1);
      gluDeleteQuadric(qobj);
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
      glEndList();

      glNewList(_s_list + 1, GL_COMPILE);
      glPushAttrib(GL_LIGHTING_BIT);
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glTranslated(0.0, 0.0, 1.3);
      glScaled(0.1, 0.1, 0.1);
      glBegin(GL_QUADS);
      Vector3d(-1.0, 0.0, 0.0).glNormal();
      Point3d(-1.0, -1.0, -1.0).glVertex();
      Point3d(-1.0, -1.0, +1.0).glVertex();
      Point3d(-1.0, +1.0, +1.0).glVertex();
      Point3d(-1.0, +1.0, -1.0).glVertex();
      Vector3d(1.0, 0.0, 0.0).glNormal();
      Point3d(1.0, -1.0, -1.0).glVertex();
      Point3d(1.0, +1.0, -1.0).glVertex();
      Point3d(1.0, +1.0, +1.0).glVertex();
      Point3d(1.0, -1.0, +1.0).glVertex();
      Vector3d(0.0, -1.0, 0.0).glNormal();
      Point3d(-1.0, -1.0, -1.0).glVertex();
      Point3d(+1.0, -1.0, -1.0).glVertex();
      Point3d(+1.0, -1.0, +1.0).glVertex();
      Point3d(-1.0, -1.0, +1.0).glVertex();
      Vector3d(0.0, 1.0, 0.0).glNormal();
      Point3d(-1.0, 1.0, -1.0).glVertex();
      Point3d(-1.0, 1.0, +1.0).glVertex();
      Point3d(+1.0, 1.0, +1.0).glVertex();
      Point3d(+1.0, 1.0, -1.0).glVertex();
      Vector3d(0.0, 0.0, -1.0).glNormal();
      Point3d(-1.0, -1.0, -1.0).glVertex();
      Point3d(-1.0, +1.0, -1.0).glVertex();
      Point3d(+1.0, +1.0, -1.0).glVertex();
      Point3d(+1.0, -1.0, -1.0).glVertex();
      Vector3d(0.0, 0.0, 1.0).glNormal();
      Point3d(-1.0, -1.0, 1.0).glVertex();
      Point3d(+1.0, -1.0, 1.0).glVertex();
      Point3d(+1.0, +1.0, 1.0).glVertex();
      Point3d(-1.0, +1.0, 1.0).glVertex();
      glEnd();
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_LIGHTING_BIT 
      glPopAttrib();//GL_TRANSFORM_BIT
      glEndList();
    }
}

/* ************************************************************************* **
** Description: Returns newly allocated modification.			     **
** _m_axis.....selected feature code of the gizmo.
** ************************************************************************* */
Transformation	*Gizmo::newTransform(const QMouseEvent	&e0,
				     const QMouseEvent	&e1)
{
  switch (_m_axis)
    {
    case 01:
    case 02:
    case 03:
      {
	const Vector3d n(_m_view->getTranslateAlong(e0, e1, _m_m->geto(),
						    getConstrainAxis()));
	return new Translation(n);
      }
    case 04:
      {
	const Vector3d n(_m_view->getTranslate(e0, e1, _m_m->geto()));
	return new Translation(n);
      }
    case 11:
    case 12:
    case 13:
      {
	const Quaternion q(_m_view->getRotateAround(e0, e1, _m_m->geto(),
						    getConstrainAxis()));
	return new Rotation(q, _m_m->geto());
      }
    case 14:
      {
	const Quaternion q(_m_view->getRotate(e0, e1, _m_m->geto()));
	return new Rotation(q, _m_m->geto());
      }
    case 21:
    case 22:
    case 23:
      {
	const double	s(_m_view->getScaleAlong(e0, e1, _m_m->geto(),
						 getConstrainAxis()));
	return new ScaleT(s, _m_m->geto(), getConstrainAxis());
      }
    case 24:
      {
	const double	s(_m_view->getScale(e0, e1));
	return new Scale(s, _m_m->geto());
      }
    default:
      return 0;
   }
}

/* ************************************************************************* **
** Description: Modifies gizmo's matrix.				     **
** ************************************************************************* */
void	Gizmo::transform(const QMouseEvent	&e0,
			 const QMouseEvent	&e1)
{
  switch (_m_axis)
    {
    case 01:
    case 02:
    case 03:
      {
	const Vector3d n(_m_view->getTranslateAlong(e0, e1, _m_m->geto(),
						    getConstrainAxis()));

	*_m_m = Translation(n) * (*_m_m);
	break;
      }
    case 04:
      {
	const Vector3d n(_m_view->getTranslate(e0, e1, _m_m->geto()));

	*_m_m = Translation(n) * (*_m_m);
	break;
      }
    case 11:
    case 12:
    case 13:
      {
	const Quaternion q(_m_view->getRotateAround(e0, e1, _m_m->geto(),
						    getConstrainAxis()));
	*_m_m = Rotation(q, _m_m->geto()) * (*_m_m);
	break;
      }
    case 14:
      {
	const Quaternion q(_m_view->getRotate(e0, e1, _m_m->geto()));

	*_m_m = Rotation(q, _m_m->geto()) * (*_m_m);
	break;
      }
    case 21:
    case 22:
    case 23:
      {
	const double	s(_m_view->getScaleAlong(e0, e1, _m_m->geto(),
						 getConstrainAxis()));
	
	*_m_m = ScaleT(s, _m_m->geto(), getConstrainAxis()) * (*_m_m);
	break;
      }
    case 24:
      {
	const double	s(_m_view->getScale(e0, e1));

	*_m_m = Scale(s, _m_m->geto()) * (*_m_m);
	break;
      }
   }
}

/* ************************************************************************* **
** Description: Returns the active axis constraining the transformation.     **
** ************************************************************************* */
Vector3d	Gizmo::getConstrainAxis() const
{ 
  switch (_m_axis)
    {
    case 01:
    case 11:
    case 21:
      return _m_m->getx().unit() * _m_size;
    case 02:
    case 12:
    case 22:
      return _m_m->gety().unit() * _m_size;
    case 03:
    case 13:
    case 23:
      return _m_m->getz().unit() * _m_size;
    default:
      return 0.0;
    }
}

/* ************************************************************************* */
void	Gizmo::selectTrans() const
{
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  Vector3d	m_x(_m_m->getx().unit()), m_y(_m_m->gety().unit());
  Vector3d	m_z(_m_m->getz().unit());
  Point3d	m_o(_m_m->geto());
  Vector3d	x(_m_view->_m_model.getx()), y(_m_view->_m_model.gety());
  Vector3d	z(_m_view->_m_model.getz());

  x.eqUnit();
  y.eqUnit();
  z.eqUnit();

  Vector3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));

  Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
  glScaled(_m_size, _m_size, _m_size);

  if (!_s_isEasy)
    {
      glLoadName(3);
      glCallList(_s_list);

      glLoadName(2);
      glRotated(-90.0, 1.0, 0.0, 0.0);
      glCallList(_s_list);

      glLoadName(1);
      glRotated(90.0, 0.0, 1.0, 0.0);
      glCallList(_s_list);
    }
  glPopMatrix();//GL_MODELVIEW

  glLoadName(4);

  px *= _m_size / 6.0;
  py *= _m_size / 6.0;
  glBegin(GL_QUADS);
  (_m_m->geto() + px - py).glVertex();
  (_m_m->geto() + px + py).glVertex();
  (_m_m->geto() - px + py).glVertex();
  (_m_m->geto() - px - py).glVertex();
  glEnd();
  glPopAttrib();//GL_TRANSFORM_BIT
}

/* ************************************************************************* */
void	Gizmo::renderTrans() const
{
  Vector3d	m_x(_m_m->getx().unit()), m_y(_m_m->gety().unit());
  Vector3d	m_z(_m_m->getz().unit());
  Point3d	m_o(_m_m->geto());
  Vector3d	x(_m_view->_m_model.getx());
  Vector3d	y(_m_view->_m_model.gety());
  Vector3d	z(_m_view->_m_model.getz());

  x.eqUnit();
  y.eqUnit();
  z.eqUnit();

  Vector3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));

  if (!_s_isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(_m_size, _m_size, _m_size);
      (_m_axis == 3)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 0.0, 1.0);
      glCallList(_s_list);
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (_m_axis == 2)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 0.0);
      glCallList(_s_list);
      glRotated(90.0, 0.0, 1.0, 0.0);
      (_m_axis == 1)? glColor3d(1.0, 1.0, 0.0): glColor3d(1.0, 0.0, 0.0);
      glCallList(_s_list);
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  px *= _m_size / 6.0;
  py *= _m_size / 6.0;
  (_m_axis == 4)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 1.0);
  glBegin(GL_LINE_LOOP);
  (_m_m->geto() + px - py).glVertex();
  (_m_m->geto() + px + py).glVertex();
  (_m_m->geto() - px + py).glVertex();
  (_m_m->geto() - px - py).glVertex();
  glEnd();
  glPopAttrib();//GL_LIGHTING_BIT
  ///////////////////////////////////////////////////////
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_GREATER);

  if (!_s_isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(_m_size, _m_size, _m_size);
      (_m_axis == 3)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.6, 0.7);
      glCallList(_s_list);
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (_m_axis == 2)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.7, 0.6);
      glCallList(_s_list);
      glRotated(90.0, 0.0, 1.0, 0.0);
      (_m_axis == 1)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.7, 0.6, 0.6);
      glCallList(_s_list);
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  (_m_axis == 4)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.7, 0.7);
  glBegin(GL_LINE_LOOP);
  (_m_m->geto() + px - py).glVertex();
  (_m_m->geto() + px + py).glVertex();
  (_m_m->geto() - px + py).glVertex();
  (_m_m->geto() - px - py).glVertex();
  glEnd();
  glPopAttrib();//GL_LIGHTING_BIT
  glPopAttrib();//GL_DEPTH_BUFFER_BIT
}

/* ************************************************************************* */
void	Gizmo::renderCircle(const Vector3d	&x,
			    const Vector3d	&y,
			    const Vector3d	&v) const
{
  unsigned	i;
  double	theta;
  Vector3d	px[32];

  for (theta = 0.0, i = 0; i < 32; theta += M_2PI / 32.0, ++i)
    px[i] = (x * cos(theta) + y * sin(theta)) * _m_size;
  glBegin(GL_LINES);
  for (i = 0; i < 32; ++i)
    if (px[i] * v > 0.0)
      px[i].glVertex(), px[(i + 1) % 32].glVertex();
  glEnd();
  glPushAttrib(GL_LINE_BIT);
  glLineStipple(1, 0xAAAA);
  glEnable(GL_LINE_STIPPLE);
  glBegin(GL_LINES);
  for (i = 0; i < 32; ++i)
    if (px[i] * v < 0.0)
      px[i].glVertex(), px[(i + 1) % 32].glVertex();
  glEnd();
  glDisable(GL_LINE_STIPPLE);
  glPopAttrib();//GL_LINE_BIT
}

/* ************************************************************************* */
void	Gizmo::renderRot() const
{
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  Vector3d	m_x(_m_m->getx().unit()), m_y(_m_m->gety().unit());
  Vector3d	m_z(_m_m->getz().unit());
  Point3d	m_o(_m_m->geto());

  Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
  const Vector3d vv(_m_m->inverse() * 
		    _m_view->unProject(Point3d(0.0, 0.0, 0.0)));

  Vector3d	x(_m_view->_m_model.getx()), y(_m_view->_m_model.gety());
  Vector3d	z(_m_view->_m_model.getz());

  x.eqUnit();
  y.eqUnit();
  z.eqUnit();

  if (!_s_isEasy)
    {
      (_m_axis == 11)? glColor3d(1.0, 1.0, 0.0): glColor3d(1.0, 0.0, 0.0);
      renderCircle(Vector3d(0.0, 1.0, 0.0), Vector3d(0.0, 0.0, 1.0), vv);
      (_m_axis == 12)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 0.0);
      renderCircle(Vector3d(0.0, 0.0, 1.0), Vector3d(1.0, 0.0, 0.0), vv);
      (_m_axis == 13)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 0.0, 1.0);
      renderCircle(Vector3d(1.0, 0.0, 0.0), Vector3d(0.0, 1.0, 0.0), vv);
    }
  glPopMatrix();//GL_MODELVIEW

  Vector3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));
  px *= _m_size * 1.2;
  py *= _m_size * 1.2;
  (_m_axis == 14)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 1.0);
  glBegin(GL_LINE_LOOP);
  for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
    (_m_m->geto() + cos(theta) * px + sin(theta) * py).glVertex();
  glEnd();
  ///////////////////////////////////////////////////////
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_GREATER);

  if (!_s_isEasy)
    {
      glPushMatrix();
      Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
      (_m_axis == 11)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.7, 0.6, 0.6);
      renderCircle(Vector3d(0.0, 1.0, 0.0), Vector3d(0.0, 0.0, 1.0), vv);
      (_m_axis == 12)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.7, 0.6);
      renderCircle(Vector3d(0.0, 0.0, 1.0), Vector3d(1.0, 0.0, 0.0), vv);
      (_m_axis == 13)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.6, 0.7);
      renderCircle(Vector3d(1.0, 0.0, 0.0), Vector3d(0.0, 1.0, 0.0), vv);
      glPopMatrix();//GL_MODELVIEW
    }

  (_m_axis == 14)? glColor3d(0.7, 0.7, 0.6): glColor3d(0.6, 0.7, 0.7);
  glBegin(GL_LINE_LOOP);
  for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
    (_m_m->geto() + cos(theta) * px + sin(theta) * py).glVertex();
  glEnd();
  glPopAttrib();//GL_DEPTH_BUFFER_BIT
  glPopAttrib();//GL_LIGHTING_BIT
  glPopAttrib();//GL_TRANSFORM_BIT
}

/* ************************************************************************* **
** Description: Renders the scaling transformation control features.	     **
** ************************************************************************* */
void	Gizmo::renderScale() const
{
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  Vector3d	x(_m_view->_m_model.getx());
  Vector3d	y(_m_view->_m_model.gety());
  Vector3d	z(_m_view->_m_model.getz());
  Vector3d	m_x(_m_m->getx().unit());
  Vector3d	m_y(_m_m->gety().unit());
  Vector3d	m_z(_m_m->getz().unit());
  Point3d	m_o(_m_m->geto());

  x.eqUnit();
  y.eqUnit();
  z.eqUnit();

  Vector3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));
  px *= _m_size * 1.2;
  py *= _m_size * 1.2;
  glColor3d(0.0, 1.0, 1.0);
  glBegin(GL_LINE_LOOP);
  for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
    (_m_m->geto() + cos(theta) * px + sin(theta) * py).glVertex();
  glEnd();

  const Point3d	tmp0(cos(M_PI / 4.0) * px + sin(M_PI / 4.0) * py);
  const Point3d	tmp1(_m_m->geto() + tmp0 * 1.2);

  glBegin(GL_LINES);
  (_m_m->geto() + tmp0).glVertex();
  tmp1.glVertex();
  glEnd();

  (_m_axis == 24)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 1.0);
  glBegin(GL_QUADS);
  (tmp1 + 0.1 * (px - py)).glVertex();
  (tmp1 + 0.1 * (3.0 * px + py)).glVertex();
  (tmp1 + 0.1 * (px + 3.0 * py)).glVertex();
  (tmp1 + 0.1 * (py - px)).glVertex();
  glEnd();

  glEnable(GL_LIGHTING);

  if (!_s_isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(_m_size, _m_size, _m_size);
      (_m_axis == 23)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 0.0, 1.0);
      glCallList(_s_list + 1);
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (_m_axis == 22)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.0, 1.0, 0.0);
      glCallList(_s_list + 1);
      glRotated(90.0, 0.0, 1.0, 0.0);
      (_m_axis == 21)? glColor3d(1.0, 1.0, 0.0): glColor3d(1.0, 0.0, 0.0);
      glCallList(_s_list + 1);
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  ///////////////////////////////////////////////////////
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_GREATER);
  glDisable(GL_LIGHTING);
  glColor3d(0.0, 0.7, 0.7);
  glBegin(GL_LINE_LOOP);
  for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
    (_m_m->geto() + cos(theta) * px + sin(theta) * py).glVertex();
  glEnd();
  glBegin(GL_LINES);
  (_m_m->geto() + tmp0).glVertex();
  tmp1.glVertex();
  glEnd();
  (_m_axis == 24)? glColor3d(0.7, 0.7, 0.0): glColor3d(0.0, 0.7, 0.7);
  glBegin(GL_QUADS);
  (tmp1 + 0.1 * (px - py)).glVertex();
  (tmp1 + 0.1 * (3.0 * px + py)).glVertex();
  (tmp1 + 0.1 * (px + 3.0 * py)).glVertex();
  (tmp1 + 0.1 * (py - px)).glVertex();
  glEnd();

  glEnable(GL_LIGHTING);

  if (!_s_isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(_m_size, _m_size, _m_size);
      (_m_axis == 23)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.6, 0.6, 0.7);
      glCallList(_s_list + 1);
      glRotated(-90.0, 1.0, 0.0, 0.0);
      (_m_axis == 22)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.6, 0.7, 0.6);
      glCallList(_s_list + 1);
      glRotated(90.0, 0.0, 1.0, 0.0);
      (_m_axis == 21)? glColor3d(1.0, 1.0, 0.0): glColor3d(0.7, 0.6, 0.6);
      glCallList(_s_list + 1);
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  glPopAttrib();//GL_DEPTH_BUFFER_BIT
  glPopAttrib();//GL_LIGHTING_BIT
}

/* ************************************************************************* */
void	Gizmo::selectScale() const
{
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  Vector3d	x(_m_view->_m_model.getx()), y(_m_view->_m_model.gety());
  Vector3d	z(_m_view->_m_model.getz());
  Vector3d	m_x(_m_m->getx().unit()), m_y(_m_m->gety().unit());
  Vector3d	m_z(_m_m->getz().unit());
  Point3d	m_o(_m_m->geto());
  x.eqUnit();
  y.eqUnit();
  z.eqUnit();
  Vector3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));
  px *= _m_size * 1.2;
  py *= _m_size * 1.2;

  const Point3d tmp1(_m_m->geto() + 
		     1.2 * (cos(M_PI / 4.0) * px + sin(M_PI / 4.0) * py));

  glLoadName(24);
  glBegin(GL_QUADS);
  (tmp1 + 0.1 * (px - py)).glVertex();
  (tmp1 + 0.1 * (3.0 * px + py)).glVertex();
  (tmp1 + 0.1 * (px + 3.0 * py)).glVertex();
  (tmp1 + 0.1 * (py - px)).glVertex();
  glEnd();

  if (!_s_isEasy)
    {
      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
      glScaled(_m_size, _m_size, _m_size);
      glLoadName(23);
      glCallList(_s_list + 1);
      glLoadName(22);
      glRotated(-90.0, 1.0, 0.0, 0.0);
      glCallList(_s_list + 1);
      glLoadName(21);
      glRotated(90.0, 0.0, 1.0, 0.0);
      glCallList(_s_list + 1);
      glPopMatrix();
      glPopAttrib();//GL_TRANSFORM_BIT
    }

  glPopAttrib();//GL_LIGHTING_BIT
}

/* ************************************************************************* */
void	Gizmo::selectRot() const
{
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  Vector3d	m_x(_m_m->getx().unit()), m_y(_m_m->gety().unit());
  Vector3d	m_z(_m_m->getz().unit());
  Point3d	m_o(_m_m->geto());

  Matrix4d(m_x, m_y, m_z, m_o).glMultMatrix();
  
  if (!_s_isEasy)
    {
      glLoadName(11);
      glBegin(GL_LINE_LOOP);
      for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
	Point3d(0.0, _m_size * cos(theta), _m_size * sin(theta)).glVertex();
      glEnd();

      glLoadName(12);
      glBegin(GL_LINE_LOOP);
      for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
	Point3d(_m_size * sin(theta), 0.0, _m_size * cos(theta)).glVertex();
      glEnd();

      glLoadName(13);
      glBegin(GL_LINE_LOOP);
      for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
	Point3d(_m_size * cos(theta), _m_size * sin(theta), 0.0).glVertex();
      glEnd();
    }

  glPopMatrix();//GL_MODELVIEW

  Vector3d	x(_m_view->_m_model.getx());
  Vector3d	y(_m_view->_m_model.gety());
  Vector3d	z(_m_view->_m_model.getz());

  x.eqUnit();
  y.eqUnit();
  z.eqUnit();

  Vector3d	px(x(0), y(0), z(0)), py(x(1), y(1), z(1));

  px *= _m_size * 1.2;
  py *= _m_size * 1.2;
  glLoadName(14);
  glBegin(GL_LINE_LOOP);
  for (double theta(0.0); theta < M_2PI; theta += M_2PI / 32.0)
    (_m_m->geto() + cos(theta) * px + sin(theta) * py).glVertex();
  glEnd();
  
  glPopAttrib();//GL_TRANSFORM_BIT
}

/* ************************************************************************* */
void	Gizmo::glSelect() const
{
  selectTrans();
  selectRot();
  selectScale();
}

/* ************************************************************************* */
void	Gizmo::glRender() const
{
  glInitLists();

  const Point3d v0(_m_view->unProject(Point3d(_m_view->_m_view(2) / 2.0,
					      _m_view->_m_view(3) / 2.0,
					      0.0)));
  const Point3d v1(_m_view->unProject(Point3d(_m_view->_m_view(2) / 2.0,
					      _m_view->_m_view(3) / 2.0,
					      1.0)));

  _m_size = (_m_m->geto() - v0) * (v1 - v0).unit() / 10.0;
  renderTrans();
  //if (_s_isEasy)
  //return;
  renderRot();
  renderScale();
}
