/* ************************************************************************* **
** Alexis Angelidis							     **
** Started Mon Jan 28 12:24:46 MET 2002					     **
** ************************************************************************* */
#ifndef __GLTRACKBALL_HH__
#define __GLTRACKBALL_HH__

/* ************************************************************************* */
#ifndef WIN32
extern "C"
{
#endif
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#ifndef WIN32
}
#endif

// ************************************************************************* //
#include <qapplication.h>
#include <qevent.h> 
#include "glView.h"
#include "gfxTypes.h"

// ************************************************************************* //
#ifndef M_2PI
#define M_2PI	6.28318530717958647692
#endif//M_2PI

// ************************************************************************* //
std::ostream &operator<<(std::ostream &os, const TrackBall &v);
std::istream &operator>>(std::istream &is, TrackBall &v);

// ************************************************************************* //
/// DESCRIPTION: TrackBall
/// Note: Far & near clipping planes are set automatically.
// ************************************************************************* //
class TrackBall
{
  friend std::ostream &operator<<(std::ostream &os, const TrackBall &v);
  friend std::istream &operator>>(std::istream &is, TrackBall &v);
protected:
  static GLuint	_s_list;
  Quaternion	_m_rot;
  Point3d	_m_org;
  double	_m_rho, _m_prox;
  const double	_m_rad1, _m_rad2;
  mutable GLdouble	_m_zNear, _m_zFar;
  mutable double	_m_rad, _m_factw, _m_facth;
  mutable View		*_m_v;
  mutable double	_m_winRad;
public:
  TrackBall(View &v);
  void	reset();
  void	resize();
  void	displayTarget() const;
  void	displayCoordSet() const;
  void	glRenderPlace() const;
  void	riPlace(std::ofstream &ofs) const;
  void	glSelectPlace() const;
  void	glRender() const;
  void	glInitLists() const;
  Point2d	screen2normal(const Point2d	&v);
  Point3d	normal2screen(const Point3d	&v);
  void	mouseMove(const QMouseEvent &e0, const QMouseEvent &e1);
  void	mousePressEvent(QMouseEvent *e1);
  GLdouble	getzNear() const { return _m_zNear; }
  GLdouble	getzFar() const { return _m_zFar; }
};

// ************************************************************************* //
inline std::ostream	&operator<<(std::ostream	&os,
				    const TrackBall	&v)
{
  return os << v._m_rot << v._m_org << " " << v._m_rho
	    << " " << v._m_prox
	    << " " << v._m_zNear
	    << " " << v._m_zFar << " " << v._m_rad << " " << v._m_factw
	    << " " << v._m_facth << " " << v._m_winRad;
}

// ************************************************************************* //
inline std::istream	&operator>>(std::istream	&is,
				    TrackBall		&v)
{
  return is >> v._m_rot >> v._m_org >> v._m_rho >> v._m_prox
	    >> v._m_zNear
	    >> v._m_zFar >> v._m_rad >> v._m_factw
	    >> v._m_facth >> v._m_winRad;
}

// __GLTRACKBALL_HH__ ****************************************************** //
#endif
