/* ************************************************************************* **
** Started Fri Jan 18 14:53:54 'MET 2002				     **
** ************************************************************************* */
#ifndef	__GLVIEW_HH__
#define	__GLVIEW_HH__

/* ************************************************************************* */
#include <qevent.h> 
#include "gfxTypes.h"

#define BUFSIZE 512

#ifndef MAXINT
#define MAXINT	2147483647
#endif//MAXINT
#ifndef MAXUINT
#define MAXUINT 0xFFFFFFFF
#endif//MAXUINT
#ifndef MAXFLOAT
#define MAXFLOAT 3.40282347e+38F
#endif//MAXFLOAT

/* ************************************************************************* **
** Description: Holds all the matrices informations for drawing an openGL    **
**   view.								     **
** ************************************************************************* */
class View
{
  friend class TrackBall;
  mutable GLuint	_m_buffer[BUFSIZE];
  mutable GLuint	_m_ret[BUFSIZE];
  mutable GLint		_m_hits;
public:
  mutable Matrix2i	_m_view;
  mutable Matrix4d	_m_proj;
  mutable Matrix4d	_m_model;
public:
  Matrix2i	&view() { return _m_view; }
  Matrix4d	&proj() { return _m_proj; }
  Matrix4d	&model() { return _m_model; }
  const Matrix2i	&view() const { return _m_view; }
  const Matrix4d	&proj() const { return _m_proj; }
  const Matrix4d	&model() const { return _m_model; }
  Vector3d	getTranslate(const QMouseEvent &e0, const QMouseEvent &e1, 
			     const Point3d &o) const;
  Vector3d	getTranslateAlong(const QMouseEvent &e0, const QMouseEvent &e1,
				  const Point3d &o, const Vector3d &x) const;
  Quaternion	getRotate(const QMouseEvent &e0, const QMouseEvent &e1, 
			  const Point3d &o) const;
  Quaternion	getRotateAround(const QMouseEvent &e0, const QMouseEvent &e1,
				const Point3d &o, const Vector3d &x) const;
  double	getScale(const QMouseEvent &e0, const QMouseEvent &e1) const;
  double	getScaleAlong(const QMouseEvent &e0, const QMouseEvent &e1,
			      const Point3d &o, const Vector3d &x) const;
  void		get();
  Point3d	unProject(const Point3d &v) const;
  Point3d	project(const Point3d &v) const;
  void		ortho() const;
  void		glOpen(const int x, const int y) const;
  GLuint	*glClose() const;
  friend std::ostream &operator<<(std::ostream &os, const View &v);
  friend std::istream &operator>>(std::istream &is, View &v);
};

// ************************************************************************* //
inline std::ostream &operator<<(std::ostream	&os,
				const View	&v)
{
  return os << v._m_view << " " << v._m_proj << " " << v._m_model;
}

// ************************************************************************* //
inline std::istream &operator>>(std::istream	&is,
				View		&v)
{
  return is >> v._m_view >> v._m_proj >> v._m_model;
}

/* ************************************************************************* **
** Description: Returns translation vector in plane at depth winz.	     **
** ************************************************************************* */
inline Vector3d	View::getTranslate(const QMouseEvent	&e0,
				   const QMouseEvent	&e1,
				   const Point3d	&o) const
{
  const double	winz(project(o)(2));
  Point3d	w0(e0.x(), _m_view(3) - e0.y(), winz);
  Point3d	w1(e1.x(), _m_view(3) - e1.y(), winz);

  w0 = unProject(w0);
  w1 = unProject(w1);
  return w1 - w0;
}

// ************************************************************************* //
/// DESCRIPTION: Returns translation vector along a line.
// ************************************************************************* //
inline Vector3d	View::getTranslateAlong(const QMouseEvent	&e0,
					const QMouseEvent	&e1,
					const Point3d		&o,
					const Vector3d		&x) const
{
  if ((x(0) == 0.0 && x(1) == 0.0 && x(2) == 0.0) ||
      (e1.x() == e0.x() && e1.y() == e0.y()))
    return Vector3d(0.0);

  Point2d	op, oxp;
  Point2d	w0(e0.x(), _m_view(3) - e0.y());
  Point2d	w1(e1.x(), _m_view(3) - e1.y());

  op = Point2d(project(o));
  oxp = Point2d(project(o + x));
  w0 = w0.proj2line(op, oxp);
  w1 = w1.proj2line(op, oxp);
  return x * ((w1 - w0) * (oxp - op) / (oxp - op).norm2());
}

/* ************************************************************************* **
** Description: Returns the rotation quaternion.			     **
** ************************************************************************* */
inline Quaternion	View::getRotate(const QMouseEvent	&e0,
					const QMouseEvent	&e1, 
					const Point3d		&o) const
{
  Point2d	op(project(o));
  Point2d	w0p(e0.x(), _m_view(3) - e0.y());
  Point2d	w1p(e1.x(), _m_view(3) - e1.y());
  Point3d	eye(unProject(Point3d(_m_view(2) * 0.5,
				       _m_view(3) * 0.5, 0.0)));
  double	angle(myacos((w0p - op).unit() * (w1p - op).unit()));
  
  if (((w1p - op)(0) * (w0p - op)(1) - (w1p - op)(1) * (w0p - op)(0)) < 0.0)
    return Quaternion::creator(-angle, o - eye);
  return Quaternion::creator(angle, o - eye);
}

/* ************************************************************************* **
** Description: Returns the quaternion of rotation around axis x, center o. **
** ************************************************************************* */
inline Quaternion	View::getRotateAround(const QMouseEvent	&e0,
					      const QMouseEvent	&e1,
					      const Point3d	&o,
					      const Vector3d	&x) const
{
  if ((x(0) == 0.0 && x(1) == 0.0 && x(2) == 0.0) ||
      (e1.x() == e0.x() && e1.y() == e0.y()))
    return Quaternion(1.0, 0.0);

  Point2d	op(project(o));
  Point2d	w0p(e0.x(), _m_view(3) - e0.y());
  Point2d	w1p(e1.x(), _m_view(3) - e1.y());
  Point3d	eye(unProject(Point3d(_m_view(2) * 0.5,
				      _m_view(3) * 0.5, 0.0)));
  Point3d	w0(unProject(Point3d(w0p(0), w0p(1), 1.0)));
  Point3d	w1(unProject(Point3d(w1p(0), w1p(1), 1.0)));
  
  Vector3d	n0 = (w0.proj2planeAlong(o, x, Vector3d(w0 - eye)) - o).unit();
  Vector3d	n1 = (w1.proj2planeAlong(o, x, Vector3d(w1 - eye)) - o).unit();

  if (((w1p(0) - op(0)) * (w0p(1) - op(1)) - 
       (w1p(1) - op(1)) * (w0p(0) - op(0))) * (x * (eye - o)) < 0.0)
    return Quaternion::creator(myacos(n1 * n0), x);
  return Quaternion::creator(-myacos(n1 * n0), x);
}

/* ************************************************************************* */
inline double	View::getScale(const QMouseEvent &e0,
			       const QMouseEvent &e1) const
{
  Vector2d	v(e1.x() - e0.x(), e0.y() - e1.y());
  Vector2d	dir(1.0, 1.0);

  if (0 <= v * dir)
    return 1.0 + sqrt(v * v) / max(_m_view(2), _m_view(3));
  return 1.0 - sqrt(v * v) / max(_m_view(2), _m_view(3));
}

/* ************************************************************************* */
inline double	View::getScaleAlong(const QMouseEvent	&e0,
				    const QMouseEvent	&e1,
				    const Point3d	&o,
				    const Vector3d	&x) const
{
  if ((x(0) == 0.0 && x(1) == 0.0 && x(2) == 0.0) ||
      (e1.x() == e0.x() && e1.y() == e0.y()))
    return 1.0;

  Point2d	op, oxp;
  Point2d	w0(e0.x(), _m_view(3) - e0.y());
  Point2d	w1(e1.x(), _m_view(3) - e1.y());
  double	s;

  op = Point2d(project(o));
  oxp = Point2d(project(o + x));
  w0 = w0.proj2line(op, oxp);
  w1 = w1.proj2line(op, oxp);
  s = 1.0 + (w1 - w0) * (oxp - op) / (oxp - op).norm2();
  if (1.0e-6 < s)
    return s;
  return 1.0e-6;
}

/* ************************************************************************* */
inline void	View::get() 
{
  _m_model.glGet(GL_MODELVIEW_MATRIX);
  _m_proj.glGet(GL_PROJECTION_MATRIX);
  _m_view.glGet(GL_VIEWPORT);
}

/* ************************************************************************* */
inline Point3d	View::unProject(const Point3d &v) const
{
  Point3d	ret;
  
  gluUnProject(v(0), v(1), v(2), _m_model._m_m, _m_proj._m_m,
	       _m_view._m_m, &ret(0), &ret(1), &ret(2));
  return ret;
}

/* ************************************************************************* */
inline Point3d	View::project(const Point3d &v) const
{
  Point3d	ret;
  
  gluProject(v(0), v(1), v(2), _m_model._m_m, _m_proj._m_m,
	     _m_view._m_m, &ret(0), &ret(1), &ret(2));
  return ret;
}

/* ************************************************************************* */
inline void	View::ortho() const
{
  glOrtho(0.0, _m_view(2), 0.0, _m_view(3), -1.0, 1.0);
}

/* ************************************************************************* */
inline void	View::glOpen(const int	x,
			     const int	y) const
{
  //Matrix2i	view(GL_VIEWPORT);

  glSelectBuffer(BUFSIZE, _m_buffer);
  glRenderMode(GL_SELECT);
  glInitNames();
  glPushName(0);
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPickMatrix(x, y, 5.0, 5.0, _m_view._m_m);
  glPopAttrib();// GL_TRANSFORM_BIT
  glClear(GL_DEPTH_BUFFER_BIT);
}

// ************************************************************************* //
/// DESCRIPTION: returns array [nb, near depth, far depth, name0, ... , 0]
// ************************************************************************* //
inline GLuint	*View::glClose() const
{
  glFlush();
  if ((_m_hits = glRenderMode(GL_RENDER)) < 1)
    return 0;

  const GLuint	*ptr(_m_buffer);
  const GLuint	*retptr(0);
  GLuint	retz(MAXUINT);
  GLuint	retnames(0);

  for (int i(0); i < _m_hits; ++i, ptr += 3 + *ptr)
    if (ptr[1] <= retz)
      {
	retnames = ptr[0];
	retz = ptr[1];
	retptr = ptr;
      }
  for (GLuint i(0); i < retnames + 3; ++i)
    _m_ret[i] = retptr[i];
  return _m_ret;
}

// ************************************************************************* //
#endif//__GLVIEW_HH__



