// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Thu Mar 20 16:31:04 NZST 2003
// ************************************************************************* //
#include <list>
#include <vector>
#include <iostream>
#include <fstream>
#include <string>

// ************************************************************************* //
#include <math.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qinputdialog.h> 
#include <qtimer.h>
#include <qgl.h>
#include <qcolordialog.h> 
#include <GL/glu.h>
#include "qtGlWidget.h"
#include "swpSurfel.h"
#include "swpShape.h"
#include "swpSweep.h"
#include "swpTool.h"
#include "swpToolPinch.h"
#include "swpToolStrangle.h"
#include "swpToolVolume.h"
#include "swpToolProto.h"
#include "swpToolZoom.h"
#include "swpToolStokes.h"
#include "gfxTypes.h"
#include "stpetersProbe.dat"

// ************************************************************************* //
GLuint	MyGLWidget::_s_envMap(0);

// ************************************************************************* //
QString	SaveThread::getNum(const QString	name,
			   unsigned		i,
			   const unsigned	base)
{
  QString	str((QChar*)0, base);

  //str.setLength(base);
  for (int k(base - 1); 0 <= k; --k)
    {
      str[k] = i % 10 + '0';
      i /= 10;
    }
  return name + str + QString(".jpeg");
}

// ************************************************************************* //
void SaveThread::run()
{
  for (unsigned i(0); i < _m_size; ++i)
    if (!_m_imageData[i].save(getNum("anim", _m_imageIds[i], 4), "JPEG", 100))
      std::cerr << "Error while saving " 
		<< getNum("file", _m_imageIds[i], 4) << std::endl;
  if (_m_imageData)
    {
      delete[] _m_imageData;
      _m_imageData = 0;
    }
  if (_m_imageIds)
    {
      delete[] _m_imageIds;
      _m_imageIds = 0;
    }
}

// ************************************************************************* //
SaveThread::SaveThread(const QImage	*imageData,
		       const unsigned	*imageIds,
		       const unsigned	size): QThread(), _m_size(size)
{
  _m_imageIds = new unsigned[size];
  _m_imageData = new QImage[size];
  for (unsigned i(0); i < size; ++i)
    _m_imageData[i] = imageData[i], _m_imageIds[i] = imageIds[i];
}

// ************************************************************************* //
SaveThread::~SaveThread()
{
  if (_m_imageData)
    delete[] _m_imageData;
  if (_m_imageIds)
    delete[] _m_imageIds;
}

// ************************************************************************* //
MyGLWidget::MyGLWidget(QWidget		*parent,
		       const char	*name):
  QGLWidget(parent, name),
  _m_trackball(_m_view),
  _m_gizmo(_m_view, _m_mGizmo),
  _m_mGizmo(Translation(Vector3d(2.0, 0.0, 0.0)) * Scale(0.5, Point3d(0.0))),
  _m_isInViewMode(false),
  _m_e0(QEvent::MouseButtonRelease, QPoint(0, 0), Qt::NoButton, Qt::NoButton),
  _m_isGouraudWire(false), _m_isShowGrid(false), _m_isUndoDisabled(true),
  _m_isEnvMap(false),
  _m_isToolColor(true), _m_sym(0), _m_fps(12), _m_iLocal(0),
  _m_iGlobal(0), _m_size(16),
  _m_imageData(new QImage[_m_size]), _m_imageIds(new unsigned[_m_size]),
  _m_isRecording(false)
{
  _m_curSweep = 0;
  _m_shape = new Shape();
  _m_shapeUndo = new Shape();
  _m_shape->initBall();
  _m_shapeUndo->initBall();
  setFocusPolicy(QWidget::StrongFocus);
  //_m_tools.push_back(new ToolBall(1.0));
  _m_tools.push_back(new ToolVolume(1.0));
  _m_tools.front()->setPosition(_m_mGizmo);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
}

// ************************************************************************* //
MyGLWidget::~MyGLWidget()
{
  std::list<SaveThread *>::iterator	mtit, mtend(_m_saveThreads.end());

  std::cout << "Please wait for all threads to terminate..." << std::flush;
  for (mtit = _m_saveThreads.begin(); mtit != mtend; ++mtit)
    (*mtit)->wait();
  std::cout << "\rThank you for using Sweepers 3.0.          " << std::endl;

  delete _m_shapeUndo;
  delete _m_shape;
  delete[] _m_imageData;
  delete[] _m_imageIds;

  std::list<Tool *>::iterator	tit, tend(_m_tools.end());
  
  for (tit = _m_tools.begin(); tit != tend; ++tit)
    delete *tit;
}

// ************************************************************************* //
void	MyGLWidget::glSelect() const
{
  std::list<Tool *>::const_iterator	tit, tend(_m_tools.end());
  unsigned	iname(0);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  if (_m_sym == 0)
    for (tit = _m_tools.begin(); tit != tend; ++tit)
      {
	glLoadName(++iname);
	SweepSingle::glRender(**tit);
      }
  else
    for (tit = _m_tools.begin(); tit != tend; ++tit)
      {
	glLoadName(++iname);
	SweepSym::glRender(**tit, _m_sym);
      }
  glFlush();
}

// ************************************************************************* //
void	MyGLWidget::initMaterials()
{
  const GLfloat	specular[] = {1.0, 1.0, 0.8, 0.2};
  const GLfloat	shininess[] = {50.0};

  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);

  if (!_s_envMap)
    {
      glGenTextures(1, &_s_envMap);
      glBindTexture(GL_TEXTURE_2D, _s_envMap);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 256, 0, GL_RGB,
		   GL_UNSIGNED_BYTE, stpetersProbe_dat);
    }
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);

  glShadeModel(GL_SMOOTH);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
}

// ************************************************************************* //
void	MyGLWidget::paintGL()
{
  std::list<Tool *>::iterator	tit, tend(_m_tools.end());

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  _m_trackball.glRenderPlace();

  glColor3d(0.7, 1.0, 0.2);

  if (_m_sym == 0)
    for (tit = _m_tools.begin(); tit != tend; ++tit)
      SweepSingle::glRender(**tit);
  else
    for (tit = _m_tools.begin(); tit != tend; ++tit)
      SweepSym::glRender(**tit, _m_sym);

  if (_m_isGouraudWire)
    {
      glPolygonMode(GL_FRONT, GL_LINE);
      glColor3d(0.0, 0.0, 0.0);
      _m_shape->glRender();
      glEnable(GL_POLYGON_OFFSET_FILL);
      glPolygonOffset(1.0f, 1.0f);
      glPolygonMode(GL_FRONT, GL_FILL);
    }

  if (_m_isEnvMap)
    {
      initMaterials();
      glEnable(GL_TEXTURE_2D);
      glEnable(GL_TEXTURE_GEN_S);
      glEnable(GL_TEXTURE_GEN_T);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glBindTexture(GL_TEXTURE_2D, _s_envMap);
    }
  if (_m_isToolColor)
    _m_shape->glRender2();
  else
    _m_shape->glRender3();
  if (_m_isEnvMap)
    {
      glDisable(GL_TEXTURE_GEN_T);
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_2D);
    }
  if (_m_isShowGrid)
    {
      glDisable(GL_LIGHTING);
      Point3d(0.5).glColor();
      glBegin(GL_LINES);
      for (double x(-4.0); x <= 4.0; x += 0.3333)
	if (0.01< fabs(x))
	  Point3d(x, -4.0, 0.0).glVertex(), Point3d(x, 4.0, 0.0).glVertex();
      for (double y(-4.0); y <= 4.0; y += 0.3333)
	if (0.01< fabs(y))
	  Point3d(-4.0, y, 0.0).glVertex(), Point3d(4.0, y, 0.0).glVertex();
      glEnd();//GL_LINES
      //glLineWidth(2.0f);
      glBegin(GL_LINES);
      Point3d(0.0).glColor();
      Point3d(0.0, -4.0, 0.0).glVertex(), Point3d(0.0, 4.0, 0.0).glVertex();
      Point3d(-4.0, 0.0, 0.0).glVertex(), Point3d(4.0, 0.0, 0.0).glVertex();
      glEnd();//GL_LINES
      glLineWidth(1.0f);
      glEnable(GL_LIGHTING);
    }

  if (_m_isInViewMode)
    _m_trackball.glRender();
  else
    _m_gizmo.glRender();
  glFlush();
}

// ************************************************************************* //
void	MyGLWidget::keyPressEvent(QKeyEvent *e)
{
  if (e->key() == Qt::Key_Space)
    _m_isInViewMode = !_m_isInViewMode;
  else if (e->key() == Qt::Key_Delete)
    {
      if (1 < _m_tools.size())
	{
	  delete _m_tools.front();
	  _m_tools.pop_front();
	  _m_mGizmo = _m_tools.front()->getPosition();
	  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
	}
    }
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::toggleRecord(bool b)
{
  _m_isRecording = b;
  if (b == true)
    _m_timerId = startTimer(1000 / _m_fps);//DEBUG
  else
    { 
      killTimer(_m_timerId);
      _m_saveThreads.push_back(new SaveThread(_m_imageData, _m_imageIds,
					      _m_iLocal + 1));
      _m_saveThreads.back()->start();
      _m_saveThreads.clear();
      _m_iLocal = 0;
      std::cout << "Last index is " << _m_iGlobal << std::endl;
    }
}

// is implemented in Qt3
QImage MyGLWidget::grabFrameBuffer ( bool withAlpha)
{
	QImage img;
	return img;
}

// ************************************************************************* //
void MyGLWidget::timerEvent(QTimerEvent *)
{
  if (_m_iLocal < _m_size)
    {
      _m_imageIds[_m_iLocal] = _m_iGlobal;
      // only in Qt3
	  _m_imageData[_m_iLocal] = grabFrameBuffer();
    }
  if (++_m_iLocal == _m_size)
    {
      _m_saveThreads.push_back(new SaveThread(_m_imageData, _m_imageIds,
					      _m_size));
      _m_saveThreads.back()->start();
      _m_iLocal = 0;
    }
  ++_m_iGlobal;
}
    
// ************************************************************************* //
void	MyGLWidget::mousePressEvent(QMouseEvent *e)
{
  const GLuint	*buf;

  if (_m_isInViewMode)
    _m_trackball.mousePressEvent(e);
  else if (e->state() & ControlButton)
    {
      _m_view.glOpen(e->x(), _m_view.view()(3) - e->y());
      _m_trackball.glSelectPlace();
      glSelect();
      buf = _m_view.glClose();
      select(buf? buf[3]: 0);
    }
  else
    {
      _m_view.glOpen(e->x(), _m_view.view()(3) - e->y());
      _m_trackball.glSelectPlace();
      _m_gizmo.glSelect();
      buf = _m_view.glClose();
      _m_gizmo.select(buf? buf[3]: 0);
    }
  _m_e0 = *e;
  updateGL();
}

// ************************************************************************* //
/// DESCRIPTION: A non-realtime move cannot be applied twice at the same
///  place... move the tool please.
// ************************************************************************* //
void	MyGLWidget::mouseMoveEvent(QMouseEvent *e)
{
  //TODO: convert rotation in translation.
  Transformation	*t;
  Sweep			*swp;
  static bool		saveFlag(false);
  static bool		processingSweep(false);
  //const clock_t		t0(clock());

  if (_m_isRecording)
    qApp->processOneEvent();
  if (_m_isInViewMode)
    _m_trackball.mouseMove(_m_e0, *e);
  else if (!(e->state() & ControlButton))
    if (e->state() & Qt::LeftButton)
      {
	if (saveFlag == false && !_m_isUndoDisabled)
	  {
	    saveFlag = true;
	    _m_shapeUndo->copy(*_m_shape);
	  }
	if (_m_curSweep)
	  {
	    _m_curSweep->sweepStep(*_m_shape);
	    if (_m_curSweep->isDone())
	      delete _m_curSweep, _m_curSweep = 0;
	  }
	else if ((!processingSweep) && (t = _m_gizmo.newTransform(_m_e0, *e)))
	  {
	    if ((swp = _m_tools.front()->newSweep(_m_sym, _m_mGizmo, *t)))
	      if (swp->isRealTime())
		{
		  swp->sweep(*_m_shape);
		  if (swp->transformGizmo())
		    {
		      swp->transform(_m_mGizmo);
		      _m_tools.front()->setPosition(_m_mGizmo);
		    }
		  delete swp;
		}
	      else
		{
		  if (swp->beginSweep(*_m_shape))
		    {
		      _m_curSweep = swp;
		      processingSweep = true;
		    }
		  else
		    delete swp;
		}
	    delete t;
	  }
	_m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
      }
    else 
      {
	saveFlag = false;
	processingSweep = false;
	_m_gizmo.transform(_m_e0, *e);
	_m_tools.front()->setPosition(_m_mGizmo);
	_m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
      }
  _m_e0 = *e;
  //std::cout << (clock() - t0) / double(CLOCKS_PER_SEC) << std::endl;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::resetView()
{
  _m_trackball.reset();
  updateGL();
}

// ************************************************************************* //
bool	MyGLWidget::saveView()
{
  QString s = QFileDialog::getSaveFileName(".", "View (*.vu)",
					   this, "save file dialog",
					   "Save to file");
  if (s == 0)
    return false;
  if (!s.contains(".vu"))
    s = s + ".vu";

  std::ofstream	ofs(s, std::ofstream::binary);

  if (!ofs)
    return false;
  ofs << _m_trackball;
  return true;
}

// ************************************************************************* //
bool	MyGLWidget::loadView()
{
  QString s = QFileDialog::getOpenFileName(".", "View (*.vu)",
					   this, "open file dialog",
					   "Load file");
  if (s == 0)
    return false;

  std::ifstream	ifs(s, std::ifstream::binary);

  if (!ifs)
    return false;
  ifs >> _m_trackball;
  updateGL();
  return true;
}

// ************************************************************************* //
void	MyGLWidget::resetTool()
{
  _m_mGizmo = Translation(Vector3d(2.0, 0.0, 0.0)) * Scale(0.5, Point3d(0.0));
  _m_tools.front()->setPosition(_m_mGizmo);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::centerTool()
{
  _m_mGizmo = Matrix4d::identity();
  _m_tools.front()->setPosition(_m_mGizmo);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::resetShape()
{
  _m_shape->reset();
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  if (_m_curSweep)
    delete _m_curSweep, _m_curSweep = 0;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::reset()
{
  _m_sym = 0;
  _m_trackball.reset();
  _m_mGizmo = Translation(Vector3d(2.0, 0.0, 0.0)) * Scale(0.5, Point3d(0.0));
  _m_tools.front()->setPosition(_m_mGizmo);
  _m_shape->reset();
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::cleanShape()
{
  _m_shape->cleanShape();
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::renormalize()
{
  _m_shape->renormalize();
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::setGouraud()
{
  glPolygonMode(GL_FRONT, GL_FILL);
  glEnable(GL_CULL_FACE);
  _m_isGouraudWire = false;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::setWireframe()
{
  glPolygonMode(GL_FRONT, GL_LINE);
  glEnable(GL_CULL_FACE);
  _m_isGouraudWire = false;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::setSolidwires()
{
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glDisable(GL_CULL_FACE);
  _m_isGouraudWire = false;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::setGouraudWire()
{
  glPolygonMode(GL_FRONT, GL_FILL);
  glEnable(GL_CULL_FACE);
  _m_isGouraudWire = true;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::toggleEasyTool(bool b)
{
  Gizmo::setEasy(b);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::toggleShowBox(bool b)
{
  _m_isShowGrid = b;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::toggleEmptyTool(bool b)
{
  Tool::setEmpty(b);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::toggleDisableUndo(bool b)
{
  _m_isUndoDisabled = b;
}

// ************************************************************************* //
void	MyGLWidget::toggleEnvMap(bool b)
{
  _m_isEnvMap = b;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::toggleZones(bool b)
{
  _m_isToolColor = b;
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::setActiveToolCoating(int k)
{
  _m_tools.front()->setCoating(k / 100.0);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::setSymmetry(int s)
{
  _m_sym = s;
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::setShapeColor()
{
  const QColor	initial(int(255 * Surfel::_s_color(0)),
			int(255 * Surfel::_s_color(1)), 
			int(255 * Surfel::_s_color(2)));
  const QColor	chosen = QColorDialog::getColor(initial, this,
						"Color Chooser");
  if (chosen.isValid())
    Surfel::_s_color = Vector3d(chosen.red() / 255.0,
				chosen.green() / 255.0,
				chosen.blue() / 255.0);
}

// ************************************************************************* //
bool	MyGLWidget::renderShape()
{
  QString	sView, sShape;
  QString	s = QFileDialog::getSaveFileName(".", "Renderman (*.rib)",
						 this, "save file dialog",
						 "Save to file");
  if (s == 0)
    return false;
  if (!s.contains(".rib"))
    s = s + ".rib";

  sView = s;
  sView.truncate(s.length() - 4);
  sView = sView + "View.rib";
  std::cout << sView << std::endl;

  sShape = s;
  sShape.truncate(s.length() - 4);
  sShape = sShape + "Shape.rib";
  std::cout << sShape << std::endl;

  std::cout << s << std::endl;
  
  std::ofstream	ofs(s, std::ofstream::binary);
  std::ofstream	vofs(sView, std::ofstream::binary);
  std::ofstream	sofs(sShape, std::ofstream::binary);

  if (!ofs || !vofs || !sofs)
    return false;
  ofs <<
    "LightSource \"distantlight\" 1\n"
    "Identity\n"
    "Projection \"perspective\" \"fov\" 50 \n"
    "ReadArchive " << "\"" << sView << "\"\n" <<
    "Display \"ri.tif\" \"tiff\" \"rgba\"\n"
    "WorldBegin\n"
    "Surface \"matte\"\n"
    "Color .2 .6 .2\n"
    "ReadArchive "  << "\"" << sShape << "\"\n" <<
    "WorldEnd\n" << std::flush;
  _m_trackball.riPlace(vofs);
  _m_shape->riRender(sofs);
  return true;
}

// ************************************************************************* //
bool	MyGLWidget::saveShape()
{
  QString s = QFileDialog::getSaveFileName(".", "Sweeper shape (*.sws)",
					   this, "save file dialog",
					   "Save to file");
  if (s == 0)
    return false;
  if (!s.contains(".sws"))
    s = s + ".sws";
  std::ofstream	ofs(s, std::ofstream::binary);

  if (!ofs)
    return false;
  _m_shape->save(ofs);
  return true;
}

// ************************************************************************* //
bool	MyGLWidget::loadShape()
{
  QString s = QFileDialog::getOpenFileName(".", "Sweeper shape (*.sws)",
					   this, "open file dialog",
					   "Load file");
  if (s == 0)
    return false;

  std::ifstream	ifs(s, std::ifstream::binary);

  if (!ifs)
    return false;
  _m_shape->load(ifs);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  return true;
}

// ************************************************************************* //
bool	MyGLWidget::exportShape()
{
  QString s = QFileDialog::getSaveFileName(".", "VRML (*.wrl)",
					   this, "export file dialog",
					   "Export to file");
  if (s == 0)
    return false;
  if (!s.contains(".wrl"))
    s = s + ".wrl";
  
  std::ofstream	ofs(s, std::ofstream::binary);

  if (!ofs)
    return false;
  _m_shape->exportVRML(ofs);
  return true;
}

// ************************************************************************* //
void	MyGLWidget::undo()
{
  if (_m_isUndoDisabled)
    return;

  Shape	*shape;

  shape = _m_shape;
  _m_shape = _m_shapeUndo;
  _m_shapeUndo = shape;
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::select(int id)
{
  if (id == 0)
    return;

  std::list<Tool *>::iterator	tit(_m_tools.begin());
  Tool	*tswap;

  std::cout << id << std::endl;
  //return;
  for (int i = 1; i < id; ++i, ++tit)
    ;
  tswap = _m_tools.front();
  _m_tools.front() = *tit;
  *tit = tswap;
  _m_mGizmo = _m_tools.front()->getPosition();
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::selectTool(int id)
{
  std::cout << "single select tool " << std::endl;

  delete _m_tools.front();
  if (id == 0)
    _m_tools.front() = new ToolBall(1.0);
  else if (id == 1)
    _m_tools.front() = new ToolPinch(1.0);
  else if (id == 2)
    _m_tools.front() = new ToolStrangle(1.0);
  else if (id == 3)
    _m_tools.front() = new ToolVolume(1.0);
  else if (id == 4)
    _m_tools.front() = new ToolProto(1.0);
  else if (id == 5)
    _m_tools.front() = new ToolZoom(1.0);
  else if (id == 6)
    _m_tools.front() = new ToolStokes(1.0);
  _m_tools.front()->setPosition(_m_mGizmo);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::selectMultipleTool(int id)
{
  std::cout << "multi select tool " << std::endl;
  if (id == 0)
    _m_tools.push_front(new ToolBall(1.0));
  else if (id == 1)
    _m_tools.push_front(new ToolPinch(1.0));
  else if (id == 2)
    _m_tools.push_front(new ToolStrangle(1.0));
  else if (id == 3)
    _m_tools.push_front(new ToolVolume(1.0));
  else if (id == 4)
    _m_tools.push_front(new ToolProto(1.0));
  else if (id == 5)
    _m_tools.push_front(new ToolZoom(1.0));
  else if (id == 5)
    _m_tools.push_front(new ToolStokes(1.0));
  _m_tools.front()->setPosition(_m_mGizmo);
  _m_tools.front()->colorize(_m_mGizmo, *_m_shape, _m_sym);
  updateGL();
}

// ************************************************************************* //
void	MyGLWidget::resetRecordIndex()
{
  _m_iGlobal = _m_iLocal = 0;
}

// ************************************************************************* //
void	MyGLWidget::setRecordIndex()
{
  bool		ok(FALSE);
  const int	res = 
    QInputDialog::getInteger(tr("Application name"),
			     tr("Please enter a number"),
			     _m_iGlobal, -2147483647, 2147483647,
			     1, &ok, this);

  if (ok)
    {
      _m_iLocal = 0;
      _m_iGlobal = res;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Prints the shape's stats.
// ************************************************************************* //
void	MyGLWidget::shapeInformation()
{
  QString	num0, num1, num2, num3;
  const double	vol(_m_shape->volume());

  num0.setNum(_m_shape->nbSurfels());
  num1.setNum(_m_shape->nbTriangles());
  num2.setNum(vol);
  num3.setNum(vol / (4.0 / 3.0 * M_PI));
  QMessageBox::about(this, "Shape information",
		     QString("vertices :") + num0 + "\n" +
		     QString("triangles :") + num1 + "\n" +
		     QString("volume: ") + num2 + "\n" +
		     QString("deviation: ") + num3);
}

#ifdef WIN32
#include "moc_qtGlWidget.cpp"
#endif
