/* ************************************************************************* **
** Alexis Angelidis							     **
** Thu Mar 20 16:31:04 NZST 2003					     **
** ************************************************************************* */
#ifndef __QT3D_GL_HH__
#define __QT3D_GL_HH__

/* ************************************************************************* */
#include <vector>
#include <list>
#include <string>
#include <qgl.h>
#include <qimage.h>
#include <qevent.h> 
#include <qthread.h> 
#include "glTrackball.h"
#include "glGizmo.h"

// ************************************************************************* //
class	Tool;
class	Sweep;
class	Shape;

// ************************************************************************* //
/// DESCRIPTION: Saves a buffer of animation.
// ************************************************************************* //
class SaveThread: public QThread
{
  QImage	*_m_imageData;
  unsigned	*_m_imageIds;
  const unsigned	_m_size;
  QString	getNum(const QString name, unsigned i, const unsigned base);
protected:
  void		run();
public:
  SaveThread(const QImage *imageData, const unsigned *imageIds,
	     const unsigned size = 64);
  virtual ~SaveThread();
  void	clean();
};

// ************************************************************************* //
class MyGLWidget: public QGLWidget
{
  Q_OBJECT
protected:
  TrackBall		_m_trackball;
  Gizmo			_m_gizmo;
  Matrix4d		_m_mGizmo;
  View			_m_view;
  bool			_m_isInViewMode;
  mutable QMouseEvent	_m_e0;
  Shape			*_m_shapeUndo;
  Shape			*_m_shape;
  std::list<Tool *>	_m_tools;
  Sweep			*_m_curSweep;
  bool			_m_isGouraudWire; // Shading mode
  bool			_m_isShowGrid; // Show grid enabled/disabled
  bool			_m_isUndoDisabled;// Undo enabled/disabled
  bool			_m_isEnvMap;// Tool color mode
  bool			_m_isToolColor;// Tool color mode
  int			_m_sym;// Tool symetry enabled/disabled
  int			_m_fps;// Frames per seconds
  void			keyPressEvent(QKeyEvent *e);
  int			_m_timerId;// Timer id, when recording.
  unsigned		_m_iLocal, _m_iGlobal;
  const unsigned	_m_size;
  QImage		*_m_imageData;
  unsigned		*_m_imageIds;
  std::list<SaveThread *>	_m_saveThreads;
  bool			_m_isRecording;
  static GLuint		_s_envMap;
public:
  MyGLWidget(QWidget *parent = 0, const char *name = 0);
  virtual ~MyGLWidget();
  void		initializeGL();
  void		resizeGL(int w, int h);
  void		paintGL();
  void		glSelect() const;
  void		mouseMoveEvent(QMouseEvent *e);
  void		mousePressEvent(QMouseEvent *e);
  void		timerEvent(QTimerEvent *e);
  static void	initMaterials();

  // is declared in Qt3
  QImage grabFrameBuffer ( bool withAlpha = FALSE );

public slots:
  void	setGouraud();
  void	setWireframe();
  void	setSolidwires();
  void	setGouraudWire();
  void	resetView();
  bool	saveView();
  bool	loadView();
  void	resetTool();
  void	centerTool();
  void	resetShape();
  void	reset();
  bool	saveShape();
  bool	loadShape();
  bool	renderShape();
  bool	exportShape();
  void	toggleEasyTool(bool b);
  void	toggleEmptyTool(bool b);
  void	toggleShowBox(bool b);
  void	toggleDisableUndo(bool b);
  void	toggleEnvMap(bool b);
  void	toggleZones(bool b);
  void	setShapeColor();
  void	setActiveToolCoating(int k);
  void	setSymmetry(int s);
  void	cleanShape();
  void	renormalize();
  void	undo();
  void	selectTool(int id);
  void	selectMultipleTool(int id);
  void	select(int id);
  void	toggleRecord(bool b);
  void	resetRecordIndex();
  void	setRecordIndex();
  void	shapeInformation();
};

/* ************************************************************************* */
inline void	MyGLWidget::initializeGL()
{
  const GLfloat	specular[] = {1.0, 1.0, 1.0, 1.0};
  const GLfloat	shininess[] = {50.0};

  glClearColor(0.647, 0.647, 0.647, 0.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_NORMALIZE);

  glShadeModel(GL_SMOOTH);
  glEnable(GL_COLOR_MATERIAL);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);

  const GLfloat	lDiffuse0[] = {1.0f, 1.0f, 1.0f, 1.0f};
  const GLfloat	lDiffuse1[] = {0.3f, 0.7f, 0.3f, 1.0f};
  const GLfloat	lDiffuse2[] = {0.7f, 0.3f, 0.3f, 1.0f};
  const GLfloat	lDiffuse3[] = {1.0f, 1.0f, 0.3f, 1.0f};

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT3);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lDiffuse0);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, lDiffuse1);
  glLightfv(GL_LIGHT2, GL_DIFFUSE, lDiffuse2);
  glLightfv(GL_LIGHT2, GL_DIFFUSE, lDiffuse3);
  glPointSize(4.0);
}

/* ************************************************************************* */
inline void	MyGLWidget::resizeGL(int	w,
				     int	h)
{
  glViewport(0, 0, w, h);
  _m_view.get();
  _m_trackball.resize();
}

/* ************************************************************************* */
#endif//__QT3D_GL_HH__
