/* ************************************************************************* */
#include <qapplication.h>
#include <qgl.h>
#include <qmainwindow.h>
#include <fstream>

/* ************************************************************************* */
#include "qtMainWidget.h"
#include "ffdsphere.xpm"
#include "ffdpinch.xpm"
#include "ffdstrangle.xpm"
//#include "ffdcustom.xpm"
#include "ffdvolume.xpm"
#include "ffdproto.xpm"
#include "ffdzoom.xpm"
#include "ffdstokes.xpm"

// ************************************************************************* //
MainWindow::MainWindow(QWidget		*parent,
		       const char	*name): QMainWindow(parent, name)
{
  MyGLWidget	*glwidget = new MyGLWidget(this);
  QPopupMenu	*menu = new QPopupMenu(this);

  menu->insertItem("&Reset", glwidget, SLOT(reset()), 0, 1);
  menu->insertSeparator();
  menu->insertItem("&Exit", qApp, SLOT(quit()), CTRL + Key_Q);
  menuBar()->insertItem("&Menu", menu, 0, 0);
  setCentralWidget(glwidget);
  glwidget->setFocus();

  _m_shape = new QPopupMenu(this);
  _m_shape->setCheckable(TRUE);
  _m_shape->insertItem("&Reset", glwidget, SLOT(resetShape()), 0, 0);
  _m_shape->insertSeparator();
  _m_shape->insertItem("&Save", glwidget, SLOT(saveShape()), 0, 1);
  _m_shape->setAccel(CTRL + Key_S, 1);
  _m_shape->insertItem("&Load", glwidget, SLOT(loadShape()), 0, 2);
  _m_shape->setAccel(CTRL + Key_L, 2);
  _m_shape->insertItem("&Export", glwidget, SLOT(exportShape()), 0, 3);

  _m_shape->insertItem("&Render", glwidget, SLOT(renderShape()), 0, 4);
  _m_shape->insertSeparator();

  _m_shape->insertItem("&Show grid", this, SLOT(slotShowBox()), 0, 5);
  connect(this, SIGNAL(toggleShowBox(bool)), 
	  glwidget, SLOT(toggleShowBox(bool)));
  _m_shape->setItemChecked(5, FALSE);
  _m_shape->insertItem("&Clean Shape", glwidget, SLOT(cleanShape()),
		       CTRL + Key_C);
  _m_shape->insertItem("&Renormalize", glwidget, SLOT(renormalize()));
  _m_shape->insertItem("&Color", glwidget, SLOT(setShapeColor()));
  _m_shape->insertItem("&Information", glwidget, SLOT(shapeInformation()));
  menuBar()->insertItem("&Shape", _m_shape, 1, 1);

  _m_tool = new QPopupMenu(this);
  _m_tool->setCheckable(TRUE);
  _m_tool->insertItem("&Reset", glwidget, SLOT(resetTool()), 0, 1);
  _m_tool->insertItem("&Center", glwidget, SLOT(centerTool()), 0, 2);
  _m_tool->insertSeparator();
  _m_tool->insertItem("&Easy", this, SLOT(slotEasyTool()), 0, 3);
  _m_tool->setItemChecked(3, TRUE);
  _m_tool->setAccel(CTRL + Key_E, 3);
  connect(this, SIGNAL(toggleEasyTool(bool)), 
	  glwidget, SLOT(toggleEasyTool(bool)));
  _m_tool->insertItem("&Empty", this, SLOT(slotEmptyTool()), 0, 4);
  _m_tool->setItemChecked(4, FALSE);
  connect(this, SIGNAL(toggleEmptyTool(bool)), 
	  glwidget, SLOT(toggleEmptyTool(bool)));
  _m_tool->insertItem("&Undo", glwidget, SLOT(undo()), 0, 5);
  _m_tool->setAccel(CTRL + Key_Z, 5);
  _m_tool->insertItem("&Disable undo", this, SLOT(slotDisableUndo()), 0, 6);
  _m_tool->setItemChecked(6, TRUE);
  connect(this, SIGNAL(toggleDisableUndo(bool)), 
	  glwidget, SLOT(toggleDisableUndo(bool)));
  menuBar()->insertItem("&Tool", _m_tool, 2, 2);

  _m_view = new QPopupMenu(this);
  _m_view->insertItem("&Reset", glwidget, SLOT(resetView()),
		      CTRL + Key_V);
  _m_view->insertItem("&Save", glwidget, SLOT(saveView()));
  _m_view->insertItem("&Load", glwidget, SLOT(loadView()));
  menuBar()->insertItem("&View", _m_view, 3, 3);

  _m_shading = new QPopupMenu(this);
  _m_shading->insertItem("&Smooth", glwidget, SLOT(setGouraud()), 0, 0);
  _m_shading->setAccel(CTRL + Key_G, 0);
  _m_shading->insertItem("&Wires", glwidget, SLOT(setWireframe()), 0, 1);
  _m_shading->insertItem("&Solid wires", glwidget, SLOT(setSolidwires()),
			 0, 2);
  _m_shading->insertItem("&Wires on smooth", glwidget, SLOT(setGouraudWire()),
			 0, 3);
  _m_shading->setAccel(CTRL + Key_W, 3);

  _m_shading->insertItem("&EnvMap", this, SLOT(slotEnvMap()), 0, 4);
  connect(this, SIGNAL(toggleEnvMap(bool)), glwidget,
	  SLOT(toggleEnvMap(bool)));

  _m_shading->insertItem("&Tool/Stretch zone", this, SLOT(slotZones()), 0, 5);
  _m_shading->setItemChecked(5, TRUE);
  connect(this, SIGNAL(toggleZones(bool)), glwidget, SLOT(toggleZones(bool)));


  menuBar()->insertItem("&Shading", _m_shading, 4, 4);

  _m_capture = new QPopupMenu(this);
  _m_capture->insertItem("&Start/Stop", this, SLOT(slotRecord()), 0, 0);
  _m_capture->insertItem("&Reset Index", glwidget, SLOT(resetRecordIndex()),
			 0, 1);
  _m_capture->insertItem("&Set Index", glwidget, SLOT(setRecordIndex()),
			 0, 2);
  connect(this, SIGNAL(toggleRecord(bool)),
	  glwidget, SLOT(toggleRecord(bool)));
  _m_capture->setItemChecked(0, FALSE);
  _m_capture->setAccel(CTRL + Key_K, 0);

  menuBar()->insertItem("&Capture", _m_capture, 5, 5);

  _m_range = new MySpinBox(20, 1000, 10, this, "WSpinBox");
  _m_range->setValue(100);
  connect(_m_range, SIGNAL(valueChanged(int)),
	  glwidget, SLOT(setActiveToolCoating(int)));
  menuBar()->insertItem(_m_range, 6, 6);

  _m_symmetry = new QComboBox(FALSE, this, "ComboBox");
  _m_symmetry->insertItem("none");
  _m_symmetry->insertItem("x");
  _m_symmetry->insertItem("y");
  _m_symmetry->insertItem("z");
  connect(_m_symmetry, SIGNAL(activated(int)), glwidget,
	  SLOT(setSymmetry(int)));
  menuBar()->insertItem(_m_symmetry, 7, 7);

#ifdef WIN32
  _m_toolbox = new QToolBar("toolbox", NULL);
#else
  _m_toolbox = new QToolBar(this, "toolbox");
#endif

  _m_toolbox->setLabel("Toolbox");

  QPixmap	sphereIcon(ffdsphere);

  new ToolButton(0, sphereIcon, "Sphere tool", glwidget,
		 _m_toolbox, "Sphere tool");

  QPixmap	pinchIcon(ffdpinch);

  new ToolButton(1, pinchIcon, "Pinch tool", glwidget,
		 _m_toolbox, "Pinch tool");

  QPixmap	strangleIcon(ffdstrangle);

  new ToolButton(2, strangleIcon,"Strangle tool", glwidget,
		 _m_toolbox, "Strangle tool");

  QPixmap	volumeIcon(ffdvolume);
  
  new ToolButton(3, volumeIcon, "Volume tool", glwidget,
		 _m_toolbox, "Volume tool");

  QPixmap	protoIcon(ffdproto);

  new ToolButton(4, protoIcon, "Proto tool", glwidget,
		 _m_toolbox, "Proto tool");

  QPixmap	zoomIcon(ffdzoom);

  new ToolButton(5, zoomIcon, "Zoom tool", glwidget,
		 _m_toolbox, "Zoom tool");

  QPixmap	stokesIcon(ffdstokes);

  new ToolButton(6, stokesIcon, "Stokes tool", glwidget,
		 _m_toolbox, "Stokes tool");
}

// ************************************************************************* //
MainWindow::~MainWindow()
{
}

// ************************************************************************* //
void MainWindow::slotEasyTool()
{
  const bool	val(!_m_tool->isItemChecked(3));

  _m_tool->setItemChecked(3, val);
  emit toggleEasyTool(val);
}

// ************************************************************************* //
void MainWindow::slotEmptyTool()
{
  const bool	val(!_m_tool->isItemChecked(4));

  _m_tool->setItemChecked(4, val);
  emit toggleEmptyTool(val);
}

// ************************************************************************* //
void MainWindow::slotShowBox()
{
  const bool	val(!_m_shape->isItemChecked(5));

  _m_shape->setItemChecked(5, val);
  emit toggleShowBox(val);
}

// ************************************************************************* //
void MainWindow::slotDisableUndo()
{
  const bool	val(!_m_tool->isItemChecked(6));

  _m_tool->setItemChecked(6, val);
  emit toggleDisableUndo(val);
}

// ************************************************************************* //
void MainWindow::slotEnvMap()
{
  const bool	val(!_m_shading->isItemChecked(4));

  _m_shading->setItemChecked(4, val);
  emit toggleEnvMap(val);
}

// ************************************************************************* //
void MainWindow::slotZones()
{
  const bool	val(!_m_shading->isItemChecked(5));

  _m_shading->setItemChecked(5, val);
  emit toggleZones(val);
}

// ************************************************************************* //
void MainWindow::slotRecord()
{
  const bool	val(!_m_capture->isItemChecked(0));

  _m_capture->setItemChecked(0, val);
  emit toggleRecord(val);
}

#ifdef WIN32
#include "moc_qtMainWidget.cpp"
#endif
// ************************************************************************* //
#ifdef WIN32
QStringData* QString::shared_null;
QApplication* qApp;
#endif
int	main(int ac, char **av)
{
  QApplication	qApp(ac, av);
  MainWindow	mainWindow;
  
  mainWindow.setGeometry(300, 200, 640, 550);
  qApp.setMainWidget(&mainWindow);
  mainWindow.show();
#ifdef WIN32
  mainWindow.showToolbar();
#endif
  qApp.exec();
  return 0;
}


