// ************************************************************************* //
/// AUTHOR: Alexis Angelidis 
/// STARTED: Fri Nov 21 16:40:28 NZDT 2003
// ************************************************************************* //
#ifndef __QTMAINWODGET_HH__
#define __QTMAINWODGET_HH__

// ************************************************************************* //
#include <qmainwindow.h> 
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qspinbox.h>
#include <qcombobox.h>
#include <qgl.h>
#include "qtGlWidget.h"

// ************************************************************************* //
class ToolButton: public QToolButton
{
  Q_OBJECT
protected:
  const unsigned	_m_idTool;
  void			mousePressEvent(QMouseEvent *e);
public:
  ToolButton(const unsigned idTool, const QIconSet &iconSet,
	     const QString &textLabel, MyGLWidget *receiver,
	     QToolBar *parent, const char *name = 0);
signals:
  void	selectMultipleTool(int idTool);
  void	selectSingleTool(int idTool);
};

/* ************************************************************************* */
inline void	ToolButton::mousePressEvent(QMouseEvent *e)
{
  if (e->state() == ControlButton)
    emit selectMultipleTool(_m_idTool);
  else
    emit selectSingleTool(_m_idTool);
  QToolButton::mousePressEvent(e);
}

/* ************************************************************************* */
inline ToolButton::ToolButton(const unsigned	idTool,
			      const QIconSet	&iconSet,
			      const QString	&textLabel, 
			      MyGLWidget	*receiver,
			      QToolBar		*parent,
			      const char	*name):
  QToolButton(parent, name), 
  _m_idTool(idTool)
{
  setIconSet(iconSet);
  setTextLabel(textLabel, true);
  
  connect(this, SIGNAL(selectSingleTool(int)), receiver,
	  SLOT(selectTool(int)));
  connect(this, SIGNAL(selectMultipleTool(int)), receiver,
	  SLOT(selectMultipleTool(int)));
}

// ************************************************************************* //
/// DESCRIPTION: Spinbox class for selecting a tool thickness.
// ************************************************************************* //
class MySpinBox: public QSpinBox
{
  Q_OBJECT
public:
  MySpinBox(QWidget *parent = 0, const char * name = 0):
    QSpinBox(parent, name) {}
  MySpinBox(int minValue, int maxValue, int step = 1,
	    QWidget *parent = 0, const char * name = 0):
    QSpinBox(minValue, maxValue, step, parent, name) {}
  QString mapValueToText(int value)
  {
    if (value == -1)
      return QString("Auto");
    if (value % 100 < 10)
      return QString("%1.0%2").arg(value / 100).arg(value % 100);
    return QString("%1.%2").arg(value / 100).arg(value % 100);
  }
  int	mapTextToValue(bool *)//ok
  {
    if (text() == "Auto")
      return -1;
    return int(100 * text().toFloat());
  }
};

// ************************************************************************* //
class MainWindow: public QMainWindow
{
  Q_OBJECT
  QPopupMenu	*_m_shape;
  QPopupMenu	*_m_tool;
  QPopupMenu	*_m_shading;
  QPopupMenu	*_m_view;
  QPopupMenu	*_m_capture;
  MySpinBox	*_m_range;
  QComboBox	*_m_symmetry;
  QToolBar	*_m_toolbox;
public:
  MainWindow(QWidget *parent = 0, const char *name = 0);
  ~MainWindow();
#ifdef WIN32
  void showToolbar() { if (_m_toolbox) _m_toolbox->show(); }
#endif
public slots:
  void	slotEasyTool();
  void	slotEmptyTool();
  void	slotShowBox();
  void	slotDisableUndo();
  void	slotEnvMap();
  void	slotZones();
  void	slotRecord();
signals:
  void	toggleEasyTool(bool);
  void	toggleEmptyTool(bool);
  void	toggleShowBox(bool);
  void	toggleDisableUndo(bool);
  void	toggleEnvMap(bool);
  void	toggleZones(bool);
  void	toggleRecord(bool);
};

// ************************************************************************* //
#endif//__QTMAINWODGET_HH__
