// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#ifndef	__SFFILTER_HH__
#define	__SFFILTER_HH__

// ///////////////////////////////////////////////////////////////////////// //
#include "gfxTypes.h"

// ///////////////////////////////////////////////////////////////////////// //
template <class T>
class Grid;
class SpeedGrid;
class DivGrid;
class MultiMatrix;

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Cross filter, defined in a [0, 1, 2]^3 cube.
// ///////////////////////////////////////////////////////////////////////// //
class Filter
{
  friend class	MultiMatrix;
protected:
  double	_m_c, _m_s[6];
public:
  Filter() {}
  Filter(const double c, const double s);
  Filter(const double c, const double smx, const double spx, const double smy,
	 const double spy, const double smz, const double spz);
  Vector3d	convolve(const SpeedGrid &w, const int x, const int y, 
			 const int z) const;
  double	convolve(const DivGrid &w, const int x, const int y, 
			 const int z) const;
  double	operator()(const unsigned x, const unsigned y, 
			   const unsigned z) const;
  double	&operator()(const unsigned x, const unsigned y,
			    const unsigned z);
  void	reset() { _m_c = 0; for (int i(0); i < 6; ++i) _m_s[i] = 0.0; }
  friend std::ostream	&operator<<(std::ostream &os, const Filter &f);
};

// ///////////////////////////////////////////////////////////////////////// //
inline Filter::Filter(const double c,
		      const double s): _m_c(c)
{
  for (unsigned i(0); i < 6; ++i)
    _m_s[i] = s;
}

// ///////////////////////////////////////////////////////////////////////// //
inline Filter::Filter(const double c,
		      const double smx,
		      const double spx,
		      const double smy,
		      const double spy,
		      const double smz,
		      const double spz): _m_c(c)
{
  _m_s[0] = smx;
  _m_s[1] = spx;
  _m_s[2] = smy;
  _m_s[3] = spy;
  _m_s[4] = smz;
  _m_s[5] = spz;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	Filter::operator()(const unsigned x,
				   const unsigned y,
				   const unsigned z) const
{
  if (x == 1 && y == 1 && z == 1)
    return _m_c;
  if (y == 1 && z == 1)
    return (x == 0)? _m_s[0]: _m_s[1];
  if (z == 1 && x == 1)
    return (y == 0)? _m_s[2]: _m_s[3];
  if (x == 1 && y == 1)
    return (z == 0)? _m_s[4]: _m_s[5];
  return 0.0;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	&Filter::operator()(const unsigned x,
				    const unsigned y,
				    const unsigned z)
{
  static double	zero;

  if (x == 1 && y == 1 && z == 1)
    return _m_c;
  if (y == 1 && z == 1)
    return (x == 0)? _m_s[0]: _m_s[1];
  if (z == 1 && x == 1)
    return (y == 0)? _m_s[2]: _m_s[3];
  if (x == 1 && y == 1)
    return (z == 0)? _m_s[4]: _m_s[5];
  zero = 0.0;
  return zero;
}

// ///////////////////////////////////////////////////////////////////////// //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Filter	&f)
{
  return os << f._m_c << " " << f._m_s[0] << " " << f._m_s[1] << " "
	    << f._m_s[2] << " " << f._m_s[3] << " " << f._m_s[4] << " "
	    << f._m_s[5];
}

// ///////////////////////////////////////////////////////////////////////// //
#endif//__SFFILTER_HH__
