// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#ifndef	__SFISOFILTER_HH__
#define	__SFISOFILTER_HH__

// ///////////////////////////////////////////////////////////////////////// //
class DivGrid;
class SpeedGrid;
class UniMatrix;

// ///////////////////////////////////////////////////////////////////////// //
#include "gfxTypes.h"

// ///////////////////////////////////////////////////////////////////////// //
class IsoFilter
{
  friend class	DivGrid;
  friend class	SpeedGrid;
  friend class	UniMatrix;
  double	_m_c, _m_s;
public:
  IsoFilter() {}
  IsoFilter(const double c, const double s);
  double	operator()(const unsigned x, const unsigned y, 
			   const unsigned z) const;
  double	&operator()(const unsigned x, const unsigned y,
			    const unsigned z);
  void	reset() { _m_c = _m_s = 0.0; }
  friend std::ostream	&operator<<(std::ostream &os, const IsoFilter &f);
};

// ///////////////////////////////////////////////////////////////////////// //
inline IsoFilter::IsoFilter(const double c,
			    const double s): _m_c(c), _m_s(s)
{
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	IsoFilter::operator()(const unsigned x,
				      const unsigned y,
				      const unsigned z) const
{
  if (x == 1 && y == 1 && z == 1)
    return _m_c;
  if (y == 1 && z == 1)
    return _m_s;
  if (z == 1 && x == 1)
    return _m_s;
  if (x == 1 && y == 1)
    return _m_s;
  return 0.0;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	&IsoFilter::operator()(const unsigned x,
				       const unsigned y,
				       const unsigned z)
{
  static double	zero;

  if (x == 1 && y == 1 && z == 1)
    return _m_c;
  if (y == 1 && z == 1)
    return _m_s;
  if (z == 1 && x == 1)
    return _m_s;
  if (x == 1 && y == 1)
    return _m_s;
  zero = 0.0;
  return zero;
}

// ///////////////////////////////////////////////////////////////////////// //
inline std::ostream	&operator<<(std::ostream	&os,
				    const IsoFilter	&f)
{
  return os << f._m_c << " " << f._m_s;
}

// ///////////////////////////////////////////////////////////////////////// //
#endif//__SFISOFILTER_HH__
