// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#ifndef	__SFSOLVER_HH__
#define	__SFSOLVER_HH__

// ///////////////////////////////////////////////////////////////////////// //
#include "gfxTypes.h"
#include "sfSparseMatrix.h"
#include "sfSpeedGrid.h"

// ///////////////////////////////////////////////////////////////////////// //
class IsoFilter;
class Object;
class Solver;

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Tool ball.
// ///////////////////////////////////////////////////////////////////////// //
class Object
{
  friend class	Solver;
  Point3d	_m_p;//position
  double	_m_rad;//radius
public:
  Object(const Point3d &p, const double rad): _m_p(p), _m_rad(rad) {}
  bool		contains(const Point3d &p) const;
  bool		intersects(const Point3d &p) const;
  bool		touches(const Point3d &p) const;
  void		glRender() const;
  Vector3d	normal(const Point3d &p) const;
};

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION:
// ///////////////////////////////////////////////////////////////////////// //
class Solver
{
private:
  const int	_m_x, _m_y, _m_z, _m_size;
  SpeedGrid	*_m_speed, *_m_speedBuf;
  SpeedGrid	_m_speedBuf0, _m_speedBuf1, _m_speedBuf2;
  DivGrid	*_m_div;//divergence
  DivGrid	*_m_pBuf0, *_m_pBuf1, *_m_pBuf2, *_m_pBuf3;
  BBox3d	_m_bbox;//Volume solved
  IsoFilter	_m_a;//Projection filter
  MultiMatrix	*_m_l;//lower diagonal cholesky decomposition of a
  UniMatrix	_m_d;//Diffuse matrix
  Object	_m_object;
  double	_m_nu, _m_rho;
public:
  Solver(const int x, const int y, const int z, const double rad,
	 const double nu, const double rho);
  virtual ~Solver();
  void		maskObject();
  void		reset();
  void		glRender() const;
  void		solve(const double dt = 1.0);
  void		convect(const double dt, SpeedGrid *&speed, const int steps);
  void		diffuse(const double dt, SpeedGrid &speed);
  void		blur(SpeedGrid *&speed);
  void		project(SpeedGrid &speed);
  void		operator()(Point3d &p, const double dt,
			   const int steps = 10) const;
  Vector3d	&operator()(const int x, const int y, const int z);
  Vector3d	&operator()(const int i);
  Vector3d	speed(const Point3d &p) const;
};

// ///////////////////////////////////////////////////////////////////////// //
#endif//__SFSOLVER_HH__
