// ************************************************************************* //
// AUTHOR: Alexis Angelidis.
// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
// ************************************************************************* //
#include "gfxTypes.h"
#include "swpSurfel.h"
#include "swpEdge.h"
#include "swpTriangle.h"

// ************************************************************************* //
const double g_prec = 0.15;

// ************************************************************************* //
bool	Edge::geq(const Edge *const e0,
		  const Edge *const e1)
{
  return ((e0->_m_s0->_m_p - e0->_m_s1->_m_p).norm2() >=
	  (e1->_m_s0->_m_p - e1->_m_s1->_m_p).norm2());
}

// ************************************************************************* //
bool	Edge::leq(const Edge *const e0,
		  const Edge *const e1)
{
  return ((e0->_m_s0->_m_p - e0->_m_s1->_m_p).norm2() <=
	  (e1->_m_s0->_m_p - e1->_m_s1->_m_p).norm2());
}

// ************************************************************************* //
Edge::Edge(Surfel &s0,
	   Surfel &s1):
  _m_s0(&s0), _m_s1(&s1), _m_t0(0), _m_t1(0), _m_qList(0)
{
}

// ************************************************************************* //
void	Edge::connectTo(Triangle &t)
{
  if (!_m_t0 && !_m_t1)
    _m_t0 = &t;
  else if (!_m_t1)
    if (((_m_s0 == &t.gets1()) && (_m_s1 == &t.gets0())) ||
	((_m_s0 == &t.gets2()) && (_m_s1 == &t.gets1())) ||
	((_m_s0 == &t.gets0()) && (_m_s1 == &t.gets2())))
      {
	_m_t1 = _m_t0;
	_m_t0 = &t;
      }
    else
      _m_t1 = &t;
}

// ************************************************************************* //
void	Edge::connectTo(Triangle	&t0,
			Triangle	&t1)
{
  if (((_m_s0 == &t0.gets1()) && (_m_s1 == &t0.gets0())) ||
      ((_m_s0 == &t0.gets2()) && (_m_s1 == &t0.gets1())) ||
      ((_m_s0 == &t0.gets0()) && (_m_s1 == &t0.gets2())))
    _m_t0 = &t0, _m_t1 = &t1;
  else
    _m_t0 = &t1, _m_t1 = &t0; 
}

// ************************************************************************* //
/// DESCRIPTION: Returns false if the edge hasn't been stetched or 
///  compressed, else tests the edge properly.
// ************************************************************************* //
bool	Edge::isTooShort() const
{
  //if (_m_s0->_m_detJac == 1.0 && _m_s1->_m_detJac == 1.0)
  //return false;

  const Vector3d	v0(_m_s0->_m_v), v1(_m_s1->_m_v);
  const Point3d		p0(_m_s0->_m_p), p1(_m_s1->_m_p);
  const Vector3d	g(p0 - p1), p(0.25 * ((g * v0) * v0 - (g * v1) * v1));
  const double		d2((p0 - p1).norm2());
  const double		v01(v0 * v1);

  return ((0.5 < v01 && 1.0 - v01 < 0.5 * sq(0.05)) && 
	  (d2 < sq(g_prec * 0.5)) &&
	  (p.norm2() < sq(g_prec * 0.5 / 40.0))) || 
    (d2 < sq(g_prec / 128.0)); //CTE
}

// ************************************************************************* //
double	sphereField(double d2)
{
  //const double	d2 = (p - c).norm2();
  const double rmax = 0.75;
  const double rmin = 0.02;

  if (d2 >= sq(rmax))
    return 1.0;
  if (d2 <= sq(rmin))
    return 0.01;

  const double dk = (sqrt(d2) - rmin) / (rmax - rmin);
  return 0.01 + 
    (1.0 - 0.01) * (-dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0));
  //return dk * (2 - dk);
}

// ************************************************************************* //
bool	Edge::isTooShort(const std::list<Surfel *>	&sharps) const
{
  return false;
  std::list<Surfel *>::const_iterator	vit, vend(sharps.end());

  for (vit = sharps.begin(); vit != vend; ++vit)
    if (_m_s0 == *vit || _m_s1 == *vit)
      return false;

  const Vector3d	v0(_m_s0->_m_v), v1(_m_s1->_m_v);
  const Point3d		p0(_m_s0->_m_p), p1(_m_s1->_m_p);
  const Vector3d	g(p0 - p1), p(0.25 * ((g * v0) * v0 - (g * v1) * v1));
  const double		d2((p0 - p1).norm2());
  const double		v01(v0 * v1);
  const double		k = pow(min(_m_s0->_m_detJac, _m_s1->_m_detJac), 1024);

  return ((0.5 < v01 && 1.0 - v01 < 0.5 * sq(0.05 * k)) && 
	  (d2 < sq(g_prec * 0.5 * k)) &&
	  (p.norm2() < sq(g_prec * 0.5 / 40.0 * k))) || 
    (d2 < sq(g_prec / 128.0)); //CTE
}

// ************************************************************************* //
/// DESCRIPTION: Returns false if the edge has been compressed,
///  else tests the edge properly.
// ************************************************************************* //
bool	Edge::isTooLong() const
{
  //if (_m_s0->_m_detJac == 1.0 && _m_s1->_m_detJac == 1.0)
  //return false;

  const Vector3d	v0(_m_s0->_m_v), v1(_m_s1->_m_v);
  const Point3d		p0(_m_s0->_m_p), p1(_m_s1->_m_p);
  const Vector3d	g(p0 - p1), p(0.25 * ((g * v0) * v0 - (g * v1) * v1));
  const double		d2((p0 - p1).norm2());
  const double		v01(v0 * v1);

  return (sq(g_prec / 20.0) < p.norm2()) || 
    (0.5 < v01 && 1.0 - v01 > 0.5 * sq(g_prec)) || 
    sq(g_prec) < d2; //CTE
}

// ************************************************************************* //
bool	Edge::isTooLong(const std::list<Surfel *>	&sharps) const
{
  return false;
  std::list<Surfel *>::const_iterator	vit, vjt, vend(sharps.end());

  for (vit = sharps.begin(); vit != vend; ++vit)
    if (_m_s0 == *vit)
      for (vjt = sharps.begin(); vjt != vend; ++vjt)
	if (_m_s1 == *vjt)
	  return false;

  //double kmin = 1.0e12, ktmp;
  //for (vit = sharps.begin(); vit != vend; ++vit)
  //if ((ktmp = (*vit)->_m_detJac) < kmin)
  //kmin = ktmp;

  const Vector3d	v0(_m_s0->_m_v), v1(_m_s1->_m_v);
  const Point3d		p0(_m_s0->_m_p), p1(_m_s1->_m_p);
  const Vector3d	g(p0 - p1), p(0.25 * ((g * v0) * v0 - (g * v1) * v1));
  const double		d2((p0 - p1).norm2());
  const double		v01(v0 * v1);
  const double		k = pow(min(_m_s0->_m_detJac, _m_s1->_m_detJac), 1024);

  return (sq(g_prec / 20.0 / k) < p.norm2()) || 
    (0.5 < v01 && 1.0 - v01 > 0.5 * sq(g_prec / k)) || 
    sq(g_prec / k) < d2; //CTE
  //return isTooLong();
}

// ************************************************************************* //
/// DESCRIPTION: The edge HAS to be in a querry list.
// ************************************************************************* //
void	Edge::remFromQueryList()
{
  if (_m_qList)
    _m_qList->erase(_m_qHandle), _m_qList = 0;
}

// ************************************************************************* //
void	Edge::pushFrontInQueryList(std::list<Edge *> &qlist)
{
  _m_qList = &qlist;
  _m_qList->push_front(this);
  _m_qHandle = _m_qList->begin();
}

// ************************************************************************* //
void	Edge::pushBackInQueryList(std::list<Edge *> &qlist)
{
  _m_qList = &qlist;
  _m_qList->push_back(this);
  _m_qHandle = --_m_qList->end();
}

/* ************************************************************************* **
** DESCRIPTION: Flips the edges.           				     **
** NOTE:    v2	            v2	        				     **
**         /^\ 	           / \						     **
**        / | \           /   \						     **
**      e2  |  e1	e2     e1					     **
**      /   |   \       /   t1  \					     **
**     / t1 | t0 \     /         \					     **
**   v3    this   v1 v3----this-->v1					     **
**     \    |    /     \         /	 				     **
**      \   |   /       \   t0  /					     **
**      e3  |  e0       e3     e0					     **
**        \ | / 	  \   / 					     **
**         \|/ 	           \ /						     **
**          v0	            v0						     **
** ************************************************************************* */
void	Edge::flip()
{
  Edge	&e0(_m_t0->gete(*this));
  Edge	&e2(_m_t1->gete(*this));
  Edge	&e1(_m_t0->gete(e0)), &e3(_m_t1->gete(e2));
  Surfel	&s1(e0.gets(*_m_s0)), &s3(e2.gets(*_m_s1));
  
  _m_t0->replace(e0, e3);
  _m_t0->replace(e1, e0);
  _m_t1->replace(e2, e1);
  _m_t1->replace(e3, e2);
  e3.replace(*_m_t1, *_m_t0);
  e1.replace(*_m_t0, *_m_t1);
  _m_s0 = &s3;
  _m_s1 = &s1;
}

// ************************************************************************* //
Triangle	&Edge::gett(const Triangle &t)
{
  return (_m_t0 == &t)? *_m_t1: *_m_t0;
}

// ************************************************************************* //
Triangle	&Edge::gett(const int i)
{
  return i? *_m_t1: *_m_t0;
}

/* ************************************************************************* **
** DESCRIPTION: If the smallest angle of the two adjacent triangles	     **
**   is bigger when edge is flipped, then edge is flipped and returns	     **
**   true, else false.							     **
** NOTE: Flip isn't done if surface is put in critical state, such as normal **
**   inversion.								     **
** ************************************************************************* */
bool	Edge::tryFlip()
{
  double	a0, a1, a2, a3, a4, a5, b0, b1, b2, b3, b4, b5;
  Vector3d	n0, n1;

  b0 = _m_t0->computeCosAngle0();
  b1 = _m_t0->computeCosAngle1();
  b2 = _m_t0->computeCosAngle2();
  b3 = _m_t1->computeCosAngle0();
  b4 = _m_t1->computeCosAngle1();
  b5 = _m_t1->computeCosAngle2();
  flip();
  _m_t0->computeNormaln(n0);
  _m_t1->computeNormaln(n1);
  if (0.0 < n0 * n1)
    {
      a0 = _m_t0->computeCosAngle0();
      a1 = _m_t0->computeCosAngle1();
      a2 = _m_t0->computeCosAngle2();
      a3 = _m_t1->computeCosAngle0();
      a4 = _m_t1->computeCosAngle1();
      a5 = _m_t1->computeCosAngle2();
      if (min(min(min(min(min(b0, b1), b2), b3), b4), b5) <
	  min(min(min(min(min(a0, a1), a2), a3), a4), a5))
	return true;
    }
  flip();
  return false;
}

// ************************************************************************* //
const Surfel	&Edge::gets(const Surfel &s) const
{
  return (_m_s0 == &s)? *_m_s1: *_m_s0;
}

// ************************************************************************* //
Surfel	&Edge::gets(Surfel &s)
{
  return (_m_s0 == &s)? *_m_s1: *_m_s0;
}

// ************************************************************************* //
void	Edge::replace(const Surfel	&s0,
		      Surfel		&s1)
{
  if (_m_s0 == &s0)
    _m_s0 = &s1;
  else
    _m_s1 = &s1;
}

// ************************************************************************* //
void	Edge::replace(const Triangle	&t0,
		      Triangle		&t1)
{
  if (_m_t0 == &t0)
    _m_t0 = &t1;
  else
    _m_t1 = &t1;
}

// ************************************************************************* //
/// DESCRIPTION: Replaces all references to u with v. Use this method for
///  replacing surfels in the shape.
// ************************************************************************* //
void	Edge::replaceAll(const Surfel	&u,
			 Surfel		&v)
{
  std::list<Edge *>	elist;
  Edge::circulator	ecir, ebeg;
  std::list<Edge *>::iterator	eit, eend;

  ecir = ebeg = Edge::circulator(u, *this);
  for (++ecir; ecir != ebeg; ++ecir)
    elist.push_back(&(*ecir));
  for (eit = elist.begin(), eend = elist.end(); eit != eend; ++eit)
    (*eit)->replace(u, v);
  replace(u, v);
}

// ************************************************************************* //
void	Edge::split(Surfel &s) const
{
  //
  /*
  Vector3d	u0, u1;

  const Vector3d	&v0(_m_s0->_m_oldv), &v1(_m_s1->_m_oldv);
  _m_t0->computeNormalp(u0);
  _m_t1->computeNormalp(u1);
  Vector3d	v = (u0 + u1).unit();
  const double	w0 = v * v0, w1 = v * v1;

  if (0.0 < w0 && 0.0 < w1)
    s._m_v = (w0 * v0 + w1 * v1).unit(),
      s._m_p = (_m_s0->_m_oldp * w1 + _m_s1->_m_oldp * w0) * (1.0 / (w0 + w1));
  else
    s._m_v = (v0 + v1).unit(),
      s._m_p = (_m_s0->_m_oldp + _m_s1->_m_oldp) * 0.5;
  s.saveSurfel();
  return;
  */

  s._m_v = (_m_s0->_m_oldv + _m_s1->_m_oldv);
  s._m_v.eqUnit();
  s._m_p = (_m_s0->_m_oldp + _m_s1->_m_oldp) * 0.5;
  s._m_detJac = (_m_s0->_m_detJac + _m_s1->_m_detJac) * 0.5;//DEBUG
  s.saveSurfel();
}


// ************************************************************************* //
void	Edge::splitDiag(Surfel &s) const
{
  s._m_v = (_m_s0->_m_oldv + _m_s1->_m_oldv);
  s._m_v.eqUnit();
  s._m_p = (_m_t0->gets(*this)._m_oldp + _m_t1->gets(*this)._m_oldp) * 0.5;
  s.saveSurfel();
}

// ************************************************************************* //
void	Edge::split(Surfel	&s,
		    const std::list<Surfel *> &sharps) const
{
  std::list<Surfel *>::const_iterator	sit, send;

  for (sit = sharps.begin(), send = sharps.end(); sit != send; ++sit)
    if (_m_s1 == *sit)
      {
	s._m_p = (_m_s0->_m_oldp + _m_s1->_m_oldp) * 0.5;
	s._m_v = _m_s0->_m_oldv;
	s.saveSurfel();
	return;
      }
    else if (_m_s0 == *sit)
      {
	s._m_p = (_m_s0->_m_oldp + _m_s1->_m_oldp) * 0.5;
	s._m_v = _m_s1->_m_oldv;
	s.saveSurfel();
	return;
      }
  split(s);
}

// ************************************************************************* //
void	Edge::collapse(Surfel &s) const
{
  s._m_v = (_m_s0->_m_oldv + _m_s1->_m_oldv);
  s._m_v.eqUnit();
  s._m_p = (_m_s0->_m_oldp + _m_s1->_m_oldp) * 0.5;
  s._m_detJac = (_m_s0->_m_detJac + _m_s1->_m_detJac) * 0.5;//DEBUG
  s.saveSurfel();
}

// ************************************************************************* //
/// DESCRIPTION: Increments the circulator.
// ************************************************************************* //
Edge::circulator	&Edge::circulator::operator++()
{
  _m_t = &_m_e->gett(*_m_t);
  _m_e = &_m_t->gete(*_m_s, *_m_e);
  return *this;
}

// ************************************************************************* //
/// DESCRIPTION: Decrements the circulator.
// ************************************************************************* //
Edge::circulator	&Edge::circulator::operator--()
{
  _m_e = &_m_t->gete(*_m_s, *_m_e);
  _m_t = &_m_e->gett(*_m_t);
  return *this;
}

// ************************************************************************* //
/// DECREMENTS: Tests if the circulators are different.
/// WARNING: The circulators MUST be 'compatible'.
// ************************************************************************* //
bool	Edge::circulator::operator!=(const Edge::circulator &eit) const
{
  return _m_e != eit._m_e;
}

