// ************************************************************************* //
// AUTHOR: Alexis Angelidis.
// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
// ************************************************************************* //
#ifndef __SWPEDGE_HH__
#define __SWPEDGE_HH__

// ************************************************************************* //
#include <list>

// ************************************************************************* //
class	Surfel;
class	Triangle;
class	Shape;

// ************************************************************************* //
extern const int	g_size;
extern const double	g_prec;

// ************************************************************************* //
class Edge
{
  friend class	Triangle;
  friend class	Shape;
  friend class	SweepPinch;
  Surfel	*_m_s0, *_m_s1;
  Triangle	*_m_t0, *_m_t1;
private:
  std::list<Edge *>::iterator	_m_meshHandle;
  std::list<Edge *>::iterator	_m_qHandle;
  std::list<Edge *>		*_m_qList;
public:
  // *********************************************************************** //
  /// DESCRIPTION: Edge Circulator.
  // *********************************************************************** //
  class circulator
  {
    const Surfel	*_m_s;
    Edge		*_m_e;
    Triangle		*_m_t;
  public:
    circulator() {}
    circulator(const circulator &ec): 
      _m_s(ec._m_s), _m_e(ec._m_e), _m_t(ec._m_t) {}
    circulator(const Surfel &s, Edge &e): _m_s(&s), _m_e(&e), _m_t(e._m_t0) {}
    circulator	&operator++();
    circulator	&operator--();
    bool	operator!=(const circulator &eit) const;
    Edge	&operator*() { return *_m_e; }
    Edge	*operator->() { return _m_e; }
  };
public:
  Edge(Surfel &s0, Surfel &s1);
  void	connectTo(Triangle &t);
  void	connectTo(Triangle &t0, Triangle &t1);
  bool	isTooLong() const;
  bool	isTooLong(const std::list<Surfel *> &sharps) const;
  bool	isTooShort() const;
  bool	isTooShort(const std::list<Surfel *> &sharps) const;
  void	pushFrontInQueryList(std::list<Edge *> &qlist);
  void	pushBackInQueryList(std::list<Edge *> &qlist);
  void	remFromQueryList();
  void	flip();
  bool	tryFlip();
  const Surfel	&gets(const Surfel &v) const;
  Surfel	&gets(Surfel &v);
  Surfel	&gets0() { return *_m_s0; }
  Surfel	&gets1() { return *_m_s1; }
  Triangle	&gett(const Triangle &t);
  Triangle	&gett(const int i);
  void		replace(const Triangle &t0, Triangle &t1); 
  void		replace(const Surfel &s0, Surfel &s1);
  void		replaceAll(const Surfel &u, Surfel &v);
  void		split(Surfel &s) const;
  void		splitDiag(Surfel &s) const;
  void		split(Surfel &s, const std::list<Surfel *> &sharps) const;
  void		collapse(Surfel &s) const;
  static bool	leq(const Edge *const e0, const Edge *const e1);
  static bool	geq(const Edge *const e0, const Edge *const e1);
  double	length2() const { return (_m_s0->_m_p - _m_s1->_m_p).norm2(); }
};

// ************************************************************************* //
#endif//__SWPEDGE_HH__
