// ************************************************************************* //
/// AUTHOR: Alexis Angelidis.
/// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
// ************************************************************************* //
#include <list>
#include <vector>
#include <map>
#include "gfxTypes.h"
#include "swpShape.h"
#include "swpSurfel.h"
#include "swpEdge.h"
#include "swpTriangle.h"

#ifndef MAXFLOAT
#define MAXFLOAT 3.40282347e+38F
#endif//MAXFLOAT

// ************************************************************************* //
Shape::Shape()
{
}

// ************************************************************************* //
Shape::~Shape()
{
  clear();
}

// ************************************************************************* //
void	Shape::copy(const Shape &s)
{
  clear();

  std::map<Surfel *, Surfel *>		smap;
  std::map<Edge *, Edge *>		emap;
  std::list<Surfel *>::const_iterator	sit, send(s._m_surfels.end());
  std::list<Edge *>::const_iterator	eit, eend(s._m_edges.end());
  std::list<Triangle *>::const_iterator	tit, tend(s._m_triangles.end());

  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    insert(*(smap[*sit] = new Surfel(**sit)));
  for (eit = s._m_edges.begin(); eit != eend; ++eit)
    emap[*eit] = &insert(*smap[(*eit)->_m_s0], *smap[(*eit)->_m_s1]);
  for (tit = s._m_triangles.begin(); tit != tend; ++tit)
    connect(*emap[(*tit)->_m_e0], *emap[(*tit)->_m_e1], *emap[(*tit)->_m_e2]);
}

// ************************************************************************* //
void	Shape::reset()
{
  clear();
  initBall();
}

// ************************************************************************* //
void	Shape::initEicosahedra()
{
  const Point3d	pi[12] = { Point3d(0.0, 0.0, 1.0),
			   Point3d(0.723607, 0.525731, 0.447214),
			   Point3d(-0.276393, 0.8506513, 0.447214),
			   Point3d(-0.894427, 0.0, 0.447214),
			   Point3d(-0.276393, -0.850651, 0.447214),
			   Point3d(0.723607, -0.525731, 0.447214),
			   Point3d(0.0, 0.0, -1.0),
			   Point3d(0.894427, 0.0, -0.447214),
			   Point3d(0.276393, 0.850651, -0.447214),
			   Point3d(-0.723607, 0.525731, -0.447214),
			   Point3d(-0.723607, -0.525731, -0.447214),
			   Point3d(0.276393, -0.850651, -0.447214) };
  Surfel	*si[12];

  for (unsigned i(0); i < 12; ++i)
    insert(*(si[i] = new Surfel(pi[i], Vector3d(pi[i]))));

  Edge	&e0_1(insert(*si[0], *si[1]));
  Edge	&e0_2(insert(*si[0], *si[2]));
  Edge	&e0_3(insert(*si[0], *si[3]));
  Edge	&e0_4(insert(*si[0], *si[4]));
  Edge	&e0_5(insert(*si[0], *si[5]));
  Edge	&e1_2(insert(*si[1], *si[2]));
  Edge	&e1_5(insert(*si[1], *si[5]));
  Edge	&e1_7(insert(*si[1], *si[7]));
  Edge	&e1_8(insert(*si[1], *si[8]));
  Edge	&e2_3(insert(*si[2], *si[3]));
  Edge	&e2_8(insert(*si[2], *si[8]));
  Edge	&e2_9(insert(*si[2], *si[9]));
  Edge	&e3_4(insert(*si[3], *si[4]));
  Edge	&e3_9(insert(*si[3], *si[9]));
  Edge	&e3_10(insert(*si[3], *si[10]));
  Edge	&e4_5(insert(*si[4], *si[5]));
  Edge	&e4_10(insert(*si[4], *si[10]));
  Edge	&e4_11(insert(*si[4], *si[11]));
  Edge	&e5_7(insert(*si[5], *si[7]));
  Edge	&e5_11(insert(*si[5], *si[11]));
  Edge	&e6_7(insert(*si[6], *si[7]));
  Edge	&e6_8(insert(*si[6], *si[8]));
  Edge	&e6_9(insert(*si[6], *si[9]));
  Edge	&e6_10(insert(*si[6], *si[10]));
  Edge	&e6_11(insert(*si[6], *si[11]));
  Edge	&e7_8(insert(*si[7], *si[8]));
  Edge	&e7_11(insert(*si[7], *si[11]));
  Edge	&e8_9(insert(*si[8], *si[9]));
  Edge	&e9_10(insert(*si[9], *si[10]));
  Edge	&e10_11(insert(*si[10], *si[11]));

  connect(e0_1, e1_2, e0_2);
  connect(e0_2, e2_3, e0_3);
  connect(e0_3, e3_4, e0_4);
  connect(e0_4, e4_5, e0_5);
  connect(e0_5, e1_5, e0_1);
  connect(e6_8, e7_8, e6_7);
  connect(e6_9, e8_9, e6_8);
  connect(e6_10, e9_10, e6_9);
  connect(e6_11, e10_11, e6_10);
  connect(e6_7, e7_11, e6_11);
  connect(e1_7, e7_8, e1_8);
  connect(e2_8, e8_9, e2_9);
  connect(e3_9, e9_10, e3_10);
  connect(e4_10, e10_11, e4_11);
  connect(e5_11, e7_11, e5_7);
  connect(e5_7, e1_7, e1_5);
  connect(e1_8, e2_8, e1_2);
  connect(e2_9, e3_9, e2_3);
  connect(e3_10, e4_10, e3_4);
  connect(e4_11, e5_11, e4_5);
}

// ************************************************************************* //
void	Shape::initBall()
{
  initEicosahedra();
  std::list<Edge *>::iterator	eit, eend(_m_edges.end());
  std::list<Surfel *>::iterator	sit;
  std::list<Edge *>		elist;
  Surfel	*newSurfel;
  Edge		*newEdges[4];

  for (eit = _m_edges.begin(); eit != eend; ++eit)
    (*eit)->pushFrontInQueryList(elist);
  while (!elist.empty())
    {
      elist.front()->tryFlip();
      elist.front()->_m_s0->_m_detJac = 
	elist.front()->_m_s1->_m_detJac = MAXFLOAT;
      if (elist.front()->isTooLong())
	{
	  std::list<Edge *>::iterator	eit(elist.begin());
	  
	  newSurfel = 0;
	  newEdges[0] = newEdges[1] = newEdges[2] = newEdges[3] = 0; 
	  if (split(eit, newSurfel, newEdges, 0))
	    {
	      newSurfel->_m_p = Point3d(newSurfel->_m_v);
	      newSurfel->_m_p.eqJitter(1.0e-6);
	      newSurfel->saveSurfel();
	      for (unsigned i(0); i < 4; ++i)
		if (newEdges[i])
		  newEdges[i]->pushBackInQueryList(elist);
	    }
	}
      else
	elist.front()->remFromQueryList();
    }
  for (sit = _m_surfels.begin(); sit != _m_surfels.end(); ++sit)
    (*sit)->_m_detJac = 1.0;
}

// ************************************************************************* //
void	Shape::glRender() const
{
  std::list<Triangle *>::const_iterator	tit, tend(_m_triangles.end());

  glBegin(GL_TRIANGLES);
  for (tit = _m_triangles.begin(); tit != tend; ++tit)
    (*tit)->glRender();
  glEnd();//GL_TRIANGLES
}

// ************************************************************************* //
void	Shape::glRender2() const
{
  std::list<Triangle *>::const_iterator	tit, tend(_m_triangles.end());

  glBegin(GL_TRIANGLES);
  for (tit = _m_triangles.begin(); tit != tend; ++tit)
    (*tit)->glRender2();
  glEnd();//GL_TRIANGLES
}

// ************************************************************************* //
void	Shape::glRender3() const
{
  std::list<Triangle *>::const_iterator	tit, tend(_m_triangles.end());

  glBegin(GL_TRIANGLES);
  for (tit = _m_triangles.begin(); tit != tend; ++tit)
    (*tit)->glRender3();
  glEnd();//GL_TRIANGLES
}

// ************************************************************************* //
void	Shape::query(std::list<Surfel *>	&qlist,
		     const BBox3d		&b)
{
  std::list<Surfel *>::iterator	sit, send(_m_surfels.end());

  for (sit = _m_surfels.begin(); sit != send; ++sit)
    if (b.contains((*sit)->_m_p))
      (*sit)->pushFrontInQueryList(qlist);
}

// ************************************************************************* //
void	Shape::insert(std::list<Surfel *> &qlist)
{
  std::list<Surfel *>::iterator	qit, qend(qlist.end());

  for (qit = qlist.begin(); qit != qend; ++qit)
    (*qit)->_m_qList = 0;
  qlist.clear();
}

// ************************************************************************* //
void	Shape::query(std::list<Edge *>	&qlist,
		     const BBox3d	&b)
{
  std::list<Edge *>::iterator	eit, eend(_m_edges.end());

  for (eit = _m_edges.begin(); eit != eend; ++eit)
    if (b.contains((*eit)->_m_s0->_m_p) || b.contains((*eit)->_m_s1->_m_p))
      (*eit)->pushFrontInQueryList(qlist);
}

// ************************************************************************* //
void	Shape::insert(std::list<Edge *> &qlist)
{
  std::list<Edge *>::iterator	qit, qend(qlist.end());

  for (qit = qlist.begin(); qit != qend; ++qit)
    (*qit)->_m_qList = 0;
  qlist.clear();
}

// ************************************************************************* //
Surfel	&Shape::insert(Surfel &news)
{
  _m_surfels.push_front(&news);
  news._m_meshHandle = _m_surfels.begin();
  return news;
}

// ************************************************************************* //
Edge	&Shape::insert(Surfel	&s0,
		       Surfel	&s1)
{
  Edge	*newe = new Edge(s0, s1);

  _m_edges.push_front(newe);
  newe->_m_meshHandle = _m_edges.begin();
  return *newe;
}

// ************************************************************************* //
Triangle	&Shape::connect(Edge	&e0,
				Edge	&e1,
				Edge	&e2)
{
  Triangle *newt = new Triangle(e0, e1, e2);

  e0.connectTo(*newt);
  e1.connectTo(*newt);
  e2.connectTo(*newt);
  _m_triangles.push_front(newt);
  newt->_m_meshHandle = _m_triangles.begin();
  return *newt;
}

// ************************************************************************* //
Triangle	&Shape::insert(Edge	&e0,
			       Edge	&e1,
			       Edge	&e2)
{
  Triangle *newt = new Triangle(e0, e1, e2);

  _m_triangles.push_front(newt);
  newt->_m_meshHandle = _m_triangles.begin();
  return *newt;
}

// ************************************************************************* //
void	Shape::remove(Surfel &s)
{
  s.remFromQueryList();
  _m_surfels.erase(s._m_meshHandle);
  delete &s;
}

// ************************************************************************* //
void	Shape::remove(Edge &e)
{
  e.remFromQueryList();
  _m_edges.erase(e._m_meshHandle);
  delete &e;
}

// ************************************************************************* */
void	Shape::remove(Triangle &t)
{
  _m_triangles.erase(t._m_meshHandle);
  delete &t;
}

// ************************************************************************* //
/// DESCRIPTION: Splits the edge (s0,s2).
///   Returns the new surfel and the 4 new edges.
// ************************************************************************* //
Surfel	&Shape::split(Edge	&e,
		      Edge	*edges[4])
{
  Triangle	&t0(*e._m_t0), &t1(*e._m_t1);
  Surfel	&s0(*e._m_s0), &s2(*e._m_s1);
  Edge		&e0(t0.gete(e)), &e1(t0.gete(e0));
  Edge		&e2(t1.gete(e)), &e3(t1.gete(e2));
  Surfel	&s1(e0.gets(s0)), &s3(e2.gets(s2));
  Surfel	&s(insert(*new Surfel));
  Edge		&e4(insert(s, s0)), &e5(insert(s, s1));
  Edge		&e6(insert(s, s2)), &e7(insert(s, s3));
  Triangle	&t4(insert(e0, e5, e4)), &t5(insert(e1, e6, e5));
  Triangle	&t6(insert(e2, e7, e6)), &t7(insert(e3, e4, e7));

  e.split(s);
  e0.replace(t0, t4);
  e1.replace(t0, t5);
  e2.replace(t1, t6);
  e3.replace(t1, t7);
  e4.connectTo(t4, t7);
  e5.connectTo(t5, t4);
  e6.connectTo(t6, t5);
  e7.connectTo(t7, t6);
  remove(e);
  remove(t1);
  remove(t0);
  if (edges)
    edges[0] = &e4, edges[1] = &e6, edges[2] = &e5, edges[3] = &e7;
  return s;
}

/* ************************************************************************* **
** DESCRIPTION: Splits the edge s0 s2.					     **
** Return value: the new vertex.					     **
**   New edges are pushed in list.					     **
** NOTE:    v2								     **
**         /|\ 								     **
**        / | \ 							     **
**      e2  e6 e1							     **
** t1   /t6 | t5\    t0							     **
**     /    |    \ 							     **
**   v3++e7+v+e5++v1							     **
**     \ t7 | t4 / 							     **
**      \   |   / 							     **
**      e3  e4 e0							     **
**        \ | / 							     **
**         \|/ 								     **
**          v0								     **
** ************************************************************************* */
bool	Shape::split(std::list<Edge *>::iterator	&eit,
		     Surfel				*&newSurfel,
		     Edge				*newEdges[4],
		     const std::list<Surfel *>		*const sharps)
{
  Edge		&e(**eit);
  Triangle	&t0(*e._m_t0), &t1(*e._m_t1);
  Surfel	&s0(*e._m_s0), &s2(*e._m_s1);
  Edge		&e0(t0.gete(e)), &e1(t0.gete(e0));
  Edge		&e2(t1.gete(e)), &e3(t1.gete(e2));
  Surfel	&s1(e0.gets(s0)), &s3(e2.gets(s2));
  Surfel	&s(insert(*new Surfel));
  Edge		&e4(insert(s, s0)), &e5(insert(s, s1));
  Edge		&e6(insert(s, s2)), &e7(insert(s, s3));
  Triangle	&t4(insert(e0, e5, e4));
  Triangle	&t5(insert(e1, e6, e5));
  Triangle	&t6(insert(e2, e7, e6));
  Triangle	&t7(insert(e3, e4, e7));

  if (sharps)
    e.split(s, *sharps);
  else
    e.split(s);
  newSurfel = &s;
  if (newEdges)
    newEdges[0] = &e4, newEdges[1] = &e5, newEdges[2] = &e6, newEdges[3] = &e7;
  e0.replace(t0, t4);
  e1.replace(t0, t5);
  e2.replace(t1, t6);
  e3.replace(t1, t7);
  e4.connectTo(t4, t7);
  e5.connectTo(t5, t4);
  e6.connectTo(t6, t5);
  e7.connectTo(t7, t6);
  ++eit;
  remove(e);
  remove(t1);
  remove(t0);
  return true;
}

/* ************************************************************************* **
** DESCRIPTION: Collapses the edge: first, tests wheather v0 or v2 is a      **
**   'peak', i.e. connected to 2 edges/2 faces only. If true, does a special **
**    collapse, else does a normal collapse (plus a little test for	     **
**    preserving topology).						     **
** NOTE:			 				             **
**      /       \|/       \    /  \	|     /	 \			     **
**     #-------- v2--------#  #.   \   	|    /	 /#		             **
**	\       /|\       /    \`.  \   |   /  ,'/		             **
**	 \ t2  / | \  t1 /      \ `. \  |  / ,'	/			     **
**	  \  e2  |  e1	/        \t2`.\ | /,'t1/			     **
**	   \ /   |   \ /          \   `\|/'   /				     **
**	   v3 t5 e t4 v1          v3-e2-v2-e1-v1			     **
**	   / \   |   / \          /   ,/|\.   \				     **
**        /  e3  |  e0  \        /t3,'/ | \`.t0\			     **
**       / t3  \ | /  t0 \      / ,' /  |  \ `.	\			     **
**      /       \|/	  \    /,'  /   |   \  `.\			     **
**     #---------v0--------#  #'   /  	|    \   `#			     **
**      \       /|\       /    \  /     |     \	 /			     **
** In case of a bug, check this:					     **
**   deleted edge, undeleted triangle					     **
**   wrongly deleted edge: not corresponding iterator/list		     **
**   order between ++eit and edge deletion				     **
** ************************************************************************* */
bool	Shape::collapse(std::list<Edge *>::iterator	&eit,
			Surfel				*&newSurfel,
			Edge				*newEdges[2])
{
  Edge		&e(**eit);
  Surfel	&s0(*e._m_s0), &s2(*e._m_s1);
  Triangle	&t4(*e._m_t0), &t5(*e._m_t1);
  Edge		&e0(t4.gete(e)), &e1(t4.gete(e0));
  Edge		&e2(t5.gete(e)), &e3(t5.gete(e2));
  Triangle	&t0(e0.gett(t4));
  Triangle	&t2(e2.gett(t5)), &t3(e3.gett(t5));

  if (&e1 == &e2)
    {
      e.collapse(s0);
      e0.replace(t4, t3);
      t3.replace(e3, e0);
      newSurfel = &s0;
      s0.remFromQueryList();
      newEdges[0] = &e0;
      e0.remFromQueryList();
      newEdges[1] = 0;
      remove(s2);
      remove(e2), remove(e3);
      remove(t4), remove(t5);
      ++eit, remove(e);
      return true;
    }
  if (&e0 == &e3)
    {
      e.collapse(s2);
      e1.replace(t4, t2);
      t2.replace(e2, e1);
      newSurfel = &s2;
      s2.remFromQueryList();
      newEdges[0] = &e1;
      e1.remFromQueryList();
      newEdges[1] = 0;
      remove(s0);
      remove(e0), remove(e2);
      remove(t4), remove(t5);
      ++eit, remove(e);
      return true;
    }

  Edge::circulator ecir0(s0, e), eend0(s0, e), eend2(s2, e);
  
  for (++++ecir0, --eend0, --eend2; ecir0 != eend0; ++ecir0)
    {
      Edge::circulator	ecir2(s2, e);

      for (++++ecir2; ecir2 != eend2; ++ecir2)
	if (ecir0->_m_s0 == ecir2->_m_s0 || ecir0->_m_s0 == ecir2->_m_s1 || 
	    ecir0->_m_s1 == ecir2->_m_s0 || ecir0->_m_s1 == ecir2->_m_s1)
	  {
	    ++eit;
	    return false;
	  }
    }
  e.collapse(s2);

  e.replaceAll(s0, s2);

  t0.replace(e0, e1);
  t3.replace(e3, e2);
  e1.replace(t4, t0);
  e2.replace(t5, t3);
  newSurfel = &s2;
  s2.remFromQueryList();
  newEdges[0] = &e1;
  e1.remFromQueryList();
  newEdges[1] = &e2;
  e2.remFromQueryList();
  remove(t4), remove(t5);
  remove(e0), remove(e3);
  remove(s0);
  ++eit, remove(e);
  return true;
}

// ************************************************************************* //
/// DESCRIPTION: Empties the mesh. All the geometry is cleared and gone.
///   Ignores the query list.
// ************************************************************************* //
void	Shape::clear()
{
  std::list<Surfel *>::iterator	sit, send(_m_surfels.end());
  std::list<Edge *>::iterator	eit, eend(_m_edges.end());
  std::list<Triangle *>::const_iterator	tit, tend(_m_triangles.end());

  for (sit = _m_surfels.begin(); sit != send; ++sit)
    delete *sit;
  _m_surfels.clear();
  for (eit = _m_edges.begin(); eit != eend; ++eit)
    delete *eit;
  _m_edges.clear();
  for (tit = _m_triangles.begin(); tit != tend; ++tit)
    delete *tit;
  _m_triangles.clear();
}

// ************************************************************************* //
/// DESCRIPTION: Saves the shape in output file stream ofs. 
///  Format is Surfels[<data>]\nTriangles[<data>]
//// Surfels format is binary.
// ************************************************************************* //
void	Shape::save(std::ofstream &ofs)
{
  std::list<Surfel *>::const_iterator	sit, send(_m_surfels.end());
  std::list<Triangle *>::const_iterator	tit, tend(_m_triangles.end());
  std::map<Surfel *, int>	smap;
  int	i(0);

  ofs << "Material [" << Surfel::_s_color << "] " << std::endl;
  ofs << "Surfels [";
  --send;
  for (sit = _m_surfels.begin(), i = 0; sit != send; ++sit, ++i)
    {
      smap[*sit] = i;
      ofs << **sit << ",";
    }
  smap[*sit] = i;
  ofs << *_m_surfels.back();
  ofs << "] " << std::endl;
  ofs << "Triangles [";
  --tend;
  for (tit = _m_triangles.begin(); tit != tend; ++tit)
    {
      ofs << smap[&(*tit)->gets0()] << " "
	  << smap[&(*tit)->gets1()] << " "
	  << smap[&(*tit)->gets2()] << ",";
    }
  ofs << smap[&(_m_triangles.back())->gets0()] << " "
      << smap[&(_m_triangles.back())->gets1()] << " "
      << smap[&(_m_triangles.back())->gets2()] << "]";
}

// ************************************************************************* //
/// DESCRIPTION: Loads what has been saved with save.
// ************************************************************************* //
void	Shape::load(std::ifstream &ifs)
{
  typedef std::pair<unsigned, unsigned>	IndexPair;

  std::list<IndexPair >	edges;
  std::list<IndexPair >::iterator	eit, eend;
  std::list<triplet<IndexPair, IndexPair, IndexPair > >	triangles;
  std::list<triplet<IndexPair, IndexPair, IndexPair > >::iterator tit, tend;
  std::map<unsigned, Surfel *>	surfelMap;
  std::map<IndexPair, Edge *>	edgeMap;
  std::string	buf;
  char		c;
  Point3d	p;
  Vector3d	v;
  unsigned	s0, s1, s2, i(0);
  IndexPair	p0, p1, p2;
  Surfel	*newSurfel;
  Edge		*newEdge;

  clear();
  ifs >> buf;//Material
  ifs >> c >> Surfel::_s_color >> c;
  ifs >> buf;//Surfels
  ifs >> c;
  while (c != ']')
    {
      ifs >> p >> v;
      surfelMap[i++] = (newSurfel = new Surfel(p, v));
      insert(*newSurfel);
      ifs >> c;
    }
  ifs >> buf;//Triangles
  ifs >> c;
  while (c != ']')
    {
      ifs >> s0 >> s1 >> s2;
      edges.push_back(p0 = (s0 < s1)? std::make_pair(s0, s1): 
		      std::make_pair(s1, s0));
      edges.push_back(p1 = (s1 < s2)? std::make_pair(s1, s2): 
		      std::make_pair(s2, s1));
      edges.push_back(p2 = (s2 < s0)? std::make_pair(s2, s0): 
		      std::make_pair(s0, s2));
      triangles.push_back(make_triplet(p0, p1, p2));
      ifs >> c;
    }
  edges.sort();
  edges.unique();
  for (eit = edges.begin(), eend = edges.end(); eit != eend; ++eit)
    {
      newEdge = &insert(*surfelMap[(*eit).first], *surfelMap[(*eit).second]);
      edgeMap[*eit] = newEdge;
    }
  for (tit = triangles.begin(), tend = triangles.end(); tit != tend; ++tit)
    connect(*edgeMap[(*tit).first], *edgeMap[(*tit).second],
	    *edgeMap[(*tit).third]);
}

// ************************************************************************* //
/// DESCRIPTION: Exports a renderman file.
// ************************************************************************* //
void	Shape::riRender(std::ofstream &ofs)
{
  std::list<Triangle *>::const_iterator	tit, tend(_m_triangles.end());

  /*
  ofs <<
    "#include <ri.h>\n"
    "RtColor Color = {.2, .6, .2};\n"
    "\n"
    "int	main()\n"
    "{\n"
    "  RiBegin(RI_NULL);\n"
    "  {\n"
    "  RiTransformBegin();\n"
    "  RiLightSource(\"distantlight\", RI_NULL);\n"
    "  RiTransformEnd();\n"
    "  RiProjection(\"perspective\", RI_NULL);\n"
    "  RiTranslate(0.0, 0.0, 1.5);\n"
    "  RiRotate(40.0, -1.0, 1.0, 0.0);\n"
    "  RiDisplay(\"ri.tif\", RI_FILE, RI_RGBA, RI_NULL);\n"
    "  RiWorldBegin();\n"
    "  RiSurface(\"matte\", RI_NULL);\n"
    "  RiColor(Color);\n" << std::endl;
  for (tit = _m_triangles.begin(), tend = _m_triangles.end(); 
       tit != tend; ++tit)
    (*tit)->riRender(ofs);
  ofs <<
    "  RiWorldEnd();\n"
    "  }\n"
    "  RiEnd();\n"
    "}\n" << std::endl;
  */

  for (tit = _m_triangles.begin(), tend = _m_triangles.end(); 
       tit != tend; ++tit)
    (*tit)->riRender(ofs);
}

// ************************************************************************* //
/// DESCRIPTION: Exports a vrml shape.
// ************************************************************************* //
void	Shape::exportVRML(std::ofstream &)
{
  //TODO: put back. Removed in earlier version.
}

// ************************************************************************* //
void	Shape::cleanShape()
{
  Vector3d	n0, n1;
  std::list<Edge *>::iterator	erit, erend(_m_edges.end());
  const double	flipAngle = 0.2;

  for (erit = _m_edges.begin(); erit != erend; ++erit)
    {
      (*erit)->_m_t0->computeNormalp(n0);
      (*erit)->_m_t1->computeNormalp(n1);
      if (n0 * n1 <= flipAngle)
	{
	  (*erit)->flip();
	  (*erit)->_m_t0->computeNormalp(n0);
	  (*erit)->_m_t1->computeNormalp(n1);
	  if (n0 * n1 <= flipAngle)
	    (*erit)->flip();
	}
    }
}

// ************************************************************************* //
void	Shape::renormalize()
{
  std::list<Triangle *>::iterator	tit(_m_triangles.begin());
  std::list<Triangle *>::iterator	tend(_m_triangles.end());
  std::list<Surfel *>::iterator		sit, send(_m_surfels.end());
  Vector3d	n;

  for (sit = _m_surfels.begin(); sit != send; ++sit)
    (*sit)->_m_v = 0.0;
  for (; tit != tend; ++tit)
    {
      (*tit)->computeNormalp(n);
      n *= (*tit)->computeArea();
      (*tit)->gets0()._m_v += n;
      (*tit)->gets1()._m_v += n;
      (*tit)->gets2()._m_v += n;
    }
  for (sit = _m_surfels.begin(); sit != send; ++sit)
    {
      (*sit)->_m_v.eqUnit();
      (*sit)->saveSurfel();
    }
}

// ************************************************************************* //
/// DESCRIPTION:
/// Inserts vertex v on triangle t.
///             v1
///            /|` 
///           / " ` 
///          / / ` ` 
///         /  | |  ` 
///        /  /   `  ` 
///       /  v3----v5 ` 
///      /  / \   / `  `  
///     / .'   \ /   `. ` 
///    / /    __v4_    ` ` 
///   /_'_,--'     `--._``` 
/// v2---------------------v0
/// ************************************************************************ //
Triangle	&Shape::insertTriangle(const Point3d	&vIns,
				       Triangle		&t,
				       Surfel		*newSurfels[3],
				       Edge		*newEdges[9],
				       Triangle		*newTriangles[7])
{
  const Vector3d	n((t.gets0()._m_v + t.gets1()._m_v +
			   t.gets2()._m_v).unit());
  Surfel	&v0(t.gets0()), &v1(t.gets1()), &v2(t.gets2());
  Surfel	&v3(insert(*new Surfel(vIns, n)));
  Surfel	&v4(insert(*new Surfel(vIns, n)));
  Surfel	&v5(insert(*new Surfel(vIns, n)));
  Edge		&e01(*t._m_e2);
  Edge		&e12(*t._m_e0);
  Edge		&e20(*t._m_e1);
  Edge		&e34(insert(v3, v4));
  Edge		&e45(insert(v4, v5));
  Edge		&e53(insert(v5, v3));
  Edge		&e04(insert(v0, v4));
  Edge		&e05(insert(v0, v5));
  Edge		&e15(insert(v1, v5));
  Edge		&e13(insert(v1, v3));
  Edge		&e23(insert(v2, v3));
  Edge		&e24(insert(v2, v4));
  Triangle	&t345(insert(e34, e45, e53));
  Triangle	&t243(insert(e24, e34, e23));
  Triangle	&t054(insert(e05, e45, e04));
  Triangle	&t135(insert(e13, e53, e15));
  Triangle	&t312(insert(e12, e23, e13));
  Triangle	&t420(insert(e20, e04, e24));
  Triangle	&t501(insert(e01, e15, e05));

  e12.replace(t, t312);
  e20.replace(t, t420);
  e01.replace(t, t501);
  e34.connectTo(t345, t243);
  e45.connectTo(t345, t054);
  e53.connectTo(t345, t135);
  e04.connectTo(t054, t420);
  e05.connectTo(t054, t501);
  e15.connectTo(t135, t501);
  e13.connectTo(t135, t312);
  e23.connectTo(t243, t312);
  e24.connectTo(t243, t420);
  remove(t);
  newSurfels[0] = &v3;
  newSurfels[1] = &v4;
  newSurfels[2] = &v5;
  newEdges[0] = &e45;
  newEdges[1] = &e53;
  newEdges[2] = &e34;
  newEdges[3] = &e04;
  newEdges[4] = &e05;
  newEdges[5] = &e15;
  newEdges[6] = &e13;
  newEdges[7] = &e23;
  newEdges[8] = &e24;
  newTriangles[0] = &t345;
  newTriangles[1] = &t243;
  newTriangles[2] = &t054;
  newTriangles[3] = &t135;
  newTriangles[4] = &t312;
  newTriangles[5] = &t420;
  newTriangles[6] = &t501;
  return t345;
}


// ************************************************************************* //
bool	Shape::isSharp(const Triangle		&t, 
		       const std::list<Surfel *>&sharpSurfels)
{
  bool	b0(false), b1(false), b2(false);
  std::list<Surfel *>::const_iterator	sit;

  for (sit = sharpSurfels.begin(); sit != sharpSurfels.end(); ++sit)
    if (*sit == &t.gets0())
      b0 = true;
    else if (*sit == &t.gets1())
      b1 = true;
    else if (*sit == &t.gets2())
      b2 = true;
  return b0 && b1 && b2;
}

// ************************************************************************* //
/// DESCRIPTION: Joins the two contours.
/// steps are:
///  - find pairs of edges which minimize distance.
///  - reorder contour1 to match contour0.
///  - replace the surfels
///  - replace triangles and edges. 
// ************************************************************************* //
void	Shape::connect(std::list<Edge *>	&contour0,
		       std::list<Edge *>	&contour1,
		       const Point3d		&center)
{
  std::list<Surfel *>	sharpSurfels;
  std::list<Edge *>::iterator	eit, ecirc, ejt, ebegin, ekt, elt;
  std::list<Triangle *>	sharpTriangles;
  std::list<Triangle *>::iterator	tit;
  std::list<Edge *>	reordered;
  double	emin = MAXFLOAT, etmp;
  Surfel	*swap[2];
  
  for (eit = contour0.begin(); eit != contour0.end(); ++eit)
    sharpSurfels.push_back((*eit)->_m_s0), 
      sharpSurfels.push_back((*eit)->_m_s1);
  sharpSurfels.sort();
  sharpSurfels.unique();
  for (ejt = contour1.begin(); ejt != contour1.end(); ++ejt)
    {
      etmp = 0.0;
      for (eit = contour0.begin(), ecirc = ejt;
	   eit != contour0.end(); ++eit)
	{
	  etmp += min(((*eit)->_m_s0->_m_p - (*ecirc)->_m_s0->_m_p).norm2() +
		      ((*eit)->_m_s1->_m_p - (*ecirc)->_m_s1->_m_p).norm2(),
		      ((*eit)->_m_s0->_m_p - (*ecirc)->_m_s1->_m_p).norm2() +
		      ((*eit)->_m_s1->_m_p - (*ecirc)->_m_s0->_m_p).norm2());
	  ++ecirc;
	  if (ecirc == contour1.end())
	    ecirc = contour1.begin();
	}
      if (etmp < emin)
	emin = etmp, ebegin = ejt;
    }
  eit = ebegin;
  do 
    {
      reordered.push_back(*eit);
      if (++eit == contour1.end())
	eit = contour1.begin();
    }
  while (eit != ebegin);
  for (eit = contour0.begin(), ejt = reordered.begin(); ejt != reordered.end();
       ++eit, ++ejt)
    {
      ekt = eit;
      if (++ekt == contour0.end())
	ekt = contour0.begin();
      elt = ejt;
      if (++elt == reordered.end())
	elt = reordered.begin();
      swap[0] = ((*ejt)->_m_s0 == (*elt)->_m_s0 || 
		 (*ejt)->_m_s0 == (*elt)->_m_s1)? (*ejt)->_m_s0: (*ejt)->_m_s1;
      swap[1] = ((*eit)->_m_s0 == (*ekt)->_m_s0 || 
		 (*eit)->_m_s0 == (*ekt)->_m_s1)? (*eit)->_m_s0: (*eit)->_m_s1;
      swap[1]->_m_p = center + 1.0e-4 * 
	((swap[0]->_m_p + swap[1]->_m_p) * 0.5 - center).unit();
      (*ejt)->replaceAll(*swap[0], *swap[1]);
    }
  for (eit = contour0.begin(), ejt = reordered.begin();
       eit != contour0.end(); ++eit, ++ejt)
    {
      Triangle *s0, *t0, *s1, *t1;

      if (isSharp(*(*eit)->_m_t0, sharpSurfels))
	s0 = (*eit)->_m_t0, t0 = (*eit)->_m_t1;
      else 
	s0 = (*eit)->_m_t1, t0 = (*eit)->_m_t0;
      sharpTriangles.push_back(s0);
      if (isSharp(*(*ejt)->_m_t0, sharpSurfels))
	s1 = (*ejt)->_m_t0, t1 = (*ejt)->_m_t1;
      else 
	s1 = (*ejt)->_m_t1, t1 = (*ejt)->_m_t0;
      sharpTriangles.push_back(s1);
      (*eit)->replace(*s0, *t1);
      t1->replace(**ejt, **eit);
    }
  sharpTriangles.sort();
  sharpTriangles.unique();
  for (tit = sharpTriangles.begin(); tit != sharpTriangles.end(); ++tit)
    remove(**tit);
  for (eit = reordered.begin(); eit != reordered.end(); ++eit)
    remove(**eit);
}

/* ************************************************************************* **
** DESCRIPTION: Connects the loops j0,j1... and k0k1... and minimizes 	     **
**   distortion.							     **
**           u3                                u3  minimizing correspondance **
**          / \                                /|\      k0 k1 k2 k3 k4 k5    **
**         /   \                              /:|:\   = j3 j4 j5 j0 j1 j2    **
**        /  s0 \                            /::|::\                         **
**      u2-------u1                        j4---j3--j2  //                   **
**      / \`.s  / \`.                      /:\`.   /:\`//                    **
**     /   \ `./   \ `.  v3               /::_j5`j1_::\/`.  v3               **
**    / s1  \ /`.s2 \  `. \              /__/::\ /`.\__\  `.|\               **
**  u4-------u0-----u5 / `.\           u4-------j0-----u5 /:|.\              **
**             `.  `. /  t0`.                     `// `. /::|:`.             **
**               `.  v2------v1                   //`.  k4--k3--k2           **
**                 `. \  t  / \                  //   `.:\     /:\           **
**                 / `.\   /   \                //    /:`.k5 k1_::\          **
**                / t1 `. /  t2 \                    /__/:`. /::\__\         **
**              v4-------v0-----v5                 v4-------k0-----v5        **
** ************************************************************************* */
void	Shape::hexaMerge(Triangle	&s,
			 Triangle	&t,
			 Surfel		*newSurfels[3],
			 Edge		*newEdges[12])
{
  Surfel	&j0(s.gets0()), &j2(s.gets1()), &j4(s.gets2());
  Edge		&u1u2(*s._m_e0), &u2u0(*s._m_e1), &u0u1(*s._m_e2);

  Surfel	&j3(insert(*new Surfel()));
  Surfel	&j5(insert(*new Surfel()));
  Surfel	&j1(insert(*new Surfel()));

  Surfel	&v0(t.gets0()), &v1(t.gets2()), &v2(t.gets1());
  Edge		&v1v2(*t._m_e0), &v2v0(*t._m_e2), &v0v1(*t._m_e1);

  Triangle	&s0(u1u2.gett(s)), &t0(v1v2.gett(t));
  Triangle	&s1(u2u0.gett(s)), &t1(v2v0.gett(t));
  Triangle	&s2(u0u1.gett(s)), &t2(v0v1.gett(t));

  Surfel	&u3(s0.gets(u1u2)), &v3(t0.gets(v1v2));
  Surfel	&u4(s1.gets(u2u0)), &v4(t1.gets(v2v0));
  Surfel	&u5(s2.gets(u0u1)), &v5(t2.gets(v0v1));

  Edge	&u3j2(s0.gete(j4)), &u3j4(s0.gete(j2));
  Edge	&u4j4(s1.gete(j0)), &u4j0(s1.gete(j4));
  Edge	&u5j0(s2.gete(j2)), &u5j2(s2.gete(j0));

  Edge	&v3k2(t0.gete(v2)), &v3k4(t0.gete(v1));
  Edge	&v4k4(t1.gete(v0)), &v4k0(t1.gete(v2));
  Edge	&v5k0(t2.gete(v1)), &v5k2(t2.gete(v0));

  Edge &j0j1(insert(j0, j1)), &j1j2(insert(j1, j2));
  Edge &j2j3(insert(j2, j3)), &j3j4(insert(j3, j4));
  Edge &j4j5(insert(j4, j5)), &j5j0(insert(j5, j0));

  Edge &j3u3(insert(j3, u3));
  Edge &j5u4(insert(j5, u4));
  Edge &j1u5(insert(j1, u5));

  Triangle	&u5j0j1(insert(j1u5, j0j1, u5j0));
  Triangle	&u5j2j1(insert(u5j2, j1j2, j1u5));
  Triangle	&u3j2j3(insert(j3u3, j2j3, u3j2));
  Triangle	&u3j4j3(insert(u3j4, j3j4, j3u3));
  Triangle	&u4j4j5(insert(j5u4, j4j5, u4j4));
  Triangle	&u4j0j5(insert(u4j0, j5j0, j5u4));

  Point3d	getsp[6], getsq[6];
  const Point3d	vSnap((j0._m_p + j2._m_p + j4._m_p +
		       v0._m_p + v1._m_p + v2._m_p) * (1.0 / 6.0));

  getsp[0] = vSnap + 1.0e-3 * (j0._m_p - vSnap).unit();
  getsp[2] = vSnap + 1.0e-3 * (j2._m_p - vSnap).unit();
  getsp[4] = vSnap + 1.0e-3 * (j4._m_p - vSnap).unit();
  getsp[1] = vSnap + 1.0e-3 * ((getsp[0] + getsp[2]) * 0.5 - vSnap).unit();
  getsp[3] = vSnap + 1.0e-3 * ((getsp[2] + getsp[4]) * 0.5 - vSnap).unit();
  getsp[5] = vSnap + 1.0e-3 * ((getsp[4] + getsp[0]) * 0.5 - vSnap).unit();

  getsq[0] = vSnap + 1.0e-3 * (v0._m_p - vSnap).unit();
  getsq[2] = vSnap + 1.0e-3 * (v1._m_p - vSnap).unit();
  getsq[4] = vSnap + 1.0e-3 * (v2._m_p - vSnap).unit();
  getsq[1] = vSnap + 1.0e-3 * ((getsq[0] + getsq[2]) * 0.5 - vSnap).unit();
  getsq[3] = vSnap + 1.0e-3 * ((getsq[2] + getsq[4]) * 0.5 - vSnap).unit();
  getsq[5] = vSnap + 1.0e-3 * ((getsq[4] + getsq[0]) * 0.5 - vSnap).unit();

  Surfel	*getsk[] = { &j0, &j1, &j2, &j3, &j4, &j5 };
  Edge		*getskvk[] = { &j0j1, &j1j2, &j2j3, &j3j4, &j4j5, &j5j0 };

  int		shift(0);
  double	dMin(MAXFLOAT), dTmp;

  for (int shiftTmp(0); shiftTmp < 6; ++shiftTmp)
    {
      dTmp = 0.0;
      for (int j(0); j < 6; ++j)
	dTmp += (getsq[j] - (getsp[(j + shiftTmp) % 6])).norm2();
      if (dTmp < dMin)
	dMin = dTmp, shift = shiftTmp;
    }

  Surfel &k0(*getsk[(0 + shift) % 6]);
  Surfel &k1(*getsk[(1 + shift) % 6]);
  Surfel &k2(*getsk[(2 + shift) % 6]);
  Surfel &k3(*getsk[(3 + shift) % 6]);
  Surfel &k4(*getsk[(4 + shift) % 6]);
  Surfel &k5(*getsk[(5 + shift) % 6]);

  Edge &k1k0(*getskvk[(0 + shift) % 6]);
  Edge &k2k1(*getskvk[(1 + shift) % 6]);
  Edge &k3k2(*getskvk[(2 + shift) % 6]);
  Edge &k4k3(*getskvk[(3 + shift) % 6]);
  Edge &k5k4(*getskvk[(4 + shift) % 6]);
  Edge &k0k5(*getskvk[(5 + shift) % 6]);

  Edge &k3v3(insert(k3, v3));
  Edge &k5v4(insert(k5, v4));
  Edge &k1v5(insert(k1, v5));

  Triangle	&v5k1k0(insert(k1v5, v5k0, k1k0));
  Triangle	&v5k2k1(insert(v5k2, k1v5, k2k1));
  Triangle	&v3k3k2(insert(k3v3, v3k2, k3k2));
  Triangle	&v3k4k3(insert(v3k4, k3v3, k4k3));
  Triangle	&v4k5k4(insert(k5v4, v4k4, k5k4));
  Triangle	&v4k0k5(insert(v4k0, k5v4, k0k5));
  
  const Vector3d vec[6] = 
    { (vSnap - (getsp[(0 + shift) % 6] + getsq[0]) * 0.5).unit(),
      (vSnap - (getsp[(1 + shift) % 6] + getsq[1]) * 0.5).unit(),
      (vSnap - (getsp[(2 + shift) % 6] + getsq[2]) * 0.5).unit(),
      (vSnap - (getsp[(3 + shift) % 6] + getsq[3]) * 0.5).unit(),
      (vSnap - (getsp[(4 + shift) % 6] + getsq[4]) * 0.5).unit(),
      (vSnap - (getsp[(5 + shift) % 6] + getsq[5]) * 0.5).unit()};

  k0._m_v = vec[0];
  k0._m_p = vSnap - vec[0] * 1.0e-2;
  k1._m_v = vec[1];
  k1._m_p = vSnap - vec[1] * 1.0e-2;
  k2._m_v = vec[2];
  k2._m_p = vSnap - vec[2] * 1.0e-2;
  k3._m_v = vec[3];
  k3._m_p = vSnap - vec[3] * 1.0e-2;
  k4._m_v = vec[4];
  k4._m_p = vSnap - vec[4] * 1.0e-2;
  k5._m_v = vec[5];
  k5._m_p = vSnap - vec[5] * 1.0e-2;

  t.reconnect(k0, k4, k2);

  u3j2.replace(s0, u3j2j3);
  u3j4.replace(s0, u3j4j3);
  u4j4.replace(s1, u4j4j5);
  u4j0.replace(s1, u4j0j5);
  u5j0.replace(s2, u5j0j1);
  u5j2.replace(s2, u5j2j1);

  v3k2.replace(t0, v3k3k2);
  v3k4.replace(t0, v3k4k3);
  v4k4.replace(t1, v4k5k4);
  v4k0.replace(t1, v4k0k5);
  v5k0.replace(t2, v5k1k0);
  v5k2.replace(t2, v5k2k1);

  j2j3.connectTo(u3j2j3, v3k3k2);
  j3j4.connectTo(u3j4j3, v3k4k3);

  j4j5.connectTo(u4j4j5, v4k5k4);
  j5j0.connectTo(u4j0j5, v4k0k5);

  j0j1.connectTo(u5j0j1, v5k1k0);
  j1j2.connectTo(u5j2j1, v5k2k1);

  j3u3.connectTo(u3j2j3, u3j4j3);
  j5u4.connectTo(u4j4j5, u4j0j5);
  j1u5.connectTo(u5j0j1, u5j2j1);
  k3v3.connectTo(v3k3k2, v3k4k3);
  k5v4.connectTo(v4k5k4, v4k0k5);
  k1v5.connectTo(v5k1k0, v5k2k1);

  remove(s);
  remove(s0);
  remove(s1);
  remove(s2);
  remove(u1u2);
  remove(u2u0);
  remove(u0u1);

  remove(t);
  remove(t0);
  remove(t1);
  remove(t2);
  remove(v1v2);
  remove(v2v0);
  remove(v0v1);

  remove(v0);
  remove(v1);
  remove(v2);

  if (newSurfels)
    newSurfels[0] = &j3, newSurfels[1] = &j5, newSurfels[2] = &j1;
  if (newEdges)
    newEdges[0] = &j0j1, newEdges[1] = &j1j2, newEdges[2] = &j2j3,
      newEdges[3] = &j3j4, newEdges[4] = &j4j5, newEdges[5] = &j5j0, 
      newEdges[6] = &j3u3, newEdges[7] = &j5u4, newEdges[8] = &j1u5, 
      newEdges[9] = &k3v3, newEdges[10] = &k5v4, newEdges[11] = &k1v5;
}

/* ************************************************************************* */
double	Shape::volume() const
{
  std::list<Triangle *>::const_iterator	tit, tend(_m_triangles.end());
  double		vol(0.0);

  for (tit = _m_triangles.begin(); tit != tend; ++tit)
    {
      const Point3d	&p0((*tit)->gets0()._m_p);
      const Point3d	&p1((*tit)->gets1()._m_p);
      const Point3d	&p2((*tit)->gets2()._m_p);

      vol += Vector3d(p0) * ((p1 - p0) ^ (p2 - p0));
    }
  return vol / 6.0;
}
