// ************************************************************************* //
// AUTHOR: Alexis Angelidis.
// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
// ************************************************************************* //
#ifndef	__GLOCTREE_HH__
#define	__GLOCTREE_HH__

// ************************************************************************* //
#include <math.h>
#include <stdlib.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>

#include <string>
#include <list>
#include <vector>
#include <fstream>

// ************************************************************************* //
#include "gfxTypes.h"

// ************************************************************************* //
class	Tool;
class	Surfel;
class	Edge;
class	Triangle;
class	BBox3d;

// ************************************************************************* //
/// DESCRIPTION: Holds a grid
/// VALUE CONSTRUCTOR ATTRIBUTE:
///  depth..... greater or equal to 1.
///  bitmask.... 2^(depth - 1), greater or equal to 1, edge length of root.
/// Rendered cube:          Data format:
///  (-.,.)--(0,.)--(.,.)	*----*----*----*(3,3)
///       |    |    |           |    |    |    |
///       |    |    |           |    |    |    | Example: depth == 2
///  (-.,0)--(0,0)--(.,0)       *----*----*----*
///       |    |    |           |    |    |    |
///       |    |    |           |    |	  |    |
/// (-.,-.)--(0,-.)-(.,-.)	*----*----*----*
///                             |    |    |    |
///                             |    |    |    |
///                        (0,0)*----*----*----*(3,0)
/// TODO: sample a sphere.
// ************************************************************************* //
class Shape
{
  friend class		Tool;
  friend class		Sweep;
  friend class		SweepSingle;
  friend class		SweepSym;
  friend class		SweepPinch;
  friend class		SweepVolume;
  friend class		SweepProto;
  friend class		SweepZoom;
  friend class		SweepStokes;
  std::list<Surfel *>	_m_surfels;
  std::list<Edge *>	_m_edges;
  std::list<Triangle *>	_m_triangles;
public:
  Shape();
  ~Shape();
  void	copy(const Shape &s);
  void	glRender() const;
  void	glRender2() const;
  void	glRender3() const;
  void	query(std::list<Surfel *> &qlist, const BBox3d &b);
  void	query(std::list<Edge *> &qlist, const BBox3d &b);
  void	insert(std::list<Surfel *> &qlist);
  void	insert(std::list<Edge *> &qlist);
  Triangle	&insertTriangle(const Point3d &vIns, Triangle &t,
				Surfel *newVertices[3], Edge *newEdges[9],
				Triangle *newTriangles[7]);
  Surfel	&insert(Surfel &s);
  Edge		&insert(Surfel &s0, Surfel &s1);
  Triangle	&insert(Edge &e0, Edge &e1, Edge &e2);
  Triangle	&connect(Edge &e0, Edge &e1, Edge &e2);
  void		remove(Surfel &s);
  void		remove(Edge &e);
  void		remove(Triangle &t);
  void	initEicosahedra();
  void	initBall();
  bool	split(std::list<Edge *>::iterator &eit, Surfel *&newSurfel,
	      Edge *newEdges[4] = 0,
	      const std::list<Surfel *> *const sharps = 0);
  Surfel	&split(Edge &e, Edge *edges[2] = 0);
  bool	collapse(std::list<Edge *>::iterator &e, Surfel *&newSurfel,
		 Edge *newEdges[2] = 0);
  void	save(std::ofstream &ofs);
  void	exportVRML(std::ofstream &ofs);
  void	riRender(std::ofstream &ofs);
  void	load(std::ifstream &ifs);
  void	clear();
  void	cleanShape();
  void	renormalize();
  void	reset();
  void	hexaMerge(Triangle &s, Triangle &t, Surfel *newSurfels[3],
		  Edge *newEdges[12]);
  void	connect(std::list<Edge *> &contour0, std::list<Edge *> &contour1,
		const Point3d &center);
  bool	isSharp(const Triangle &t, const std::list<Surfel *> &sharpSurfels);
  unsigned	nbSurfels() const { return _m_surfels.size(); }
  unsigned	nbTriangles() const { return _m_triangles.size(); }
  double	volume() const;
};

// ************************************************************************* //
#endif//__GLOCTREE_HH__
