// ************************************************************************* //
// AUTHOR: Alexis Angelidis.
// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
// ************************************************************************* //
#include "swpSurfel.h"

// ************************************************************************* //
Vector3d	Surfel::_s_color(0.2, 0.4, 1.0);

// ************************************************************************* //
Surfel::Surfel(const Point3d	&p,
	       const Vector3d	&v): 
  _m_p(p), _m_oldp(p), _m_v(v), _m_oldv(v), _m_color(1.0), _m_detJac(1.0), 
  _m_qList(0)
{
}

// ************************************************************************* //
Surfel::Surfel(): _m_color(1.0), _m_detJac(1.0), _m_qList(0)
{
}

// ************************************************************************* //
/// DESCRIPTION: Renders a patch arrount _m_p, of size 1.0 / 2.0 (moraless).
// ************************************************************************* //
void	Surfel::glRender() const
{
  _m_v.glNormal();
  _m_p.glVertex();
}

// ************************************************************************* //
/// DESCRIPTION: Renders a patch arrount _m_p, of size 1.0 / 2.0 (moraless).
// ************************************************************************* //
void	Surfel::glRender2() const
{
  static const Vector3d	colin(0.75);

  if (_m_color == 0.0)
    _s_color.glColor();
  else if (_m_color == 1.0)
    colin.glColor();
  else if (0.2 <= _m_color)
    {
      double dk = (_m_color - 0.2) / (1.0 - 0.2);

      dk = 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
      (0.75 * (dk * _s_color + Vector3d(1.0 - dk))).glColor();
    }
  else
    {
      double dk = (0.2 - _m_color) / 0.2;
	  
      dk = 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
      ((0.75 * dk - dk + 1.0) * _s_color).glColor();
    }
  _m_v.glNormal();
  _m_p.glVertex();
}

// ************************************************************************* //
/// DESCRIPTION: Renders a patch arrount _m_p, of size 1.0 / 2.0 (moraless).
// ************************************************************************* //
void	Surfel::glRender3() const
{
  static const Vector3d	colin(0.75);

  const double k = _m_detJac;

  if (k <= 1.0)
    {
      const double kmax  = max(0.0, 1.0 - k);

      Point3d(2.0 * kmax, 0.0, 1.0).glColor();
    }
  else 
    {
      const double kmin  = min(1.0, k - 1.0);

      Point3d(0.0, 2.0 * kmin, 1.0).glColor();
    }
  _m_v.glNormal();
  _m_p.glVertex();
}

// ************************************************************************* //
void	Surfel::saveSurfel()
{
  _m_oldp = _m_p;
  _m_oldv = _m_v;
  _m_detJac = 1.0;
}

// ************************************************************************* //
/// DESCRIPTION: The edge HAS to be in a querry list.
// ************************************************************************* //
void	Surfel::remFromQueryList()
{
  if (_m_qList)
    _m_qList->erase(_m_qHandle), _m_qList = 0;
}

// ************************************************************************* //
void	Surfel::pushFrontInQueryList(std::list<Surfel *> &qlist)
{
  _m_qList = &qlist;
  _m_qList->push_front(this);
  _m_qHandle = _m_qList->begin();
}

// ************************************************************************* //
void	Surfel::pushBackInQueryList(std::list<Surfel *> &qlist)
{
  _m_qList = &qlist;
  _m_qList->push_back(this);
  _m_qHandle = --_m_qList->end();
}

// ************************************************************************* //
std::ostream	&operator<<(std::ostream	&os,
			    const Surfel	&s)
{
  return os << s._m_p << s._m_v;
}
