// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Nov 11 08:48:06 NZDT 2000
// ************************************************************************* //
#include "swpShape.h"
#include "swpSweep.h"
#include "swpSurfel.h"
#include "swpEdge.h"
#include "swpTriangle.h"
#include "swpTransform.h"
#include "swpTool.h"

// ************************************************************************* //
#ifndef MAXINT
#define MAXINT	2147483647
#endif//MAXINT
#ifndef MAXUINT
#define MAXUINT 0xFFFFFFFF
#endif//MAXUINT
#ifndef MAXFLOAT
#define MAXFLOAT 3.40282347e+38F
#endif//MAXFLOAT

// ************************************************************************* //
/// DESCRIPTION: Sweeps just a few steps.
// ************************************************************************* //
void	Sweep::sweepStep(Shape &) const
{
  //TODO: sweep a few steps.
}

// ************************************************************************* //
/// DESCRIPTION: Sweeps the entire path of any tool over the mesh.
// ************************************************************************* //
void	Sweep::sweep(Shape	&s) const
{
  if (!_m_steps)
    return;

  std::list<Surfel *>		slist;
  std::list<Surfel *>::iterator	sit, send(slist.end());
  std::list<Edge *>		elist;
  std::list<Edge *>::iterator	eit, eend(elist.end());
  Surfel	*newSurfel;
  Edge		*newEdges[4];
  unsigned	c(0), cmax;

  query(s, slist);
  query(s, elist);
  cmax = 10 * elist.size() / _m_steps;
  for ( ; _m_istep < _m_steps; ++_m_istep, c = 0)
    {
      for (sit = slist.begin(), send = slist.end(); sit != send; ++sit)
	{
	  (*sit)->saveSurfel();
	  sweepStep(**sit);
	}
      //elist.sort(Edge::geq);
      for (eit = elist.begin(), eend = elist.end(); eit != eend && c < cmax; )
	if (!(*eit)->isTooLong())
	  ++eit;
	else if (s.split(eit, newSurfel, newEdges))
	  {
	    ++c;
	    newSurfel->pushBackInQueryList(slist);
	    sweepStep(*newSurfel);
	    for (unsigned i(0); i < 4; ++i)
	      if (newEdges[i])
		newEdges[i]->pushBackInQueryList(elist);
	  }
      //elist.sort(Edge::leq);
      for (eit = elist.begin(), eend = elist.end(); eit != eend; )
	if (!(*eit)->isTooShort())
	  ++eit;
	else if (s.collapse(eit, newSurfel, newEdges))
	  {
	    newSurfel->pushBackInQueryList(slist);
	    sweepStep(*newSurfel);
	    for (unsigned i(0); i < 2; ++i)
	      if (newEdges[i])
		newEdges[i]->pushBackInQueryList(elist);
	  }
      //s.cleanShape();
    }
  s.insert(slist);
  s.insert(elist);
}

// ************************************************************************* //
/// DESCRIPTION: Surfel query. This function can be overloaded.
// ************************************************************************* //
void	Sweep::query(Shape			&s,
		     std::list<Surfel *>	&slist) const
{
  s.query(slist, _m_bbox);
}

// ************************************************************************* //
/// DESCRIPTION: Edge query. This function can be overloaded.
// ************************************************************************* //
void	Sweep::query(Shape		&s,
		     std::list<Edge *>	&elist) const
{
  s.query(elist, _m_bbox);
}

// ************************************************************************* //
SweepSingle::SweepSingle(const Tool		&tool,
			 const Matrix4d		&msrc,
			 const Transformation	&transf):
  Sweep(msrc, transf), _m_tool(tool)
{
  Transformation	*tcur = transf.copy();
  Matrix4d		m, invm;

  _m_bbox = tool.getBBox(msrc, msrc.inverse());
  _m_steps = 2 * transf.nbSteps(_m_bbox, tool.getCoating());
  std::cout << "\rsteps = " << _m_steps << " " << std::flush;
  _m_tstep = transf.copy();
  *_m_tstep ^= 1.0 / _m_steps;
  for (unsigned istep = 0; istep < _m_steps; ++istep)
    {
#ifndef WIN32
		tcur->cast(transf);
#else
		delete tcur;
		tcur = transf.copy();
#endif
		*tcur ^= istep / double(_m_steps);
		m = (*tcur) * msrc;
		invm = m.inverse();
		_m_sweptbbox.push_back(tool.getBBox(m, invm));
		_m_bbox.fill(_m_sweptbbox.back());
		_m_msteps.push_back(m);
		_m_invmsteps.push_back(invm);
	}
  delete tcur;
}

// ************************************************************************* //
SweepSingle::~SweepSingle()
{
  delete _m_tstep;
}

// ************************************************************************* //
void	SweepSingle::sweepStep(Point3d	&p) const
{
  (*_m_tstep)(p, _m_tool.field(p, _m_msteps[_m_istep],
			       _m_invmsteps[_m_istep]));
}

// ************************************************************************* //
void	SweepSingle::sweepStep(Surfel	&s) const
{
  const double	f(_m_tool.field(s._m_p, _m_msteps[_m_istep], 
				_m_invmsteps[_m_istep]));

  if (f == 0.0)
    return;
  /*
  if (f == 1.0)
    {
      (*_m_tstep)(s._m_p, s._m_v);//???
      s._m_detJac = _m_tstep->det();//???
    }
  else*/
    {
      Point3d	&p(s._m_p), px(p), py(p), pz(p);
      
      (*_m_tstep)(p, f);
      
      px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;
      sweepStep(px);
      sweepStep(py);
      sweepStep(pz);

      Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);

      s._m_detJac = j.det();
      j.eqCommatrix();
      s._m_v *= j;
      s._m_v.eqUnit();
    }
}

// ************************************************************************* //
void	SweepSingle::sweep(Shape &s) const
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());

  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    (*sit)->saveSurfel();
  Sweep::sweep(s);
}

// ************************************************************************* //
void	SweepSingle::glRender(const Tool	&tool)
{
  glPushMatrix();
  tool.glRender();
  glPopMatrix();
}

// ************************************************************************* //
SweepSym::SweepSym(const Tool		&tool,
		   const Matrix4d	&msrc,
		   const Transformation	&transf,
		   const unsigned	sym):
  Sweep(msrc, transf), _m_tool(tool), _m_sym(sym)
{
  Transformation	*tcur = transf.copy();
  Matrix4d		m, invm;
  Matrix4d		mSym, invmSym;
  BBox3d		bboxSym;

  _m_bbox = tool.getBBox(msrc, msrc.inverse());
  _m_steps = 2 * transf.nbSteps(_m_bbox, tool.getCoating());
  std::cout << "\rsteps = " << _m_steps << " " << std::endl;
  _m_tstep = transf.copy();
  _m_tstepSym = transf.copy();
  *_m_tstep ^= 1.0 / _m_steps;
  _m_mlogstep = _m_tstep->asMatrix();
  _m_mlogstep.eqLog();
  *_m_tstepSym ^= 1.0 / _m_steps;
  _m_tstepSym->eqSym(sym);
  _m_mlogstepSym = _m_tstepSym->asMatrix();
  _m_mlogstepSym.eqLog();
  for (unsigned istep = 0; istep < _m_steps; ++istep)
    {
#ifndef WIN32
		tcur->cast(transf);
#else
		delete tcur;
		tcur = transf.copy();
#endif
      *tcur ^= istep / double(_m_steps);
      mSym = m = (*tcur) * msrc;
      mSym.eqSym(sym);
      invm = m.inverse();
      invmSym = mSym.inverse();
      _m_sweptbbox.push_back(tool.getBBox(m, invm));
      _m_bbox.fill(_m_sweptbbox.back());
      _m_msteps.push_back(m);
      _m_mstepsSym.push_back(mSym);
      _m_invmsteps.push_back(invm);
      _m_invmstepsSym.push_back(invmSym);
    }
  delete tcur;
  bboxSym = _m_bbox;
  bboxSym.eqSym(sym);
  if ((_m_intersect = bboxSym.intersects(_m_bbox)))
    _m_bbox.fill(bboxSym);
}

// ************************************************************************* //
SweepSym::~SweepSym()
{
  delete _m_tstep;
  delete _m_tstepSym;
}

// ************************************************************************* //
void	SweepSym::glRender(const Tool		&tool,
			   const unsigned	sym)
{
  glFrontFace(GL_CW);
  glPushMatrix();
  Matrix4d::sym(sym).glMultMatrix();
  tool.glRender();
  glPopMatrix();

  glFrontFace(GL_CCW);
  glPushMatrix();
  tool.glRender();
  glPopMatrix();
}

// ************************************************************************* //
void	SweepSym::sweepFarAway(Shape &s) const
{
  Transformation	*tSym = _m_transf.copy();

  tSym->eqSym(_m_sym);

  SweepSingle	swpSym(_m_tool, Matrix4d::sym(_m_sym) * _m_msrc, *tSym);
  SweepSingle	swp(_m_tool, _m_msrc, _m_transf);

  swpSym.Sweep::sweep(s);
  swp.Sweep::sweep(s);
  delete tSym;
}

// ************************************************************************* //
void	SweepSym::sweepStepUni0(Point3d	&p) const
{
  (*_m_tstep)(p, _m_tool.field(p, _m_msteps[_m_istep], 
			       _m_invmsteps[_m_istep]));
}

// ************************************************************************* //
void	SweepSym::sweepStepUni1(Point3d	&p) const
{
  (*_m_tstepSym)(p, _m_tool.field(p, _m_mstepsSym[_m_istep], 
				  _m_invmstepsSym[_m_istep]));
}

// ************************************************************************* //
void	SweepSym::sweepStepSym(Point3d	&p) const
{
  const double	f(_m_tool.field(p, _m_msteps[_m_istep], 
				_m_invmsteps[_m_istep]));
  const double	fSym(_m_tool.field(p, _m_mstepsSym[_m_istep],
				   _m_invmstepsSym[_m_istep]));
  const double	co(1.0 - (f * fSym) / (f + fSym));

  p *= (_m_mlogstep * (co * f) + _m_mlogstepSym * (co * fSym)).exp();
}

// ************************************************************************* //
void	SweepSym::sweepStep(Surfel	&s) const
{
  const double	f(_m_tool.field(s._m_p, _m_msteps[_m_istep], 
				_m_invmsteps[_m_istep]));
  const double	fSym(_m_tool.field(s._m_p, _m_mstepsSym[_m_istep],
				   _m_invmstepsSym[_m_istep]));

  if (f == 0.0 && fSym == 0.0)
    return;
  else if (f == 1.0 && fSym == 0.0)
    {
      (*_m_tstep)(s._m_p, s._m_v);
      s._m_detJac = _m_tstep->det();
    }
  else if (f == 0.0 && fSym == 1.0)
    {
      (*_m_tstepSym)(s._m_p, s._m_v);
      s._m_detJac = _m_tstepSym->det();
    }
  else
    {
      Point3d	&p(s._m_p), px(p), py(p), pz(p);

      px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;
      if (fSym == 0.0)
	{
	  (*_m_tstep)(p, f);
	  sweepStepUni0(px);
	  sweepStepUni0(py);
	  sweepStepUni0(pz);
	}
      else if (f == 0.0)
	{
	  (*_m_tstepSym)(p, fSym);
	  sweepStepUni1(px);
	  sweepStepUni1(py);
	  sweepStepUni1(pz);
	}
      else
	{
	  const double	co(1.0 - (f * fSym) / (f + fSym));

	  p *= (_m_mlogstep * (co * f) + _m_mlogstepSym * (co * fSym)).exp();
	  sweepStepSym(px);
	  sweepStepSym(py);
	  sweepStepSym(pz);
	}

      Matrix3d	j((px - p) * 1.0e12,
		  (py - p) * 1.0e12,
		  (pz - p) * 1.0e12);
      
      s._m_detJac = j.det();
      j.eqCommatrix();
      s._m_v *= j;
      s._m_v.eqUnit();
    }
}

// ************************************************************************* //
void	SweepSym::sweep(Shape &s) const
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());
  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    (*sit)->saveSurfel();

  if (!_m_intersect)
    sweepFarAway(s);
  else
    Sweep::sweep(s);
}

// ************************************************************************* //
SweepPinch::SweepPinch(const Tool		&tool,
		       const Matrix4d		&msrc,
		       const Transformation	&transf,
		       const unsigned		)://sym
  Sweep(msrc, transf), _m_tool(tool), _m_invmsrc(msrc.inverse())
{
  _m_steps = 1024;
  _m_bbox = tool.getBBox(msrc, _m_invmsrc);
  _m_vSnap = _m_msrc * Point3d(0.0);
  _m_vPos[0] = _m_msrc * Point3d(0.0, 0.0, 1.0 - 1.0e-6);
  _m_vPos[1] = _m_msrc * Point3d(0.0, 0.0, -1.0 + 1.0e-6);
  _m_zAxis = _m_msrc * Vector3d(0.0, 0.0, 1.0 / (_m_steps / 2.0));
}

// ************************************************************************* //
bool	SweepPinch::beginSweep(Shape &s)
{
  // *********************************************************************** //
  Point3d	vInt[2];

  intersect(s, vInt);
  if (!_m_toFuse[0] || !_m_toFuse[1])
    return false;

  //std::list<Edge *>	edges[2];
  //std::list<Edge *>::iterator	eit;

  // *********************************************************************** //
  // Snap vertex.
  for (int j(0); j < 2; ++j)
    {
#if 0
		Triangle* tj = _m_toFuse[j];
		Surfel& tj0 = tj->gets0();
		Surfel& tj1 = tj->gets1();
		Surfel& tj2 = tj->gets2();
		tj0._m_p = vInt[j] + (tj0._m_p - vInt[j]).unit() * 1.0e-2;
		tj1._m_p = vInt[j] + (tj1._m_p - vInt[j]).unit() * 1.0e-2;
		tj2._m_p = vInt[j] + (tj2._m_p - vInt[j]).unit() * 1.0e-2;

		_m_sharpSurfels.push_back(&tj0);
		_m_sharpSurfels.push_back(&tj1);
		_m_sharpSurfels.push_back(&tj2);
		_m_sharpContour[j].push_back(tj->_m_e0);
		_m_sharpContour[j].push_back(tj->_m_e1);
		_m_sharpContour[j].push_back(tj->_m_e2);
#else
		_m_toFuse[j]->gets0()._m_p = 
		  vInt[j] + (_m_toFuse[j]->gets0()._m_p - vInt[j]).unit() * 1.0e-2;
		_m_toFuse[j]->gets1()._m_p = 
		  vInt[j] + (_m_toFuse[j]->gets1()._m_p - vInt[j]).unit() * 1.0e-2;
		_m_toFuse[j]->gets2()._m_p = 
		  vInt[j] + (_m_toFuse[j]->gets2()._m_p - vInt[j]).unit() * 1.0e-2;

      _m_sharpSurfels.push_back(&_m_toFuse[j]->gets0());
      _m_sharpSurfels.push_back(&_m_toFuse[j]->gets1());
      _m_sharpSurfels.push_back(&_m_toFuse[j]->gets2());
      _m_sharpContour[j].push_back(_m_toFuse[j]->_m_e0);
      _m_sharpContour[j].push_back(_m_toFuse[j]->_m_e1);
      _m_sharpContour[j].push_back(_m_toFuse[j]->_m_e2);
#endif
    }
  _m_sharpContour[1].reverse();

  std::list<Surfel *>::iterator	sit;

  for (sit = s._m_surfels.begin(); sit != s._m_surfels.end(); ++sit)
    (*sit)->saveSurfel();

  //Edge		*newEdges[2];
  //Surfel	*oldSurfels[2];

  //TODO: Nice Split algorithm

  //Split connecting polygon
  /*
  for (int k(0); k < 2; ++k)
    for (int j(0); j < 2; ++j)
      {
	edges[j] = _m_sharpContour[j];
	_m_sharpContour[j].clear();
	for (eit = edges[j].begin(); eit != edges[j].end(); ++eit)
	  {
	    Surfel	&s0(s.split(**eit, newEdges));

	    s0._m_p = vInt[j] + (s0._m_p - vInt[j]).unit() * 1.0e-2;
	    _m_sharpSurfels.push_back(&s0);
	    if (newEdges[0]->_m_s0 == _m_sharpContour[j].back()->_m_s0 ||
		newEdges[0]->_m_s0 == _m_sharpContour[j].back()->_m_s1 ||
		newEdges[0]->_m_s1 == _m_sharpContour[j].back()->_m_s0 ||
		newEdges[0]->_m_s1 == _m_sharpContour[j].back()->_m_s1)
	      _m_sharpContour[j].push_back(newEdges[0]),
		_m_sharpContour[j].push_back(newEdges[1]);
	    else
	      _m_sharpContour[j].push_back(newEdges[1]),
		_m_sharpContour[j].push_back(newEdges[0]);
	  }
      }
  */
  return true;
}

// ************************************************************************* //
void	SweepPinch::intersect(Shape	&s,
			      Point3d	vInt[2]) const
{
  std::list<Edge *>	elist;
  std::list<Edge *>::iterator	eit, eend(elist.end());
  Triangle	*tmp;

  _m_toFuse[0] = _m_toFuse[1] = 0;
  s.query(elist, _m_bbox);
  for (eit = elist.begin();
       eit != eend && !(_m_toFuse[0] && _m_toFuse[1]); ++eit)
    for (int i(0); i < 4; ++i)
      {
	tmp = &(*eit)->gett(i);
	if (!_m_toFuse[0] &&
	    tmp->intersects(_m_vSnap, _m_vPos[0], vInt[0]))
	  _m_toFuse[0] = tmp;
	else if (!_m_toFuse[1] &&
		 tmp->intersects(_m_vSnap, _m_vPos[1], vInt[1]))
	  _m_toFuse[1] = tmp;
      }
  s.insert(elist);
}

// ************************************************************************* //
void	SweepPinch::sweepPinch(Point3d	&p) const
{
  const double	f = _m_tool.field(p, _m_msrc, _m_invmsrc, _m_istep);

  if (f == 0.0)
    return;
  else if (f == 1.0)
    p -= _m_zAxis;
  else
    p -= f * _m_zAxis;
}

// ************************************************************************* //
void	SweepPinch::sweepPinch(Surfel	&s) const
{
  const double	f(_m_tool.field(s._m_p, _m_msrc, _m_invmsrc, _m_istep));

  if (f == 0.0)
    return;
  else
    {
      Point3d	&p(s._m_p), px(p), py(p), pz(p);

      p -= f * _m_zAxis;

      px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;
      sweepPinch(px);
      sweepPinch(py);
      sweepPinch(pz);

      Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);
      
      s._m_detJac = j.det();
      j.eqCommatrix();
      s._m_v *= j;
      s._m_v.eqUnit();
    }
}

// ************************************************************************* //
void	SweepPinch::sweepPinch(Shape &s) const
{
  std::list<Surfel *>		slist;
  std::list<Surfel *>::iterator	sit, send(slist.end());
  std::list<Edge *>		elist;
  std::list<Edge *>::iterator	eit, eend(elist.end());
  Surfel	*newSurfel;
  Edge		*newEdges[4];
  unsigned	c(0), cmax(2048);

  for (sit = s._m_surfels.begin(), send = s._m_surfels.end(); sit != send;
       ++sit)
    (*sit)->saveSurfel();
  s.query(slist, _m_bbox);
  s.query(elist, _m_bbox);
  for (unsigned istep = 0; _m_istep < (_m_steps / 2) && istep < 128;
       ++istep, ++_m_istep, c = 0)
    {
      for (sit = slist.begin(), send = slist.end(); sit != send; ++sit)
	{
	  (*sit)->saveSurfel();
	  sweepPinch(**sit);
	}
      elist.sort(Edge::geq);
      for (eit = elist.begin(), eend = elist.end(); eit != eend && c < cmax; )
	if (!(*eit)->isTooLong(_m_sharpSurfels))
	  ++eit;
	else if (s.split(eit, newSurfel, newEdges, &_m_sharpSurfels))
	  {
	    ++c;
	    newSurfel->pushBackInQueryList(slist);
	    sweepPinch(*newSurfel);
	    for (unsigned i(0); i < 4; ++i)
	      if (newEdges[i])
		newEdges[i]->pushBackInQueryList(elist);
	  }
      elist.sort(Edge::leq);
      for (eit = elist.begin(), eend = elist.end(); eit != eend; )
	if (!(*eit)->isTooShort(_m_sharpSurfels))
	  ++eit;
	else if (s.collapse(eit, newSurfel, newEdges))
	  {
	    sweepPinch(*newSurfel);
	    newSurfel->pushBackInQueryList(slist);
	    for (unsigned i(0); i < 2; ++i)
	      if (newEdges[i])
		newEdges[i]->pushBackInQueryList(elist);
	  }
    }
  s.insert(slist);
  s.insert(elist);
}

// ************************************************************************* //
void	SweepPinch::sweepPull(Point3d	&p) const
{
  const double		f(_m_tool.field(p, _m_msrc, _m_invmsrc, _m_istep));
  const Point3d		c(_m_invmsrc * p);
  double		xy;
  const double		xyOut(1.0e-6), xylen(hypot(c(0), c(1)));
  const Vector3d	axis(_m_msrc * Vector3d(c(0), c(1), 0.0).unit());
  const double		invsteps(0.5 / _m_steps);

  if (f == 0.0)
    return;
  if (xylen == 0.0)
    return;
  if (xylen < xyOut)
    {
      xy = xylen / xyOut;
      p += (1.0 - xy * xy * (3.0 - 2.0 * xy)) * invsteps * f * axis;
      return;
    }
  if (f == 1.0)
    p += axis * invsteps;
  else
    p += axis * (invsteps * f);
}

// ************************************************************************* //
void	SweepPinch::sweepPull(Surfel	&s) const
{
  const double	f(_m_tool.field(s._m_p, _m_msrc, _m_invmsrc, _m_istep));

  if (f == 0.0)
    return;

  Point3d	&p(s._m_p), px(p), py(p), pz(p);

  px(0) += 1.0e-12, py(1) += 1.0e-12, pz(2) += 1.0e-12;
  sweepPull(p);
  sweepPull(px);
  sweepPull(py);
  sweepPull(pz);
  
  Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);

  s._m_detJac = j.det();
  j.eqCommatrix();
  s._m_v *= j;
  s._m_v.eqUnit();
}

// ************************************************************************* //
void	SweepPinch::sweepPull(Shape &s) const
{
  std::list<Surfel *>		slist;
  std::list<Surfel *>::iterator	sit, send(slist.end());
  std::list<Edge *>		elist;
  std::list<Edge *>::iterator	eit, eend(elist.end());
  Surfel	*newSurfel;
  Edge		*newEdges[4];
  unsigned	c(0), cmax(512);

  for (sit = s._m_surfels.begin(), send = s._m_surfels.end(); sit != send;
       ++sit)
    (*sit)->saveSurfel();
  s.query(slist, _m_bbox);
  s.query(elist, _m_bbox);
  for (unsigned istep = 0; _m_istep < _m_steps && istep < 32;
       ++istep, ++_m_istep, c = 0)
    {
      for (sit = slist.begin(), send = slist.end(); sit != send; ++sit)
	{
	  (*sit)->saveSurfel();
	  sweepPull(**sit);
	}
      elist.sort(Edge::geq);
      for (eit = elist.begin(), eend = elist.end(); eit != eend && c < cmax; )
	{
	  const Vector3d	p0((*eit)->_m_s0->_m_p - _m_vSnap);
	  const Vector3d	p1((*eit)->_m_s1->_m_p - _m_vSnap);
	  const Vector3d	n(_m_zAxis.unit());

	  if (!(*eit)->isTooLong() ||
	      ((p0 - (p0 * n) * n).norm2() < sq(2.0e-1) &&
	       (p1 - (p1 * n) * n).norm2() < sq(2.0e-1)))
	    ++eit;
	  else if (s.split(eit, newSurfel, newEdges, &_m_sharpSurfels))
	    {
	      ++c;
	      newSurfel->pushBackInQueryList(slist);
	      sweepPull(*newSurfel);
	      for (unsigned i(0); i < 4; ++i)
		if (newEdges[i])
		  newEdges[i]->pushBackInQueryList(elist);
	    }
	}
      elist.sort(Edge::leq);
      for (eit = elist.begin(), eend = elist.end(); eit != eend; )
	{
	  const Vector3d	p0((*eit)->_m_s0->_m_p - _m_vSnap);
	  const Vector3d	p1((*eit)->_m_s1->_m_p - _m_vSnap);
	  const Vector3d	n(_m_zAxis.unit());

	  if (!(*eit)->isTooShort() ||
	      ((p0 - (p0 * n) * n).norm2() < sq(4.0e-1) &&
	       (p1 - (p1 * n) * n).norm2() < sq(4.0e-1)))
	    ++eit;
	  else if (s.collapse(eit, newSurfel, newEdges))
	    {
	      sweepPinch(*newSurfel);
	      newSurfel->pushBackInQueryList(slist);
	      for (unsigned i(0); i < 2; ++i)
		if (newEdges[i])
		  newEdges[i]->pushBackInQueryList(elist);
	    }
	}
    }
  s.insert(slist);
  s.insert(elist);
}

// ************************************************************************* //
void	SweepPinch::sweepStep(Shape &s) const
{
  std::cout << _m_istep << " / " << _m_steps << std::endl;
  std::cout << "Pinching... steps = " << _m_steps << " " << std::endl;

  if (_m_istep < _m_steps / 2)
    sweepPinch(s);

  if (_m_istep == _m_steps / 2)
    {
      std::cout << "DRILLING" << std::endl;
      s.connect(_m_sharpContour[0], _m_sharpContour[1], _m_vSnap);
      ++_m_istep;
    }
  /*
    if (_m_steps / 2 <= _m_istep)
    sweepPull(s);*/
  std::cout << "done." << std::endl;
}

