// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sun Dec 21 10:18:20 NZDT 2003
// ************************************************************************* //
#ifndef	__SWPSWEEPPROTO_HH__
#define	__SWPSWEEPPROTO_HH__

// ************************************************************************* //
#include "swpSweep.h"

// ************************************************************************* //
class RigidLog;

// ************************************************************************* //
template <class T>
class Table
{
  const int	_m_i, _m_j;
  T		*_m_m;
public:
  Table(const int i, const int j): _m_i(i), _m_j(j), _m_m(new T[i * j]) {}
  virtual ~Table() { delete[] _m_m; }
  T		&operator()(const int i, const int j)
  { return _m_m[i * _m_j + j]; }
  const T	&operator()(const int i, const int j) const 
  { return _m_m[i * _m_j + j]; }
};

// ************************************************************************* //
class SweepProto: public Sweep
{
  const ToolProto	&_m_tool;
  const unsigned 	_m_nbSpheres;
  Table<Matrix4d>	*_m_mlog;
  double		*_m_fields;
  const unsigned	_m_sym;
  bool			_m_intersect;//true if symmetry intersects this.
  Table<Point3d>	*_m_sphereCenters;
  Point3d		*_m_c;
  Vector3d		_m_t;
  const double		_m_r;
public:
  SweepProto(const ToolProto &tool, const Matrix4d &msrc,
	      const Transformation &transf, const unsigned sym);
  virtual ~SweepProto();
  void	swirlStep(Point3d &s) const;
  void	sweepStep(Point3d &s) const;
  void	sweepStep(Surfel &s) const;
  void	sweep(Shape &s) const;
  virtual bool	transformGizmo() const { return true; }
  virtual void	query(Shape &s, std::list<Surfel *> &slist) const;
  virtual void	query(Shape &s, std::list<Edge *> &elist) const;
  double	field(const Point3d &p, const unsigned i) const;
  unsigned	nbSteps() const;
};

// ************************************************************************* //
#endif//__SWPSWEEPPROTO_HH__

