// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sat Jan 17 19:30:36 NZDT 2004
// ************************************************************************* //
#include "swpSweepZoom.h"
#include "swpToolZoom.h"
#include "swpSurfel.h"
#include "swpShape.h"

// ************************************************************************* //
#ifndef SQRT3
#define SQRT3 1.7320508075688772935274
#endif//SQRT3

// ************************************************************************* //
SweepZoom::SweepZoom(const ToolZoom		&tool,
		     const Matrix4d		&msrc,
		     const Transformation	&transf,
		     const unsigned		): 
  Sweep(msrc, transf), _m_tool(tool), _m_invmsrc(msrc.inverse()),
  _m_r(1.0)
{
  const Matrix4d	mlog(transf.mlog());
  Vector3d		t, v;

  _m_c = _m_msrc * Point3d(0.0);
 
  Point3d	c(_m_c);
  Point3d	u(_m_c + Point3d(1.0, 0.0, 0.0));
  
  transf(c);
  transf(u);

  _m_t = (c - u).norm() - 1.0;
  if (_m_t < 0.0)
    _m_t = log((c - u).norm());

  t = (_m_world2solver * Vector3d(mlog * Point3d(0.0)));
  _m_steps = max(1, int(ceil(t.norm() / v.norm())));
  
  std::cout << _m_t << std::endl;
  std::cout << _m_steps << std::endl;

  _m_tstep = transf.copy();
  (*_m_tstep) ^= 1.0 / _m_steps;
  _m_bbox = tool.getBBox(msrc, _m_invmsrc);
}

// ************************************************************************* //
SweepZoom::~SweepZoom()
{
  delete _m_tstep;
}

// ************************************************************************* //
void	SweepZoom::sweepStep(Point3d &p) const
{
  Vector3d	u(p - _m_c);
  const double	rho(u.norm());
  const double	l(pow(_m_t * _m_r * _m_r * _m_r + rho * rho * rho, 1.0 / 3.0));

  u /= rho;
  p = _m_c + u * l;
}

// ************************************************************************* //
void	SweepZoom::sweepStep(Surfel &s) const
{
  Point3d	&p(s._m_p), px(p), py(p), pz(p);

  px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;

  sweepStep(p);
  sweepStep(px);
  sweepStep(py);
  sweepStep(pz);

  Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);

  j.eqCommatrix();
  s._m_v *= j;
  s._m_v.eqUnit();
}

// ************************************************************************* //
void	SweepZoom::sweep(Shape &s) const
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());

  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    (*sit)->saveSurfel();
  Sweep::sweep(s);
}

