// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Thu Nov 20 17:49:55 NZDT 2003
// ************************************************************************* //
#ifndef __SWPTOOL_HH__
#define __SWPTOOL_HH__

// ************************************************************************* //
#include "gfxTypes.h"

// ************************************************************************* //
class Sweep;
class Shape;
class SweepProto;

// ************************************************************************* //
/// DESCRIPTION: Super class for coated tools (or 1 varying parameter).
/// transformGizmo	- returns true if sweeping transforms gizmo.
/// newSweep		- creates a new sweep for that tool.
/// field(..., u)	- for animating a field.
// ************************************************************************* //
class Tool
{
  friend class SweepProto;
protected:
  static bool	_s_isEmpty;
  double	_m_k, _m_invk, _m_k2;
  Matrix4d	_m_m;
public:
  Tool(const double k);
  virtual double	field(const Point3d &p, const Matrix4d &m,
			      const Matrix4d &invm) const = 0;
  virtual double	field(const Point3d &p, const Matrix4d &m,
			      const Matrix4d &invm, const unsigned) const
  { return field(p, m, invm); }
  virtual BBox3d	getBBox(const Matrix4d &m,
				const Matrix4d &invm) const = 0;
  virtual void		glRender() const {}
  static void		setEmpty(const bool ie) { _s_isEmpty = ie; }
  void			setCoating(const double k)
  { _m_k = k; _m_invk = 1.0 / k; _m_k2 = k * k; }
  virtual void		setPosition(const Matrix4d &m) { _m_m = m; }
  Matrix4d	getPosition() const { return _m_m; }
  virtual Sweep	*newSweep(const int sym, const Matrix4d &msrc,
			  const Transformation &transf) const = 0;
  void		colorize(const Matrix4d &msrc, Shape &s,
			 const unsigned	sym = 0);
  double	getCoating() const { return _m_k; }
};

// ************************************************************************* //
/// DESCRIPTION: Sweeping tool
// ************************************************************************* //
class ToolBall: public Tool
{
public:
  ToolBall(const double k);
  double	field(const Point3d &p, const Matrix4d &m,
		      const Matrix4d &invm) const;
  BBox3d	getBBox(const Matrix4d &m, const Matrix4d &invm) const;
  double	field2Ellipse(const double a2, const double b2,
			      const double c2, const Point3d &p) const;
  double	field2Ball(const double r2, const Point3d &p) const;
  void		glRender() const;
  Sweep		*newSweep(const int sym, const Matrix4d &msrc,
			  const Transformation &transf) const;
};

// ************************************************************************* //
#endif//__SWPTOOL_HH__
