// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Mon Dec  8 09:50:04 NZDT 2003
// ************************************************************************* //
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>

// ************************************************************************* //
#include "swpToolProto.h"
#include "swpSweepProto.h"

#ifndef MAXFLOAT
#define MAXFLOAT 3.40282347e+38F
#endif//MAXFLOAT

// ************************************************************************* //
/// DESCRIPTION: Holds extra information in the case tool is symmetric.
// ************************************************************************* //
ToolProto::ToolProto(const double k): 
  Tool(k), _m_nbSpheres(8)
{
}

// ************************************************************************* //
/// DESCRIPTION: Desctructor.
// ************************************************************************* //
ToolProto::~ToolProto()
{
}

// ************************************************************************* //
/// DESCRIPTION: Updates the position of the tool.
// ************************************************************************* //
void	ToolProto::setPosition(const Matrix4d	&m)
{
  Tool::setPosition(m);
  Tool::setCoating(2.0 * (m * Vector3d(1.0, 0.0, 0.0)).norm());
}

// ************************************************************************* //
/// DESCRIPTION: Returns the field for coloring the shape.
// ************************************************************************* //
double	ToolProto::field(const Point3d		&p,
			  const Matrix4d	&m,
			  const Matrix4d	&) const
{
  const Point3d	c(m * Point3d(0.0));
  const double	l2(Vector3d(p - c).norm2());
  
  if (4.0 * _m_k2 <= l2)
    return 0.0;

  const double	l(sqrt(l2)), dk(0.5 * l * _m_invk);

  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
/// DESCRIPTION: Returns bbox, without symmetry.
// ************************************************************************* //
BBox3d	ToolProto::getBBox(const Matrix4d	&m,
			   const Matrix4d	&) const //invm
{
  BBox3d	ret(m * Point3d(0.0));

  //ret.expand(2.0 * _m_k);
  ret.expand(4.0 * _m_k);
  return ret;
}

// ************************************************************************* //
Sweep	*ToolProto::newSweep(const int			sym,
			      const Matrix4d		&msrc,
			      const Transformation	&transf) const
{
  return new SweepProto(*this, msrc, transf, sym);
}

// ************************************************************************* //
void	ToolProto::glRender() const
{
  glPushMatrix();

  _m_r = (_m_m * Vector3d(1.0, 0.0, 0.0)).norm();

  const Point3d	c(_m_m * Point3d(0.0));

  glTranslated(c.x, c.y, c.z);

  GLUquadric *quad = gluNewQuadric();

  gluSphere(quad, _m_r, 32, 24);

  gluDeleteQuadric(quad);
  
  glPopMatrix();
}
