// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sat Jan 17 18:54:28 NZDT 2004
// ************************************************************************* //
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include "swpToolStokes.h"
#include "swpSweepStokes.h"

// ************************************************************************* //
ToolStokes::ToolStokes(const double k): Tool(k)
{
}

// ************************************************************************* //
double	ToolStokes::field(const Point3d		&p,
			  const Matrix4d	&m,
			  const Matrix4d	&) const
{
  const Point3d	c(m * Point3d(0.0));
  const double	r2((m * Vector3d(1.0, 0.0, 0.0)).norm2());
  const double	l2((p - c).norm2());
  double	dk, l, r;

  if (l2 <= r2)
    return 1.0;
  l = sqrt(l2);
  r = sqrt(r2);
  if (_m_k2 <= sq(l - r))
    return 0.0;
  dk = (l - r) * _m_invk;
  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
BBox3d	ToolStokes::getBBox(const Matrix4d &m,
			    const Matrix4d &) const
{
  BBox3d	ret(m * Point3d(0.0));

  ret.expand(_m_rad * 16.0 / 12.0 * 2.0);
  return ret;
}

// ************************************************************************* //
void	ToolStokes::glRender() const
{
  Vector3d	x, y, z(_m_v.unit());
  const double	size(_m_rad * 16.0 / 12.0);

  x = ((fabs(z.x) < 0.5)? Vector3d(1.0, 0.0, 0.0): Vector3d(0.0, 1.0, 0.0));
  x = (x - (x * z) * z).unit();
  y = (z ^ x);

  glColor3d(1.0, 0.0, 0.0);
  glBegin(GL_LINE_LOOP);
  (_m_p - x * size - y * size - z * size).glVertex();
  (_m_p - x * size + y * size - z * size).glVertex();
  (_m_p + x * size + y * size - z * size).glVertex();
  (_m_p + x * size - y * size - z * size).glVertex();
  glEnd();//GL_LINE_LOOP
  glColor3d(1.0, 1.0, 0.0);
  glBegin(GL_LINE_LOOP);
  (_m_p - x * size - y * size + z * size).glVertex();
  (_m_p - x * size + y * size + z * size).glVertex();
  (_m_p + x * size + y * size + z * size).glVertex();
  (_m_p + x * size - y * size + z * size).glVertex();
  glEnd();//GL_LINE_LOOP
  glBegin(GL_LINES);
  glColor3d(1.0, 0.0, 0.0);
  (_m_p - x * size - y * size - z * size).glVertex();
  glColor3d(1.0, 1.0, 0.0);
  (_m_p - x * size - y * size + z * size).glVertex();
  glColor3d(1.0, 0.0, 0.0);
  (_m_p - x * size + y * size - z * size).glVertex();
  glColor3d(1.0, 1.0, 0.0);
  (_m_p - x * size + y * size + z * size).glVertex();
  glColor3d(1.0, 0.0, 0.0);
  (_m_p + x * size + y * size - z * size).glVertex();
  glColor3d(1.0, 1.0, 0.0);
  (_m_p + x * size + y * size + z * size).glVertex();
  glColor3d(1.0, 0.0, 0.0);
  (_m_p + x * size - y * size - z * size).glVertex();
  glColor3d(1.0, 1.0, 0.0);
  (_m_p + x * size - y * size + z * size).glVertex();
  glEnd();//GL_LINES

  glPushMatrix();
  _m_m.glMultMatrix();

  GLUquadric *quad = gluNewQuadric();

  gluSphere(quad, 1.0, 32, 24);
  gluDeleteQuadric(quad);
  glPopMatrix();
}

// ************************************************************************* //
Sweep	*ToolStokes::newSweep(const int			sym,
			     const Matrix4d		&msrc,
			     const Transformation	&transf) const
{
  _m_rad = (msrc * Vector3d(1.0, 0.0, 0.0)).norm();
  _m_p = msrc * Point3d(0.0);
  _m_v = Vector3d(transf.mlog() * Point3d(0.0));

  return new SweepStokes(*this, msrc, transf, sym);
}
