// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Mon Dec  8 09:50:04 NZDT 2003
// ************************************************************************* //
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>

// ************************************************************************* //
#include "swpToolVolume.h"
#include "swpSweepVolume.h"

#ifndef MAXFLOAT
#define MAXFLOAT 3.40282347e+38F
#endif//MAXFLOAT

// ************************************************************************* //
/// DESCRIPTION: Holds extra information in the case tool is symmetric.
// ************************************************************************* //
ToolVolume::ToolVolume(const double k): 
  Tool(k), _m_nbSpheres(8)
{
}

// ************************************************************************* //
/// DESCRIPTION: Desctructor.
// ************************************************************************* //
ToolVolume::~ToolVolume()
{
}

// ************************************************************************* //
/// DESCRIPTION: Updates the position of the tool.
// ************************************************************************* //
void	ToolVolume::setPosition(const Matrix4d	&m)
{
  Tool::setPosition(m);
  Tool::setCoating(2.0 * (m * Vector3d(1.0, 0.0, 0.0)).norm());
}

// ************************************************************************* //
/// DESCRIPTION: Returns the field for coloring the shape.
// ************************************************************************* //
double	ToolVolume::field(const Point3d		&p,
			  const Matrix4d	&m,
			  const Matrix4d	&) const
{
  const Point3d	c(m * Point3d(0.0));
  const double	l2(Vector3d(p - c).norm2());
  
  if (4.0 * _m_k2 <= l2)
    return 0.0;

  const double	l(sqrt(l2)), dk(0.5 * l * _m_invk);

  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
/// DESCRIPTION: Returns bbox, without symmetry.
// ************************************************************************* //
BBox3d	ToolVolume::getBBox(const Matrix4d	&m,
			    const Matrix4d	&) const //invm
{
  BBox3d	ret(m * Point3d(0.0));

  ret.expand(2.0 * _m_k);
  return ret;
}

// ************************************************************************* //
Sweep	*ToolVolume::newSweep(const int			sym,
			      const Matrix4d		&msrc,
			      const Transformation	&transf) const
{
  return new SweepVolume(*this, msrc, transf, sym);
}

// ************************************************************************* //
void	ToolVolume::glRender() const
{
  glPushMatrix();
  _m_m.glMultMatrix();

  GLUquadric *quad = gluNewQuadric();

  gluSphere(quad, 0.05, 32, 24);

  const unsigned	n(32);
  const double	ainc(2.0 * M_PI / n);
  double	sina[n], cosa[n], a(0.0);

  glBegin(GL_LINE_LOOP);
  for (unsigned i(0); i < n; ++i, a += ainc)
    {
      cosa[i] = cos(a); sina[i] = sin(a);
      glNormal3d(0.0, -sina[i], cosa[i]);
      glVertex3d(0.0, cosa[i], sina[i]);
    }
  glEnd();
  glBegin(GL_LINE_LOOP);
  for (unsigned i(0); i < n; ++i)
    {
      glNormal3d(-sina[i], cosa[i], 0.0);
      glVertex3d(cosa[i], sina[i], 0.0);
    }
  glEnd();
  glBegin(GL_LINE_LOOP);
  for (unsigned i(0); i < n; ++i)
    {
      glNormal3d(-sina[i], 0.0, cosa[i]);
      glVertex3d(cosa[i], 0.0, sina[i]);
    }
  glEnd();
  gluDeleteQuadric(quad);
  
  glPopMatrix();
}
