// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sat Jan 17 18:54:28 NZDT 2004
// ************************************************************************* //
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include "swpToolZoom.h"
#include "swpSweepZoom.h"

// ************************************************************************* //
ToolZoom::ToolZoom(const double k): Tool(k)
{
}

// ************************************************************************* //
double	ToolZoom::field(const Point3d	&p,
			const Matrix4d	&m,
			const Matrix4d	&) const
{
  const Point3d	c(m * Point3d(0.0));
  const double	r2((m * Vector3d(1.0, 0.0, 0.0)).norm2());
  const double	l2((p - c).norm2());
  double	dk, l, r;

  if (l2 <= r2)
    return 1.0;
  l = sqrt(l2);
  r = sqrt(r2);
  if (_m_k2 <= sq(l - r))
    return 0.0;
  dk = (l - r) * _m_invk;
  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
BBox3d	ToolZoom::getBBox(const Matrix4d &m,
			  const Matrix4d &) const
{
  BBox3d	ret(m * Point3d(0.0));
  const double	size(4.0);

  ret.expand(size);
  return ret;
}

// ************************************************************************* //
void	ToolZoom::glRender() const
{
  glColor3d(1.0, 0.0, 0.0);

  const Point3d	c(_m_m * Point3d(0.0));

  glPushMatrix();
  glTranslated(c.x, c.y, c.z);

  GLUquadric *quad = gluNewQuadric();

  gluSphere(quad, 0.1, 32, 24);
  gluDeleteQuadric(quad);
  glPopMatrix();
}

// ************************************************************************* //
Sweep	*ToolZoom::newSweep(const int			sym,
			     const Matrix4d		&msrc,
			     const Transformation	&transf) const
{
  return new SweepZoom(*this, msrc, transf, sym);
}
