// ************************************************************************* //
// AUTHOR: Alexis Angelidis.
// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
// ************************************************************************* //
#include "swpSurfel.h"
#include "swpEdge.h"
#include "swpTriangle.h"
#include <fstream>

// ************************************************************************* //
//#define USE_RENDERMAN
#ifdef	USE_RENDERMAN
#include <ri.h>
#endif//USE_RENDERMAN

// ************************************************************************* //
/// DESCRIPTION: Renderman interface rendering.
// ************************************************************************* //
void	Triangle::glRender() const
{
  gets0().glRender();
  gets1().glRender();
  gets2().glRender();
}

// ************************************************************************* //
/// DESCRIPTION: Renderman interface rendering.
// ************************************************************************* //
void	Triangle::riRender(std::ofstream &ofs) const
{
  /*
  const Surfel	&s0(gets0()), &s1(gets1()), &s2(gets2());

  ofs << "{\n";
  ofs << "  RtPoint	triangle[3] = { "
      << "{" << s0._m_p.x << ", " << s0._m_p.y << ", " << s0._m_p.z << "},"
      << "{" << s1._m_p.x << ", " << s1._m_p.y << ", " << s1._m_p.z << "},"
      << "{" << s2._m_p.x << ", " << s2._m_p.y << ", " << s2._m_p.z << "}"
      << "};" << std::endl;
  ofs << "  RtPoint	normals[3] = { "
      << "{" << s0._m_v.x << ", " << s0._m_v.y << ", " << s0._m_v.z << "},"
      << "{" << s1._m_v.x << ", " << s1._m_v.y << ", " << s1._m_v.z << "},"
      << "{" << s2._m_v.x << ", " << s2._m_v.y << ", " << s2._m_v.z << "}"
      << "};" << std::endl;
  ofs << "RiPolygon(3, RI_P, (RtPointer)triangle, RI_N, "
      << "(RtPointer)normals, RI_NULL);" << std::endl;
  ofs << "}\n" << std::endl;
  */

  const Surfel	&s0(gets0()), &s1(gets1()), &s2(gets2());

  ofs << "Polygon \"P\" ["
      << s0._m_p.x << " " << s0._m_p.y << " " << s0._m_p.z << " "
      << s1._m_p.x << " " << s1._m_p.y << " " << s1._m_p.z << " "
      << s2._m_p.x << " " << s2._m_p.y << " " << s2._m_p.z << "] "
      << "\"N\" ["
      << s0._m_v.x << " " << s0._m_v.y << " " << s0._m_v.z << " "
      << s1._m_v.x << " " << s1._m_v.y << " " << s1._m_v.z << " "
      << s2._m_v.x << " " << s2._m_v.y << " " << s2._m_v.z << "]" << std::endl;
}

// ************************************************************************* //
void	Triangle::glRender2() const
{
  gets0().glRender2();
  gets1().glRender2();
  gets2().glRender2();
}

// ************************************************************************* //
void	Triangle::glRender3() const
{
  gets0().glRender3();
  gets1().glRender3();
  gets2().glRender3();
}

/* ************************************************************************* */
Edge &Triangle::gete(const Edge &e)
{
  if (_m_e0 == &e)
    return *_m_e1;
  else if (_m_e1 == &e)
    return *_m_e2;
  else
    return *_m_e0;
}

// ************************************************************************* //
Surfel	&Triangle::gets(const Edge &e)
{
  if (&e == _m_e0)
    return gets0();
  if (&e == _m_e1)
    return gets1();
  return gets2();
}

// ************************************************************************* //
Edge	&Triangle::gete(const Surfel	&s)
{
  if (&s == &gets0())
    return *_m_e0;
  if (&s == &gets1())
    return *_m_e1;
  return *_m_e2;
}

// ************************************************************************* //
Edge	&Triangle::gete(const Surfel	&s,
			const Edge	&e)
{
  if (&s == &gets0())
    return (&e == _m_e1)? *_m_e2: *_m_e1;
  if (&s == &gets1())
    return (&e == _m_e2)? *_m_e0: *_m_e2;
  return (&e == _m_e0)? *_m_e1: *_m_e0;
}

// ************************************************************************* //
void	Triangle::replace(Edge	&e0,
			  Edge	&e1)
{
  if (_m_e0 == &e0)
    _m_e0 = &e1;
  else if (_m_e1 == &e0)
    _m_e1 = &e1;
  else
    _m_e2 = &e1;
}

/* ************************************************************************* **
** DESCRIPTION: Computes angle (v1, v0, v2).				     **
** ************************************************************************* */
double	Triangle::computeCosAngle0() const
{
  const Surfel	&s0(gets0()), &s1(gets1()), &s2(gets2());
  Vector3d	n1, n2;

  n1 = (s1._m_p - s0._m_p);
  n1.eqUnit();
  n2 = (s2._m_p - s0._m_p);
  n2.eqUnit();
  return n1 * n2;
}

/* ************************************************************************* **
** DESCRIPTION: Computes angle (s0, s1, s2).				     **
** ************************************************************************* */
double	Triangle::computeCosAngle1() const
{
  const Surfel	&s0(gets0()), &s1(gets1()), &s2(gets2());
  Vector3d	n1, n2;

  n1 = (s2._m_p - s1._m_p);
  n1.eqUnit();
  n2 = (s0._m_p - s1._m_p);
  n2.eqUnit();
  return n1 * n2;
}

/* ************************************************************************* **
** DESCRIPTION: Computes angle (s0, s2, s1).				     **
** ************************************************************************* */
double	Triangle::computeCosAngle2() const
{
  const Surfel	&s0(gets0()), &s1(gets1()), &s2(gets2());
  Vector3d	n1, n2;

  n1 = (s0._m_p - s2._m_p);
  n1.eqUnit();
  n2 = (s1._m_p - s2._m_p);
  n2.eqUnit();
  return n1 * n2;
}

/* ************************************************************************* */
void	Triangle::computeNormalp(Vector3d &n) const
{
  Vector3d	tmpn1, tmpn2;

  tmpn1 = (gets1()._m_p - gets0()._m_p);
  tmpn2 = (gets2()._m_p - gets0()._m_p);
  n = tmpn1 ^ tmpn2;
  n.eqUnit();
}

/* ************************************************************************* */
void	Triangle::computeNormaln(Vector3d &n) const
{
  Vector3d	tmpn1, tmpn2;

  tmpn1 = (gets1()._m_p - gets0()._m_p);
  tmpn2 = (gets2()._m_p - gets0()._m_p);
  n = tmpn2 ^ tmpn1;
  n.eqUnit();
}

// ************************************************************************* //
double	Triangle::computeArea() const
{
  const double	a((gets0()._m_p - gets1()._m_p).norm());
  const double	b((gets0()._m_p - gets2()._m_p).norm());
  const double	c((gets1()._m_p - gets2()._m_p).norm());
  const double	s(0.5 * (a + b + c));

  return sqrt(s * (s - a) * (s - b) * (s - c));
}

// ************************************************************************* //
/// DESCRIPTION: Returns true if the triangle intersects the segment (v0, v1).
/// If the intersection is too close to a vertex, it is jittered.
/// That's disgusting, but who cares...
// ************************************************************************* //
bool	Triangle::intersects(const Point3d	&u0,
			     const Point3d	&u1,
			     Point3d		&vInt) const
{
  Point3d	 v[3] = { gets0()._m_p, gets1()._m_p, gets2()._m_p };
  const Vector3d n((v[1] - v[0]) ^ (v[2] - v[0]));
  const Point3d	 p(u0 + (((v[0] - u0) * n) / ((u1 - u0) * n)) * (u1 - u0));
  int	nc(0);
  bool	sh, nsh;

  if (fabs(n(1)) < fabs(n(0)) && fabs(n(2)) < fabs(n(0)))
    for (int i(0); i < 3; ++i)
      v[i](1) -= p(1), v[i](0) = v[i](2) - p(2);
  else if (fabs(n(2)) < fabs(n(1)))
    for (int i(0); i < 3; ++i)
      v[i](1) = v[i](2) - p(2), v[i](0) -= p(0);
  else
    for (int i(0); i < 3; ++i)
      v[i](0) -= p(0), v[i](1) -= p(1);
  sh = v[0](1) < 0.0;
  for (int i(0), j(1); i < 3; ++i, j = ((i + 1) % 3), sh = nsh)
    if (sh != (nsh = v[j](1) < 0.0))
      if (0.0 < v[i](0) && 0.0 < v[j](0))
	++nc;
      else if (0.0 < v[i](0) || 0.0 < v[j](0))
	if (0.0 < v[i](0) - v[i](1) * (v[j](0) - v[i](0)) /(v[j](1) - v[i](1)))
	  ++nc;
  vInt = p;
  return nc % 2 && ((u0 - p) * n) * ((u1 - p) * n) < 0.0;
}

// ************************************************************************* //
/// DESCRIPTION: The vertices (v0, v1, v2) are substituted for the vertices
///  (u0, u1, u2).
// ************************************************************************* //
void	Triangle::reconnect(Surfel	&u0,
			    Surfel	&u1,
			    Surfel	&u2)
{
  Surfel	&v0(gets0()), &v1(gets1()), &v2(gets2());
  Edge		&e0(gete0()), &e1(gete1()), &e2(gete2());

  Edge::circulator	ecir, ebeg;
  std::list<Edge *>	list0, list1, list2;
  std::list<Edge *>::iterator	eit, eend;

  ecir = ebeg = Edge::circulator(v0, e2);
  for (++ecir; ecir != ebeg; ++ecir)
    list0.push_back(&(*ecir));
  ecir = ebeg = Edge::circulator(v1, e0);
  for (++ecir; ecir != ebeg; ++ecir)
    list1.push_back(&(*ecir));
  ecir = ebeg = Edge::circulator(v2, e1);
  for (++ecir; ecir != ebeg; ++ecir)
    list2.push_back(&(*ecir));
  for (eit = list0.begin(), eend = list0.end(); eit != eend; ++eit)
    (*eit)->replace(v0, u0);
  for (eit = list1.begin(), eend = list1.end(); eit != eend; ++eit)
    (*eit)->replace(v1, u1);
  for (eit = list2.begin(), eend = list2.end(); eit != eend; ++eit)
    (*eit)->replace(v2, u2);
  e0.replace(v1, u1);
  e0.replace(v2, u2);
  e1.replace(v2, u2);
  e1.replace(v0, u0);
  e2.replace(v0, u0);
  e2.replace(v1, u1);
}

// ************************************************************************* //
Triangle::circulator::circulator(Surfel		&s,
				 Triangle	&t): _m_s(&s), _m_t(&t)
{
  if (&s == &t.gets0())
    _m_e = t._m_e1;
  else if (&s == &t.gets1())
    _m_e = t._m_e2;
  else
    _m_e = t._m_e0;
}

// ************************************************************************* //
/// DESCRIPTION: Increments the circulator.
// ************************************************************************* //
Triangle::circulator &Triangle::circulator::operator++()
{
  _m_t = &_m_e->gett(*_m_t);
  _m_e = &_m_t->gete(*_m_s, *_m_e);
  return *this;
}

/* ************************************************************************* **
** DESCRIPTION: Decrements the circulator.				     **
** ************************************************************************* */
Triangle::circulator &Triangle::circulator::operator--()
{
  _m_e = &_m_t->gete(*_m_s, *_m_e);
  _m_t = &_m_e->gett(*_m_t);
  return *this;
}

/* ************************************************************************* **
** DESCRIPTION: Tests if the circulators are different.			     **
** WARNING: The circulators MUST be 'compatible'.			     **
** ************************************************************************* */
bool	Triangle::circulator::operator!=(const Triangle::circulator &tit) const
{
  return _m_t != tit._m_t;
}

/* ************************************************************************* **
** DESCRIPTION: Tests if the circulators are different.			     **
** WARNING: The circulators MUST be 'compatible'.			     **
** ************************************************************************* */
bool	Triangle::circulator::operator==(const Triangle::circulator &tit) const
{
  return _m_t == tit._m_t;
}
