// ************************************************************************* //
// AUTHOR: Alexis Angelidis.
// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
// ************************************************************************* //
#ifndef __SWPTRIANGLE_HH__
#define __SWPTRIANGLE_HH__

// ************************************************************************* //
class	Edge;
class	Shape;
class	SweepPinch;

// ************************************************************************* //
class	Triangle
{
  friend class	Shape;
  friend class	SweepPinch;
  Edge		*_m_e0, *_m_e1, *_m_e2;
private:
  std::list<Triangle *>::iterator	_m_meshHandle;
public:
  class circulator
  {
    Surfel	*_m_s;
    Edge	*_m_e;
    Triangle	*_m_t;
  public:
    circulator() {}
    circulator(const circulator &ec): 
      _m_s(ec._m_s), _m_e(ec._m_e), _m_t(ec._m_t) {}
    circulator(Surfel &s, Triangle &t);
    circulator	&operator++();
    circulator	&operator--();
    bool	operator!=(const circulator &eit) const;
    bool	operator==(const circulator &eit) const;
    Triangle	&operator*() { return *_m_t; }
    Triangle	*operator->() { return _m_t; }
  };
public:
  Triangle(Edge &e0, Edge &e1, Edge &e2): _m_e0(&e0), _m_e1(&e1), _m_e2(&e2) {}

  Surfel	&gets0()
  {
    return (((_m_e1->_m_s0 == _m_e0->_m_s0) || (_m_e1->_m_s0 == _m_e0->_m_s1))?
	  *_m_e1->_m_s1: *_m_e1->_m_s0);
  }
  Surfel	&gets1()
  {
    return (((_m_e2->_m_s0 == _m_e1->_m_s0) || (_m_e2->_m_s0 == _m_e1->_m_s1))?
  	  *_m_e2->_m_s1: *_m_e2->_m_s0);
  }
  Surfel	&gets2()
  {
    return (((_m_e0->_m_s0 == _m_e2->_m_s0) || (_m_e0->_m_s0 == _m_e2->_m_s1))?
  	  *_m_e0->_m_s1: *_m_e0->_m_s0);
  }
  const Surfel	&gets0() const
  {
    return (((_m_e1->_m_s0 == _m_e0->_m_s0) || (_m_e1->_m_s0 == _m_e0->_m_s1))?
  	  *_m_e1->_m_s1: *_m_e1->_m_s0);
  }
  const Surfel	&gets1() const
  {
    return (((_m_e2->_m_s0 == _m_e1->_m_s0) || (_m_e2->_m_s0 == _m_e1->_m_s1))?
  	  *_m_e2->_m_s1: *_m_e2->_m_s0);
  }
  const Surfel	&gets2() const
  {
    return (((_m_e0->_m_s0 == _m_e2->_m_s0) || (_m_e0->_m_s0 == _m_e2->_m_s1))?
  	  *_m_e0->_m_s1: *_m_e0->_m_s0);
  }

  Edge		&gete0() { return *_m_e0; }
  Edge		&gete1() { return *_m_e1; }
  Edge		&gete2() { return *_m_e2; }
  const Edge	&gete0() const { return *_m_e0; }
  const Edge	&gete1() const { return *_m_e1; }
  const Edge	&gete2() const { return *_m_e2; }
  Surfel	&gets(const Edge &e);
  void		glRender() const;
  void		riRender(std::ofstream &ofs) const;
  void		glRender2() const;
  void		glRender3() const;
  Edge		&gete(const Edge &e);
  Edge		&gete(const Surfel &s, const Edge &e);
  Edge		&gete(const Surfel &s);
  void		replace(Edge &e0, Edge &e1);
  double	computeCosAngle0() const;
  double	computeCosAngle1() const;
  double	computeCosAngle2() const;
  void		computeNormaln(Vector3d &v) const;
  void		computeNormalp(Vector3d &v) const;
  double	computeArea() const;
  bool		intersects(const Point3d &v0, const Point3d &v1,
			   Point3d &vInt) const;
  void		reconnect(Surfel &u0, Surfel &u1, Surfel &u2);
};

// ************************************************************************* //
/// DESCRIPTION: Class for handling three indices.
// ************************************************************************* //
template <class TFIRST, class TSECOND, class TTHIRD>
struct	triplet
{
  TFIRST	first;
  TSECOND	second;
  TTHIRD	third;
public:
  triplet(const TFIRST &t0, const TSECOND &t1, const TTHIRD &t2):
    first(t0), second(t1), third(t2) {}
};

// ************************************************************************* //
template <class TFIRST, class TSECOND, class TTHIRD>
triplet<TFIRST, TSECOND, TTHIRD>	make_triplet(const TFIRST	&t0,
						     const TSECOND	&t1,
						     const TTHIRD	&t2)
{
  return triplet<TFIRST, TSECOND, TTHIRD>(t0, t1, t2);
}

// ************************************************************************* //
#endif//__SWPTRIANGLE_HH__
